#!/usr/bin/eperl

use vars qw( %bin $dumpDir $ptrArgs $pwd $workDir);
use strict;
use DB_File;

my $VERSION='1.0.1';

my $dbName='appldb';
my $dbHost='localhost';
my $dbType='Pg';

#
# parse command line
#
my %arg_opts = ('--help|-h'=>'',
				'--debug|-d'=>'i?',
				'--get-status|-s'=>'',
				'--dump-accounts|-dc'=>'s',
				'--dump-domains|-dd'=>'s',
				'--dump-all|-da'=>'',
				'--get-content-list|-lc'=>'',
				'--no-content|-nc'=>'',
				'--no-compress|-nz'=>'',
				'--output|-o'=>'s',
				'--status-file|-sf'=>'s',
				'--version|-v'=>'',
			   );

%bin = ('tar'=>'tar','split'=>'split','cp'=>'cp','rm'=>'rm','echo'=>'echo',
		'mysql'=>'mysql','mysqldump'=>'mysqldump',
		'psql'=>'psql','uptime'=>'uptime','uname'=>'uname',
		'cat'=>'cat','pwd'=>'pwd','md5sum'=>'md5sum',
		'grep'=>'grep','find'=>'find','gzip'=>'gzip','wc'=>'wc','rpm'=>'rpm',
		'locale'=>'locale', 'iconv'=>'iconv',
		'python'=>'python');

my $returnCode = 0;

#
# global variables
#

my $kilobyte = 1024;
my $megabyte = $kilobyte*$kilobyte;

my $defaultDomainsAccount = "DefaultEnsimDomains";
my $defaultDomainsFullname = "Default Ensim Domains";

my $ensimVersion;

my $ignoredUsers_ = {'root' => '1',
					 'bin' => '1',
					 'daemon' => '1',
					 'adm' => '1',
					 'lp' => '1',
					 'sync' => '1',
					 'shutdown' => '1',
					 'halt' => '1',
					 'mail' => '1',
					 'uucp' => '1',
					 'operator' => '1',
					 'games' => '1',
					 'gopher' => '1',
					 'ftp' => '1',
					 'news' => '1',
					 'mysql' => '1',
					 'nobody' => '1',
					 'zope' => '1',
					 'majordomo' => '1',
					 'tomcat4' => '1',
					 'apache' => '1',
					 'ensimrootmail' => '1',
					 'sshd' => '1',
					};

my %ignoredUsers = %{$ignoredUsers_};



my $quotaHome = "/home/virtual";

my $quotaDev;
if (eval "require Quota;") {
  $quotaDev = Quota::getqcarg($quotaHome);
}

#
# end global variables
#



#==============================================================
#
# common subroutines
#

sub printToLog {
  my ($msg, $noeol) = @_;
  if ($msg) {
	print $msg;
  }
  unless ($noeol) {
	print "\n";
  }
}

sub printToError {
  my ($msg, $noeol, $errCode) = @_;
  if ($msg) {
	print STDERR $msg;
  }
  unless ($noeol) {
	print STDERR "\n";
  }
  if (defined($errCode)) {
    $returnCode = $errCode; ## translate an error code to the return code
  } else {
    $returnCode = 127;
  }
}

#BEGINNOTINRAQ2

sub makeFileReader
{
	my ($filename, $doNotFail) = @_;
	local (*INPUT);

	unless (-T $filename) {

		if (defined($doNotFail)) {
			&printToLog("Error: file '$filename' is not a text file");
		} else {
			&printToError("Error: file '$filename' is not a text file");
		}

		return undef;
	}

	unless (open(INPUT,"<$filename")) {

		binmode INPUT;
		if (defined($doNotFail)) {
			&printToLog("Error: unable to open file '$filename' to read");
		} else {
			&printToError("Error: unable to open file '$filename' to read");
		}

		return undef;
	}

	my $input = *INPUT;

	my $this = {
		'READ' => sub {
				my $a = <$input>;
				return $a;
			},

		'CLOSE' => sub { close($input); }

	};

	return $this;
}

sub makeArrayReader {
  my (@arr) = @_;
  my $size= 0;
  my $count = scalar(@arr);

  my $this={'READ'=>sub {
			  if($count == $size) {
				return undef;
			  }

			  return $arr[$count++];
			},
			'CLOSE'=>sub {
			}
		   };
  return $this;
}

sub makeParser {
  my $reader;
  my ($this,$ptrRows,$keySep,$valSep,$valKeySep,$comment,
      $typeRows,$maxIndex,$rowIndex);
  {
	$reader = shift;
	unless(defined($reader)) {
	  return undef;
	}

    my %args=(@_);
    $keySep = $args{'KEYSEPARATOR'};
    $valSep = $args{'VALUESEPARATOR'};
    $valKeySep = $args{'KEYVALUESEPARATOR'};
  }
  $this={'READER'=>sub {
		   if(@_){
			 $reader=shift;
		   }
		   return $reader;
		 },
		 'FILE'=>sub {
		   if(@_){
			 $reader = makeFileReader(@_);
		   }
		   return undef;
		 },
		 'KEYSEPARATOR'=>sub{
		   if(@_){
			 $keySep=shift;
		   }
		   return $keySep;
		 },
		 'VALUESEPARATOR'=>sub{
		   if(@_){
			 $valSep=shift;
		   }
		   return $valSep;
		 },
		 'KEYVALUESEPARATOR'=>sub{
		   if(@_){
			 $valKeySep=shift;
		   }
		   return $valKeySep;
		 },
		 'COMMENT'=>sub{
		   if(@_){
			 $comment=shift;
		   }
		   return $comment;
		 },
		 'PARSE'=>sub{
		   my %args=(@_);

		   $keySep=$args{'KEYSEPARATOR'} if exists $args{'KEYSEPARATOR'};
		   $valSep=$args{'VALUESEPARATOR'} if exists $args{'VALUESEPARATOR'};
		   $valKeySep=$args{'KEYVALUESEPARATOR'} if exists $args{'KEYVALUESEPARATOR'};
		   $comment=$args{'COMMENT'} if exists $args{'COMMENT'};

		   $ptrRows = &parseFile('reader'=>$reader,'keysep'=>$keySep,
								 'valsep'=>$valSep,'keyvalsep'=>$valKeySep,
								 'comment'=>$comment);
		   $typeRows = ref($ptrRows);
		   if($typeRows=~/ARRAY/){
			 $rowIndex=-1;
			 $maxIndex=scalar(@{$ptrRows})-1;
		   }
		   return $ptrRows;
		 },
		 'ROW'=>sub{
		   if($typeRows=~/HASH/){
			 my $key=shift;
			 return $ptrRows->{$key};
		   }elsif($typeRows=~/ARRAY/){
			 if($rowIndex<$maxIndex){
			   $rowIndex++;
			   return $ptrRows->[$rowIndex];
			 }else{
			   return undef;
			 }
		   }else{
			 return undef;
		   }
		 },
		};

  return $this;
}

sub makeFileParser {
  my $filename = shift;
  return makeParser(makeFileReader($filename), @_);
}

sub makeSafeFileParser {
  my $filename = shift;
  return makeParser(makeFileReader($filename, 'doNotFail'), @_);
}

sub makeStringParser {
  my $content = shift;
  return makeParser(makeArrayReader($content), @_);
}

sub parseFile {
  my %arg = (@_);
  my $reader = $arg{'reader'};
  my $keySep = $arg{'keysep'};
  my $valSep = $arg{'valsep'};
  my $valKeySep = $arg{'keyvalsep'};
  my $comment = $arg{'comment'};

  my ($ret,$key,$value,$reKey,$reValue,@values,$keyBlank,$valBlank,
     $vkey,$vvalue,$reKeyValue,$reComment);

  if($keySep){
    $ret = {};
    if(ref($keySep)=~/regexp/i || ($keySep eq ' ')){
      $reKey = $keySep;
    }else{
      $reKey = qr/$keySep/;
    }
  }else{
    $ret = [];
  }
  if($valSep){
    if(ref($valSep)=~/regexp/i || ($valSep eq ' ')){
      $reValue = $valSep;
    }else{
      $reValue = qr/$valSep/;
    }
  }
  if($valKeySep){
    if(ref($valKeySep)=~/regexp/i || ($valKeySep eq ' ')){
      $reKeyValue = $valKeySep;
    }else{
      $reKeyValue = qr/$valKeySep/;
    }
  }
  if($comment){
    if(ref($comment)=~/regexp/i){
      $reComment = $comment;
    }else{
      $reComment = qr/$comment/;
    }
  }

  while($_ = $reader->{'READ'}->()){
    chomp;
    next unless $_;
    if($comment && /$reComment/){  ## commentar
	  next;
    }
    if ($keySep){
      ($key,$value)=split($reKey,$_,2);
      if($key){
		$key=~s/^\s+//;
		$key=~s/\s+$//;
		if($valSep){

		  if ($valKeySep){
			$ret->{$key}={};
			foreach $value (split($reValue,$value)){
			  ($vkey,$vvalue)=split($reKeyValue,$value);
			  $vkey=~s/^\s+//;
			  $vkey=~s/\s+$//;
			  $ret->{$key}->{$value}=$vvalue;
			}
		  }else{
			push @{$ret->{$key}},split($reValue,$value);
		  }
		}else{
		  $value=~s/^\s+//;
		  $value=~s/\s+$//;
		  $ret->{$key}=$value;
		}
      }
    }else{
      if($valSep){
		push @{$ret},[split($valSep,$_)];
      }else{
		push @{$ret},$_;
      }
    }
  }

  $reader->{'CLOSE'}->();

  return $ret;
}

#ENDNOTINRAQ2

sub makePasswordNode {
  my $value= shift;
  my $type = shift||'encrypted';

  my $item = &makeXmlNode('password');
  $item->{'CONTENT'}->($value);  ## html encode
  $item->{'ATTRIBUTE'}->('type',$type);
  return $item;
}

sub mangleName {
  my $name = shift;

  if($name =~ /^\d/) {
    $name = 'user' . $name;
  }
  $name =~ s/\./_/g;

  return $name;
}

sub makeSysUserNode {
  my ($name, $password, $type) = @_;

  unless(defined($name) && defined($password) && defined($type)) {
	printToError("makeSysUserNode: Username, password or password type is not set.");
  }

  my $sysUserNode = makeXmlNode('sysuser');
  $sysUserNode->{'ATTRIBUTE'}->('name', mangleName($name));
  $sysUserNode->{'ADDCHILD'}->(makePasswordNode($password, $type));

  return $sysUserNode;
}

# Returns home directory for the system account $user
sub getHomeDir {
	my ($user) = @_;
	my $home;
	my @pw = getpwnam($user);
	if (-1 != $#pw) {
		$home = $pw[7];
	}
	unless ($home) {
		# how it is by default
		$home = '/home/$account';
	}
	return $home;
}

# Returns password for system account $user
sub getSystemPassword {
	my ($user) = @_;
	my $password;
	my @pw = getpwnam($user);
	if (-1 != $#pw) {
		$password = $pw[1];
	}
	unless ($password) {
		$password = '';
	}
	return $password;
}

sub createUserNameFromDomain {
  my ($name) = @_;

  if ($name =~ /^\d/) {
	$name = 'u' . $name;
  }

  $name =~ s/\./_/g;
  $name =~ s/^www\.//g;
  $name =~ s/[aoeui]//g if length($name) > 16;
  $name = substr($name, 0, 16) if length($name) > 16;

  return $name;
}

my $compress = 1;

sub setCompress {
  ($compress) = @_;
}

sub makeDumpDb
{
  my ($dbName, $dbType, $dbUser, $dbPasswd, $dbHost, $ptrDomainBin, $variables, $domainSocket) = @_;
  $dbType ||= 'mysql';

  my ($dumpFile,$dumpSql,$dumpName,$cmd,$gzip,$split,$mysql,$psql, $retCode,$out);

  $dumpName = "$dbName.$dbType.sql";
  $gzip = $compress == 1 ? ($bin{'gzip'} || 'gzip') : undef;

  if (-x $gzip) {
	$dumpName .= ".gz";
  }
  $dumpFile="$dumpDir/$dumpName";

  printToLog(">>Checking database presence");

  if ($dbType eq 'mysql') {
	$mysql = $bin{'mysql'} || 'mysql';
	if (-x $mysql) {
	  $cmd  = "$mysql -u '$dbUser'";
	  $cmd .= " -h $dbHost" if $dbHost;
	  $cmd .= " -p'$dbPasswd'" if $dbPasswd;
	  $cmd .= " -e '' '$dbName' 2>&1";

	  $out = `$cmd`;
	  chomp $out;
	  if ($out) {
		$retCode = $?>>8;
		printToLog("warning: mysql datatbase '$dbName' not found. mysql has returned '$retCode'");
		return "";
	  }
	}
  } elsif ($dbType eq 'Pg') {
	$psql = $bin{'psql'} || 'psql';
	if (-x $psql) {
	  $cmd  = "export PGUSER='$dbUser'";
	  $cmd .= " PGPASSWORD='$dbPasswd'" if $dbPasswd;
	  $cmd .= "; $psql '$dbName' ";
	  $cmd .= " -h $dbHost" if $dbHost;
	  $cmd .= " -c '' 2>&1";

	  $out = `$cmd`;
	  chomp $out;
	  if ($out) {
		$retCode = $?>>8;
		printToLog("warning: PostgreSQL datatbase '$dbName' not found. psql has returned '$retCode'");
		return "";
	  }
	}
  }

  printToLog("..done");

  unless (exists($ptrArgs->{'no-content'})) {
	$gzip = $bin{'gzip'} || 'gzip';
	$split = $bin{'split'} || 'split';

	if ($dbType eq 'mysql') {
	  my ($prefx, $postfix) = ("", "");
	  my $commonprefix = "LD_LIBRARY_PATH=$variables " if $variables;

	  my $commonopts;
	  $commonopts = " --socket='$domainSocket'" if $domainSocket;
	  $commonopts .= " -h $dbHost" if $dbHost;
	  $commonopts .= " -p'$dbPasswd'" if $dbPasswd;
	  $commonopts .= " -u '$dbUser' '$dbName'";

	  my $dumpcmd;
	  if ($ptrDomainBin) {
		$dumpcmd = $ptrDomainBin->{'mysqldump'};
	  } else {
		$dumpcmd = $bin{'mysqldump'} || 'mysqldump';
	  }

	  my $dumpopts = " --quick --quote-names --add-drop-table";

	  my $mysqlversion = `$commonprefix $bin{'mysql'} $commonopts --version`;
	  if ($mysqlversion =~ /Distrib (\d+)\.(\d+)\./) {
		if ($1 > 4 or ($1 == 4 and $2 > 0)) {
		  $dumpopts .= " --compatible=mysql323";
		}
	  }

	  $cmd = $commonprefix . $dumpcmd . $commonopts . $dumpopts;

	} elsif ($dbType eq 'Pg') {
	  $cmd  = "export PGUSER='$dbUser'";
	  $cmd .= " PGPASSWORD='$dbPasswd'" if($dbPasswd);
	  $cmd .= ' ; ' . ($bin{'pg_dump'} || 'pg_dump') . ' -d -O';
	  $cmd .= " -h '$dbHost'" if $dbHost;
	  $cmd .= " '$dbName'";
	}

	if ($compress and $gzip and -x $gzip) {
	  $cmd .= " | $gzip";
	}
	if (-x $split) {
	  $cmd .= " | $split -b 2040m - $dumpFile.";
	} else {
	  $cmd .= " > $dumpFile";
	}

	local $SIG{'PIPE'} = 'DEFAULT' if $SIG{'PIPE'} eq 'IGNORE';

	unless (system($cmd) == 0) {
	  printToError("Error execute '$cmd'");
	  return undef;
	}
  }
  return $dumpName;
}

sub addDbUsers {
  my ($dbNode, $dbName, $wrapUserMysql) = @_;

  printToLog("Dupming $dbName database users");

  $dbName =~ s/_/\\_/g;

  my (@users, $dbUser, $ptrRow);
  my $sql = "SELECT DISTINCT User FROM db WHERE Db = '$dbName'";
  $wrapUserMysql->{'EXECUTE'}->($sql);
  while ($ptrRow = $wrapUserMysql->{'FETCHROW'}->()) {
		push @users, (@{$ptrRow})[0];
  }
  $wrapUserMysql->{'FINISH'}->();

  foreach $dbUser (@users) {
		addDbUser($dbNode, $dbUser, $wrapUserMysql);
  }
}

sub addDbUser {
  my ($dbNode, $user, $wrapUserMysql, $passwordNode) = @_;

  &printToLog("Dumping mysql user '$user'");

	my ( $item );

  if ( ref( $wrapUserMysql ) =~ /HASH/ &&
			 ref( $wrapUserMysql->{'EXECUTE'} ) =~ /CODE/ ) {

    my ($sql,$ptrRow,@accesshosts,$accesshost);

    $sql = "SELECT password,host FROM user WHERE user='$user'";

    if ( $wrapUserMysql->{'EXECUTE'}->( $sql ) ) {

			$item = makeXmlNode( 'dbuser' );
			$item->{'ATTRIBUTE'}->('name', $user);

      while ( $ptrRow = $wrapUserMysql->{'FETCHROW'}->() ) {
				unless ( ref($passwordNode) =~ /HASH/ ) {
					$passwordNode = makePasswordNode( $ptrRow->[0] );
				}
				push @accesshosts, $ptrRow->[1];
      }

			if ( $passwordNode ) {
				$item->{'ADDCHILD'}->( $passwordNode );
			}

			foreach $accesshost ( @accesshosts ) {
				$item->{'ADDCHILD'}->( makeXmlNode( 'accesshost', $accesshost ) );
			}

			$dbNode->{'ADDCHILD'}->( $item );

    }

    $wrapUserMysql->{'FINISH'}->();
  }

	return $item;
}

my $lastDumpId = 0;

sub generateDumpId {
  return $$ . "-" . $lastDumpId++;
}

#
# sub makeDumpFile - make a dump of dir tree (use 'tar')
#
# Arguments:
#          $destFile - name of destination file.
#                      If the file has not '.tgz' or '.tar.gz' extension
#                      '.tgz' will be added
#          $scrDir   - path to source dir
#
#          $ptrNames - pointer to an array with names of files and dirs to
#                      include into dump
#                      can be a single name
#                      by default '.' - include all entries
#
#          $ptrExcludes - pointer to an array with names of files and dirs to
#                      exclude from dump
#                      can be a single name
#
#          $extraParams - extra params to run tar
#                         ('-h' - don't dump symlinks; dump the files they point to)
#
sub makeDumpFile {
  my ($destFile, $srcDir, $ptrNames,$ptrExclude,$extraParams)=@_;

  my @destPath = split(/\//, $destFile);
  if (length($destPath[$#destPath]) > 32) {
	$destPath[$#destPath] = generateDumpId();
  }
  $destFile = join('/', @destPath);

  my $tar = $bin{'tar'}||'tar';
  my $split = $bin{'split'}||'split';
  my $gzip = $compress == 1 ? ($bin{'gzip'} || 'gzip') : undef;

  my ($cmd,$ret,@names,@exclude);

  if(ref($ptrNames)=~/ARRAY/){
    @names = @{$ptrNames};
  }else{
    @names=($ptrNames) if $ptrNames;
  }

  if(ref($ptrExclude)=~/ARRAY/){
    @exclude = @{$ptrExclude};
  }else{
    @exclude=($ptrExclude) if $ptrExclude;
  }

  unless(-x $tar ){
    &printToError("Error: '$tar' is not executable");
    return undef;
  }

#
# check extension
#
  if(-x $gzip){
    unless($destFile=~/\.tgz$/ ||$destFile=~/\.tar\.gz$/){
      $destFile .= '.tar.gz';
    }
  }else{
    unless($destFile=~/\.tar$/){
      $destFile .= '.tar';
    }
  }
#
# end check extension
#

  unless( exists $ptrArgs->{'no-content'}){

    $cmd=$tar;
    if($srcDir){
      if( -d $srcDir){
		$cmd.=" -C $srcDir -f - -c --ignore-failed-read ";
      }else{
		return undef;
      }
    }


    if($extraParams){
      $cmd .= $extraParams.' ';
    }

    if(@names){
      if($srcDir){
		$cmd .= join( ' ', map {s/^\Q$srcDir\E\///;$_} @names); ## extract domain name
      }else{
		$cmd .= join( ' ', @names );
      }
    }else{
      $cmd .= '.';  ## dump all files & dirs
    }
    if(@exclude){
      $cmd .= ' '.join( ' ', map {s/^\Q$srcDir\E\///;'--exclude '.$_} @exclude); ## extract domain name
    }
    if(-x $gzip){
      $cmd .= " | $gzip";
    }
    if(-x $split){
      $cmd .= " | $split -b 2040m - $destFile.";
    }else{
      $cmd .= " >$destFile";
    }

		local $SIG{'PIPE'} = 'DEFAULT' if $SIG{'PIPE'} eq 'IGNORE';

    unless(system($cmd) == 0){
      &printToError("Error execute command: '$cmd': $!");
      $destFile='';
    }
  }

  $destFile=~s/^\Q$dumpDir\E\///;  ### remove dir name
  return $destFile;
}

sub getDumpDir {
  my ($subdir,$clean) = @_;
  my ($dumpdir);

  if($subdir=~/^\//){
    $dumpdir=$subdir;
  }else{
    my $util = $bin{'pwd'}||'pwd';
    $dumpdir = $pwd||`$util`;
    chomp $dumpdir;
    if($subdir){
      $dumpdir.='/'.$subdir;
    }
  }
  if(-d $dumpdir && $clean){
    unless(system("rm -rf $dumpdir")==0){
      &printToError("Error: can not clean temporary dir '$dumpdir': $!");
    }
  }
  unless(-d $dumpdir){
    mkdir ($dumpdir,0755) or
      die "Error: can not create temporary dir '$dumpdir': $!\n";
  }
  return $dumpdir;
}

sub printXml {
  my ($root,$outFh,$dtd) = @_;
  unless($outFh){
    $outFh = *STDOUT;
  }

  unless(ref($root)=~/HASH/ && ref($root->{'PRINT'})=~/CODE/){
    print $outFh "Error: a xml root is not found\n";
    &printToError("Error: sub printXml: bad arguments");
    return 0;
  }
  print $outFh "<?xml version='1.0' encoding=\"UTF-8\"?>\n";

  my  $doctype = $root->{'NAME'}->();
  if($dtd){
    print $outFh "<!DOCTYPE $doctype SYSTEM '$dtd'>\n";
  }
  $root->{'PRINT'}->($outFh,'  ');
  return 1;
}

#
# class for handle with XML
#
sub makeXmlNode {
  my ($this,$parent,%attributes,@content);
  my ($name,$plain)=@_;

  $plain = toUtf8(encodeHtml($plain));

  unless($name||$plain){
    return undef;
  }
  $this = {
	   'ATTRIBUTE' => sub {
	     my $key = shift;
	     if(@_){
	       my $value=shift;
	       if(@_){
					 $attributes{$key} = $value;
	       }else{
					 $attributes{$key} = toUtf8(normalizeAttribute($value));
	       }
	     }
	     return $attributes{$key};
	   },
	   'PARENT' => sub {
	     if(@_){
	       my $arg = shift;
	       if(ref($arg)=~/HASH/){
					 $parent=shift;
	       }
	     }
	     return $parent;
	   },
	   'NAME' => sub {
	     if(@_){
	       $name = shift;
	     }
	     return $name;
	   },
	   'CONTENT' => sub {
	     if(@_){
	       $plain = toUtf8(encodeHtml(shift));
	     }
	     return $plain;
	   },
	   'ADDCHILD'=> sub {
	     my $kind = shift;
	     if(
					(ref($kind)=~/HASH/) &&
					(ref($kind->{'PARENT'})=~/CODE/)
	       ) {
	       push @content,$kind;
	       $kind->{'PARENT'}->($this);
	     }
	     return $#content;
	   },
	   'KIND'=> sub {
	     my $index = shift;
			 my ($ret);
			 if ( $index =~ /^\d+$/ ) {
				 if($index >= 0 && $index <= $#content ) {
					 $ret =  $content[$index];
				 }
			 } else {
				 foreach my $item (@content){
					 if ( $item->{'NAME'}->() eq $index ) {
						 $ret = $item;
						 last;
					 }
				 }
			 }
			 return $ret;
	   },
	   'KINDSREF'=> sub {
			 my $name = shift;
			 my $ret = \@content;
			 if ( $name ) {
				 $ret = [];
				 foreach my $item (@content) {
					 if ( $item->{'NAME'}->() eq $name ) {
						 push @{$ret}, $item;
					 }
				 }
			 }
	     return $ret;
	   },
	   'ATTRIBUTESREF' => sub {
	     return \%attributes;
	   },
	   'PRINT'=>sub {
	     my $fh = shift || *STDOUT;
	     my $addPrefix = shift || '';
	     my $prefix = shift || '';
	     if($name){
	       my ($out,$item);
	       if(@content || defined($plain)){

					 $out = $prefix . &getTag($name,\%attributes,'start');
					 print $fh $out;
					 if(@content){
						 print $fh "\n";
						 foreach $item (@content){
							 if(ref($item)=~/HASH/ && ref($item->{'PRINT'})=~/CODE/){
								 $item->{'PRINT'}->($fh,$addPrefix,$prefix.$addPrefix);
							 }
						 }
					 }
					 if(defined($plain)){
						 print $fh $plain;
						 if(@content){
							 print $fh "\n$prefix";
						 }
					 }else{
						 if($prefix){
							 print $fh $prefix;
						 }
					 }
					 $out = &getTag($name,\%attributes,'stop') . "\n";
					 print $fh $out;

	       }else{
					 $out = $prefix . &getTag($name,\%attributes) . "\n";
					 print $fh $out;
	       }
	     }else{
	       if(defined($plain)){
					 print $fh $plain;
	       }
	     }
	   }
	 };

  return $this;
}

sub getTag {
  my ($name,$ptrAttr,$mode) = @_;
  my ($out,$key,$value);
  $mode||='';
  if ($mode eq 'stop'){
    $out="</$name>";
  }else{
    $out="<$name";
    while (($key,$value)=each(%{$ptrAttr})){
      $out .= ' '.$key;
      if (defined($value)){
		$out .= '="'.$value.'"';
      }
    }
    if($mode eq 'start'){
      $out .= ">";
    }else{
      $out .= "/>";
    }
  }
  return $out;
}

#
# end class for handle with XML
#


#
# class for handle with database
#
sub getDbConnect {
  my $ptrBin = shift;
  my $myDbType = shift;
  my $myDbAddress = shift ;
  my $myDbUser = shift ;
  my $myDbPasswd = shift ;
  my $myDbDB = shift ;
  my $myDbHost = shift ;

  #added by Honig for Sphera
  my $myDbSocket = shift;
  my $variables = shift;

  my ($dbh,$this,$sth);
  my ($shell,$args);
  my $echo = $ptrBin->{'echo'}||'echo';
  my ($sql,@result,$resultIndex,$ptrResult);

  unless(defined($myDbPasswd) && defined($myDbUser)){
    if($myDbType eq 'mysql'){
      foreach my $myCnf ('/etc/my.cnf',$ENV{'HOME'}.'/.my.cnf'){
		next unless (-T $myCnf);
		next unless open(CNF,"<$myCnf");
		binmode CNF;
		my($key,$value);
		my $state=0;
		while (<CNF>){
		  next unless $_;
		  next if (/^#/);
		  if($state){
			if(/^\s*\[/){
			  last;
			}
			if(/^\s*(\S+)\s*=\s*(\S+)/){
			  $key=$1;
			  $value=$2;
			  if($key =~ /^user$/){
				$myDbUser = $value;
				# sometimes strings are quoted
				$myDbUser =~ s/^('|")(.*)\1$/$2/;
			  }elsif($key =~ /^pass$/){
				$myDbPasswd = $value;
				# sometimes strings are quoted
				$myDbPasswd =~ s/^('|")(.*)\1$/$2/;
			  }
			}
		  }else{
			if(/^\s*\[client\]/){
			  $state=1;
			}
		  }
		}
		close (CNF);
      }
    }elsif($myDbType eq 'Pg'){
      $myDbPasswd ||= $ENV{'PGPASSWORD'};
      $myDbUser ||= $ENV{'PGUSER'};
    }
  }

  if (eval "require DBI"){
    if (eval "require DBD::$myDbType"){
      if($myDbAddress){
		if($myDbUser && $myDbPasswd){
		  $dbh = DBI->connect($myDbAddress,$myDbUser,$myDbPasswd) or
			die "Error: unable to connect to database: $DBI::errstr\n";
		}elsif($myDbUser){
		  $dbh = DBI->connect($myDbAddress,$myDbUser) or
			die "Error: unable to connect to database: $DBI::errstr\n";
		}else{
		  $dbh = DBI->connect($myDbAddress) or
			die "Error: unable to connect to database: $DBI::errstr\n";
		}
      }
    }
  }
#  unless($dbh){
#    if ($myDbType eq 'Pg'){
#      if(eval "require Pg"){
#        $myDbAddress="user=$myDbUser dbname=$myDbDB";
#        if($myDbPasswd){
#          $myDbAddress.=" password=$myDbPasswd";
#        }
#        $dbh = Pg::connectdb($myDbAddress);
#      }
#    }
# }
	unless ($dbh) {

		if ($myDbType eq 'mysql') {
			$shell = $ptrBin->{'mysql'} || 'mysql';
			$args  = " --socket='$myDbSocket'" if ($myDbSocket);
			if ($myDbPasswd) {
				$args .= " -B -u '$myDbUser' -p'$myDbPasswd' '$myDbDB' ";
			} else {
				$args .= " -B -u $myDbUser $myDbDB ";
			}
			$shell = "LD_LIBRARY_PATH=$variables $shell" if ($variables);
    }elsif($myDbType eq 'Pg'){
      $shell = $ptrBin->{'psql'}||'psql';
	  if($myDbPasswd) {
		$shell = "PGPASSWORD=$myDbPasswd $shell";
	  }
	  if($myDbUser) {
		$shell = "PGUSER=$myDbUser $shell";
	  }

      $args = " -d $myDbDB -c ";
    }
#
# check connection to Confixx databnase
#
    if($myDbDB=~/confixx/){
      $sql = "select hostname from admin";
      @result = `$echo "$sql"| $shell $args`;
      unless($result[0]=~/^\w+$/){
		$args = ' -h $myDbHost'.$args;
		@result = `$echo "$sql"| $shell $args`;
		unless($result[0]=~/^\s*hostname/){
		  die "Error: Can not connect to Confixx database (unable to execute sql: $sql)\n@result\n";
		}
      }
    }
#
# end check connection to Confixx databnase
#
    $shell .= $args;
  }
  $this = {
	   'RECONNECT'=>sub{
	     if($dbh){
	       $dbh->disconnect();
		   if($myDbPasswd){
			 $dbh = DBI->connect($myDbAddress,$myDbUser,$myDbPasswd) or
			   die "Error: unable to connect to database: $DBI::errstr\n";
		   }else{
			 $dbh = DBI->connect($myDbAddress,$myDbUser) or
			   die "Error: unable to connect to database: $DBI::errstr\n";
		   }
	       return $dbh;
	     }else{
	       return 1;
	     }
	   },
	   'EXECUTE'=> sub {
	     $sql = shift;
	     if($dbh){
	       $sth = &executeDbh($dbh,$sql);
	       if($sth){
		 return $sth->rows;
	       }else{
		 return 0;
	       }
	     }else{
	       $resultIndex=0;
	       &executeDbShell($shell,\@result,$echo,$sql,$myDbType);
	       return scalar(@result);
	     }

	   },
	   'FETCHROW'=> sub {
	     if($sth){
	       if(@result = $sth->fetchrow){
		 return \@result;
	       }else{
		 return undef;
	       }
	     }else{
	       $resultIndex++;
	       if($resultIndex>$#result){
		 return undef;
	       }else{
		 return $result[$resultIndex];
	       }
	     }
	   },
	   'FETCHHASH'=>sub {
	     if($sth){
	       $ptrResult = $sth->fetchrow_hashref();
	     }else{
	       $resultIndex++;
	       if($resultIndex>$#result){
		 $ptrResult = undef;
	       }else{
			 my ($i,$key);
			 $ptrResult = {};

			 for($i=0; $i<@{$result[0]}; $i++){
			   $ptrResult->{$result[0]->[$i]}=$result[$resultIndex]->[$i];
			 }
	       }
	     }
	     return $ptrResult;
	   },
	   'FINISH'=>sub {
	     if($sth){
	       $sth->finish();
	     }
	     @result=();
	     $ptrResult = undef;
	   },
	   'PRINTERROR'=>sub {
	     if($dbh){
	       if(@_){
			 $dbh->{'PrintError'}=shift;
	       }
	       return $dbh->{'PrintError'};
	     }else{
	       return 0;
	     }
	   },
	   'PING'=>sub{
	     if($dbh){
	       return $dbh->ping();
	     }else{
	       return 1;
	     }
	   },
	   'DISCONNECT'=>sub{
	     if($dbh){
	       return $dbh->disconnect();
	     }else{
	       return 1;
	     }
	   },
	   'USER'=>sub{
	     return $myDbUser;
	   },
	   'PASSWORD'=>sub{
	     return $myDbPasswd;
	   },
	   'HOST'=>sub{
	     return $myDbHost;
	   },
	   'TYPE'=>sub{
	     return $myDbType;
	   },
	  };

  return $this;
}

sub executeDbh {
  my $dbh = shift;
  unless(ref($dbh)=~/DBI/){
    &printToError ("Error: DBI object is not found");
  }
  my $sql = shift;

  my $sth = $dbh->prepare($sql) or
    &printToError ("Error: prepare sql: $sql\n$DBI::errstr");
  if($sth->execute){
    return $sth;
  }else{
    &printToError ("Error: exesute sql: $sql\n$DBI::errstr");
  }
}

sub executeDbShell {
  my ($shell,$ptrRes,$echo,$sql, $type) = @_;
  @{$ptrRes}=();
  my (@out,@row,$ret);
  $ret=0;
  if($type eq 'Pg'){
    @out = `$shell "$sql"`;
  }else{
    @out = `$echo "$sql"|$shell`;
  }
  if ($out[0]=~/^error/i){
    &printToErrror ("Error: execute sql: $sql\n@out");
  }else{
    if($type eq 'Pg'){
      foreach my $line (@out){
	chomp $line;
	next if ($line=~/^-[-+]*$/); # skip delimiter
	if($line=~/^\(\s*\d+\s+\w+\)$/){
	    last;  ## (6 rows)
	}
	$line=~s/^\s+//;
	$line=~s/\s+$//;
	@row = split(/\s*\|\s*/,$line);
	push @{$ptrRes}, [@row];
	$ret++;
      }
    }else{
      if($out[0]=~/^\+[-+]*$/){
	foreach my $line (@out){
	  chomp $line;
	  next if ($line=~/^\+[-+]*$/); # skip delimiter
	  $line=~s/^\|\s*//;
	  $line=~s/\s+$//;
	  @row = split(/\s*\|\s*/,$line);
	  @row = map { $_ eq 'NULL' ? undef : $_ } @row;
	  push @{$ptrRes}, [@row];
	  $ret++;
	}
      }else{  ## batch mode
	foreach my $line (@out){
	  chomp $line;
	  @row = split(/\t/,$line);
	  @row = map { $_ eq 'NULL' ? undef : $_ } @row;
	  push @{$ptrRes}, [@row];
	  $ret++;
	}
      }
    }
  }
  return $ret;
}

#
# end class for handle with database
#

sub makeMIMEBase64 {
  my ($useModule,$this);
  if (eval "require MIME::Base64"){
    $useModule = 1;
  }else{
    $useModule = 0;
  }

  $this={
	 'ENCODE'=>sub{
	   my $src = shift;
	   my ($encoded);
	   if($useModule){
	     $encoded = MIME::Base64::encode($src);
		 chomp $encoded;
	   }else{
	     $encoded = &encodeBase64($src);
	   }
	   return $encoded;
	 },
	 'USEMODULE'=>sub{
	   return $useModule;
	 },
	};

  return $this;
}

#
# url decode (certificates in Plesk DB encoded by the url encoding)
#

sub urlDecode {
    my $url = $_[0];
    $url =~ tr/+/ /;
    $url =~ s/%([a-fA-F0-9]{2,2})/chr(hex($1))/eg;
    return $url;
}

sub encodeHtml {
  my(@ret,$src);
  if ($src=shift){
    if(ref $src){
      if(${$src}=~/[&"'><]/){
		${$src}=~s/&/&amp;/sg;
		${$src}=~s/>/&gt;/sg;
		${$src}=~s/</&lt;/sg;
		${$src}=~s/\'/&#39;/sg;
		${$src}=~s/\"/&quot;/sg;
      }
      return ${$src};
    }else{
      if($src=~/[&"'><]/){
		$src=~s/&/&amp;/sg;
		$src=~s/>/&gt;/sg;
		$src=~s/</&lt;/sg;
		$src=~s/\'/&#39;/sg;
		$src=~s/\"/&quot;/sg;
      }
    }
  }
  return $src;
}

# Normalization according to the XML 1.0 (Third Edition),
# clauses 2.3 "Common Syntactic Constructs"
# and 3.3.3 "Attribute-value normalization"

sub normalizeAttribute {
  $_ = $_[0];
  s/\r/&#xD;/g;
  s/\n/&#xA;/g;
  s/\t/&#x9;/g;
  s/\</&lt;/g;
  s/\&/&amp;/g;
  s/\"/&quot;/g;
  s/\'/&#39;/g;
  return $_;
}

############################
#
#	Non-ascii characters handling
{
	# static
	my $charset;

	sub setCharset {
	  $charset = shift;
	}

	# try to determine the local charmap
	sub _determineCharset()
	{
	  my $charset;
	  my $locale = $bin{'locale'} || 'locale';

	  if (exists $ENV{'LANG'}) {
		my @pieces = split /\./, $ENV{'LANG'};
		if (defined $pieces[1]) {
		  $charset = $pieces[1];
		  chomp $charset;
		}
	  }
	  if (not $charset && open(CM, "$locale charmap |")) {
	  	binmode CM;
		$charset = <CM>;
		close(CM);
		chomp $charset;
	  }
	  if (not $charset) {
		$charset = 'ISO-8859-1';
	  }

	  return $charset;
	}

	sub toUtf8
	{
		my ($subj, $myCharset) = @_;

		$charset = _determineCharset() if (not $charset);

		$myCharset = $charset unless ( $myCharset );

		if ( (not $subj) ||
				 (not $bin{'iconv'}) ||
				 ( $myCharset =~ /UTF-?8/i )
			 ) {
			return $subj;
		}

		if ($subj =~ /^[\x00-\x7f]*$/) { return $subj; }

		my $converted;
		my $escaped = $subj;
		$escaped =~ s/'/'"'"'/g;

		if (open(IC, "echo '$escaped' | " . $bin{'iconv'} . " -f '$myCharset' -t 'UTF-8' |")) {
			binmode IC;
			while (my $string = <IC>) { $converted .= $string; }
			close(IC);
			chomp $converted;
			if ($converted) {
			  return $converted;
			}
		}

		return "";
	}
}

#
# Based on the b64.c by Bob Trower (base64.sf.net)
#

my @cb64 = ('A', 'B', 'C', 'D', 'E', 'F', 'G', 'H',
			'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P',
			'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X',
			'Y', 'Z', 'a', 'b', 'c', 'd', 'e', 'f',
			'g', 'h', 'i', 'j',	'k', 'l', 'm', 'n',
			'o', 'p', 'q', 'r', 's', 't', 'u', 'v',
			'w', 'x', 'y', 'z', '0', '1', '2', '3',
			'4', '5', '6', '7', '8', '9', '+', '/');

@cb64 = map { ord } @cb64;

sub encodeBlock {
  my ($len, @in) = @_;
  my @res;
  push @res, $cb64[$in[0] >> 2];
  push @res, $cb64[(($in[0] & 0x3)<<4)|(($in[1]&0xf0)>>4)];
  push @res, $len>1 ? $cb64[(($in[1]&0xf)<<2)|(($in[2]&0xc0)>>6)] : ord('=');
  push @res, $len>2 ? $cb64[$in[2]&0x3f] : ord('=');
  return @res;
}

sub encodeBase64 {
  my @in = unpack "C*", $_[0];

  my @out;
  while (scalar(@in)) {
	my @block = (0, 0, 0);
	my $length = 0;
	while (scalar(@in) && $length < 3) {
	  $block[$length++] = shift(@in);
	}
	push @out, encodeBlock($length, @block);
  }
  return pack "C*", @out;
}

sub getBins {
  my ($ptrBin,$path) = @_;
  my (%searchPath,$key,$prefix,$bin,$fullPath);
  my ($notFound,$i,@prefixes);

  $notFound=0;

  $i=0;
  foreach $prefix (split(/:/,$path)){
    next unless $prefix;
    unless(exists($searchPath{$prefix})){
      $searchPath{$prefix}=$i++;
      push @prefixes,$prefix;
    }
  }

  foreach $key (keys %{$ptrBin}){
	$ptrBin->{$key} = undef;
    foreach $prefix (@prefixes){
      $prefix.='/' unless $prefix=~/\/$/;
      $bin = $key;
      $bin =~ s/^\///;
      $fullPath = $prefix.$bin;
      if (-x $fullPath){
		$ptrBin->{$key}=$fullPath;
		last;
      }
    }
    $notFound++;
  }
  return $notFound;
}

sub getArguments {
  my ($ptrArgv,$ptrOpts) = @_;

  my (@keys,$firstKey,%prepKeys,$key,$value,$ptrArr,$arg,$state);
  my (%retHash,$pat,$found,$used,@rest,$fullArg,$prevKey);

  while (($key,$value)=each(%{$ptrOpts})){
    @keys = split(/\|/,$key);
    $firstKey = $keys[0];
    $firstKey =~s/^-*//;
    push @{$prepKeys{$firstKey}},[@keys];
    push @{$prepKeys{$firstKey}},$value;
  }

  $state =0;
  foreach $arg (@{$ptrArgv}){

#    print "arg: $arg\n";

    $used = 0;
    if($state==1){
      if  ($arg =~ /^-/){
	$state=0;
      }else{
	$used = 1;
	$retHash{$prevKey}=$arg;
	next;
      }
    }
    if ($state == 2){
      $retHash{$prevKey}=$arg;
      $state = 0;
      $used = 1;
    }else{
      $fullArg = $arg;
      if ($arg =~ /^(-\S+)=(\S+)/){
	$arg = $1;
	$value = $2;
      }else{
	$value = undef;
      }
      foreach $key (keys %prepKeys){

	$ptrArr = $prepKeys{$key};

#	print "key: $key, ",join('|',@{$ptrArr->[0]}),"\n";


	$found = 0;
	foreach $pat (@{$ptrArr->[0]}){
	  if ($pat eq $arg){
	    $found = 1;
	    last;
	  }
	}
	if($found){
	  $used = 1;
	  if(defined($value)){
	    $retHash{$key}=$value;
	  }else{
	    if($ptrArr->[1]){
	      if($ptrArr->[1] =~ /\?$/){
		$state = 1;
	      }else{
		$state = 2;
	      }
	      $prevKey = $key;
	    }else{
	      $retHash{$key}=undef;
	    }
	  }
	  last;
	}
      }
    }
    unless($used){
      if($value){
	push @rest,$fullArg;
      }else{
	push @rest,$arg;
      }
    }
  }
  @{$ptrArgv}=@rest;
  return \%retHash;

}

sub makeDumpStatus($) {
  my $statusFile = shift;
  my($pid,$currAccount,$currDomain,$leftAccounts,$leftDomains,$this);
  $pid=$$;
  $this = {'FILE'=>sub {
	     if(@_){
	       $statusFile = shift;
	     }
	     return $statusFile;
	   },
	   'PID'=>sub {
	     if(@_){
	       $pid = shift;
	     }
	     return $pid;
	   },
	   'ACCOUNT'=>sub{
	     if(@_){
	       $currAccount = shift;
	       if($leftAccounts){
		 $leftAccounts--;
	       }
	     }
	     return $currAccount;
	   },
	   'DOMAIN'=>sub{
	     if(@_){
	       $currDomain = shift;
	       if($leftDomains){
		 $leftDomains--;
	       }
	     }
	     return $currDomain;
	   },
	   'COUNTS'=>sub{
	     if(@_){
	       my $newValue = shift;
	       if (defined $newValue){
		 $leftAccounts=$newValue;
	       }
	       $newValue = shift;
	       if (defined $newValue){
		 $leftDomains=$newValue;
	       }
	     }
	     return ($leftAccounts,$leftDomains);
	   },
	   'PRINT'=>sub{
	     if(@_){
	       $statusFile=shift;
	     }
	     open(STATUS,">$statusFile.tmp");

		 binmode STATUS;
	     my $root = &makeXmlNode('agent-dump-status');
	     my $item = &makeXmlNode('pid',$pid);
	     $root->{'ADDCHILD'}->($item);
	     $item = &makeXmlNode('account',$currAccount||'');
	     $root->{'ADDCHILD'}->($item);
	     $item = &makeXmlNode('domain',$currDomain||'');
	     $root->{'ADDCHILD'}->($item);
	     $item = &makeXmlNode('accounts-left',
				  defined($leftAccounts)?$leftAccounts:'');
	     $root->{'ADDCHILD'}->($item);
	     $item = &makeXmlNode('domains-left',
				  defined($leftDomains)?$leftDomains:'');
	     $root->{'ADDCHILD'}->($item);

	     &printXml($root,*STATUS,'agent-dump-status.dtd');

	     close(STATUS);

		 rename("$statusFile.tmp", $statusFile);
	   }

	  };
  return $this;
}

sub makeContentList {
  my ($dirDump,$xmlDump) = @_;
  my ($name,$fullPath,%files,%fromXML,$size,@cids,$root,$item,$key,$cmd,
      $gzip,$wc,$md5sum);

  if(opendir(DUMP,$dirDump)){
    while ($name=readdir(DUMP)){
      $fullPath="$dirDump/$name";
      unless(-f $fullPath && $name=~/\.\w\w$/){
	next;
      }
      $key = $fullPath;
      $key =~ s/\.\w\w$//;
      $size=(stat($fullPath))[7];
      if(defined $size){
	push @{$files{$key}},[$fullPath,$size];
      }
    }
    closedir(DUMP);
  }

	if (-T $xmlDump) {
		if (open(DUMP, "<$xmlDump")) {
			binmode DUMP;
			while (<DUMP>) {
				if (@cids = /cid(?:_\w+)?=\"([^"]+)\"/g) {
					foreach $fullPath (@cids) {
						unless ($fullPath =~ /^\//) {
							$fullPath = "$dirDump/$fullPath";
						}
						$fromXML{$fullPath} = undef;
					}
				}
			}
			close(DUMP);
		}

		foreach $key (keys %files) {
			unless (exists($fromXML{$key})) {
				delete $files{$key};
			}
		}
	}

  $gzip = $bin{'gzip'}||'gzip';
  $wc = $bin{'wc'}||'wc';
  $md5sum = $bin{'md5sum'}||'md5sum';

  $root=&makeXmlNode('contentlist');

  my($compressed,$uncompressed,$list,$count,$ptrList,@sortedList,
     $ptrItem,$out);
  my($totalUncomp,$totalSize)=(0,0);
  foreach $key (keys %files){

#
# sort by extension  .aa .ab ...
#
    @sortedList =  sort {&getExt($a->[0]) cmp &getExt($b->[0])} @{$files{$key}};

#
# output list of files
#
    $compressed=0;
    foreach $ptrItem (@sortedList){
      ($fullPath,$size)=@{$ptrItem};
      $item=&makeXmlNode('file',$fullPath);
      $item->{'ATTRIBUTE'}->('size',$size,'noencode');
      $root->{'ADDCHILD'}->($item);

#
# get md5sum
#
      if(-x $md5sum){
	$out=`$md5sum $fullPath`;
	if($out=~/\s*([0-9a-fA-F]+)\s+/){
	  $item->{'ATTRIBUTE'}->('md5sum',$1,'noencode');
	}
      }

      $compressed += $size;

    }

    $uncompressed = undef;
#
# get unpacked size of the archive
#
    if ($key=~/\.tgz$|\.gz$/){
      if(-x $gzip){
	if(scalar(@sortedList)==1){
	  ($fullPath,$size)=@{$sortedList[0]};
	  ($uncompressed,$compressed)=&getZipSize($fullPath,$gzip);
	}else{
	  $list = join(' ',map {$_->[0]} @sortedList);
	  if(-x $wc){
	    $cmd = "cat $list | $gzip | $wc -c";
	    $uncompressed = `$cmd`;
	    chomp $uncompressed;
	    if ($uncompressed =~ /^\s*(\d+)\s*$/){
	      $uncompressed = $1;
	    }else{
	      $uncompressed = undef;
	    }
	  }else{
	    $cmd = "cat $list >$key";
	    if(system($cmd)==0){
	      ($uncompressed,$compressed)=&getZipSize($key,$gzip);
	      unlink $key;
	    }
	  }
	}
      }
    }
#
# calc total size
#
    $totalSize += $compressed;
    $totalUncomp += $uncompressed if $uncompressed;
  }

  $root->{'ATTRIBUTE'}->('size',$totalSize,'noencode');
  if ($totalUncomp && $totalUncomp != $totalSize){
    $root->{'ATTRIBUTE'}->('unpacked',$totalUncomp,'noencode');
  }

  return $root;
}

sub getExt {
  my $path = shift;
  if($path=~/\.([^.]*)$/){
    return $1;
  }
  return $path;
}

sub getZipSize {
  my ($fullPath,$gzip) = @_;
  $gzip ||= 'gzip';
  my $cmd = "$gzip -l $fullPath";
  my ($uncompressed,$compressed,@out);
  if (open(OUTPUT,"$cmd|")){
    ($compressed,$uncompressed)=(0,0);
    while(<OUTPUT>){
      @out = split(' ',$_);
      if($out[0]=~/^\d+$/ && $out[1]=~/^\d+$/){
	$uncompressed+=$out[1];
	$compressed+=$out[0];
      }
    }
    close (OUTPUT);
  }
  return $uncompressed,$compressed;
}

my $outputIsFile;

sub openOutput
{
	my ($outPath) = @_;
	my ($outFh);

	if ($outPath) {
		open(OUTPUT, ">>$outPath") or die "Error: can not open '$outPath' for write: $!\n";
		binmode OUTPUT;
		$outFh = *OUTPUT;
		$outputIsFile = 1;
	} else {
		$outFh = *STDOUT;
		$outputIsFile = 0;
	}

	return $outFh;
}

sub closeOutput{
  if($outputIsFile){
    close (OUTPUT) or
      &printToError("Error: unable to close OUTPUT: $!\n");
    $outputIsFile=0;
  }
}

#=========================================================================
#
# dtd-specific (but common) subroutines
#
sub createPasswordNode {
  my ($parent, $password, $type) = @_;

  my $item = makeXmlNode('password', $password);
  $item->{'ATTRIBUTE'}->('type', $type);
  $parent->{'ADDCHILD'}->($item);
}

sub createSysuserNode {
  my ($parent, $name, $password, $type) = @_;

  my $item = makeXmlNode('sysuser');
  $item->{'ATTRIBUTE'}->('name', $name);
  createPasswordNode($item, $password, $type);
  $parent->{'ADDCHILD'}->($item);
}

sub readFile {
  my ($fileName) = @_;

  unless (-f $fileName) {
	return undef;
  }

  open(INPUT, "< $fileName");
  binmode INPUT;
  my $content = join('', <INPUT>);
  close(INPUT);

  return $content;
}

1;

#
# end common subroutines
#
#=========================================================================


$ptrArgs = &getArguments (\@ARGV,\%arg_opts);

#
# search binaries
#

&getBins(\%bin,'/bin:/usr/bin:/usr/local/bin:/sbin:/usr/sbin:'.
		 '/usr/local/sbin:'.$ENV{'PATH'});

my ($outPath,$outFh,$outHas,$dumpFile,$statusFile,$app);

$app = $0;
if ($app=~/(.+)\/([^\/]+)$/) {
  $workDir = $1;
  $app = $2;
}
$app=~s/\.pl$//;

$pwd=`$bin{'pwd'}`||'.';

chop $pwd;

if ($workDir eq '.') {
  $workDir = $pwd;
} elsif ($workDir !~ /^\//) {
  $workDir = $pwd.'/'.$workDir;
}

$dumpFile = $workDir.'/EnsimX.xml';
$statusFile = $workDir.'/EnsimX.status';
$dumpDir = $workDir.'/pma';

my $objDumpStatus = &makeDumpStatus($ptrArgs->{'status-file'}||$statusFile);


if (!$ptrArgs->{'output'} &&
	(exists $ptrArgs->{'dump-all'} ||
	 exists $ptrArgs->{'dump-accounts'}||
	 exists $ptrArgs->{'dump-domains'}
	)
   ) {
  $ptrArgs->{'output'}=$dumpFile;
}

if ($outPath = $ptrArgs->{'output'}) {

	open(OUTPUT, ">$outPath") or die "Error: can not open '$outPath' for write: $!\n";
	binmode OUTPUT;
	$dumpFile = $outPath;
	close(OUTPUT);
}

if (exists $ptrArgs->{'version'}) {
  print $outFh $VERSION,"\n";
  exit 0;
}

my $rootName = 'EnsimXdump';
my $dtdName = 'EnsimX.dtd';

my $wrapDbh = getDbConnect(\%bin, $dbType,
						   "DBI:$dbType:dbname=$dbName",
						   undef, undef,
						   $dbName,$dbHost);

if (exists $ptrArgs->{'get-status'}) {
  my $outFh = &openOutput($ptrArgs->{'output'});

  &printAgentStatus($outPath=$ptrArgs->{'output'}||$dumpFile,
					$objDumpStatus->{'FILE'}->(), $outFh);

  &closeOutput();
  exit 0;
}

unless (ref($wrapDbh)=~/HASH/ && ref($wrapDbh->{'EXECUTE'})=~/CODE/) {
  die "Unable to connect to Ensim DB. Giving up.";
}

#
# get MIME Base64 encoder
#
my $wrapBase64 = makeMIMEBase64();


if (exists ($ptrArgs->{'dump-all'}) ||
	exists ($ptrArgs->{'dump-accounts'}) ||
	exists ($ptrArgs->{'dump-domains'})) {

  my ($root,@accounts,@domains,$ptrAccounts,$ptrDomains, $value);
  if (exists $ptrArgs->{'no-content'}) {
	$::dumpDir = &getDumpDir("$workDir/pma");
  } else {
	$::dumpDir = &getDumpDir("$workDir/pma",'clean');
  }

  if (exists $ptrArgs->{'no-compress'}) {
	setCompress(0);
  }

  &printToLog("Work dir: $workDir");
  &printToLog("Dump file: $dumpFile");
  &printToLog("Status file: ".$objDumpStatus->{'FILE'}->());


  if ($value = $ptrArgs->{'dump-accounts'}) {
	if ($value eq "-") {
	  $value = <STDIN>;
	  chomp $value;
	}
	@accounts = split(/\s*,\s*/,$value);
	$ptrAccounts = \@accounts;
  }

  if ($value = $ptrArgs->{'dump-domains'}) {
	if ($value eq "-") {
	  $value = <STDIN>;
	  chomp $value;
	}
	@domains = split(/\s*,\s*/,$value);
	$ptrDomains = \@domains;
  }
  #
  # generate a xml dump
  #

  $root = &getEnsimXDump($dumpDir, $ptrAccounts, $ptrDomains, $rootName);

  #
  # print dump to output
  #
  &printToLog('Print dump file ...',1);
  if (ref($root)=~/HASH/ && ref($root->{'PRINT'})=~/CODE/) {
	my $outFh = &openOutput($ptrArgs->{'output'});

	&printXml($root,$outFh,$dtdName);

	&closeOutput();
  }
  &printToLog(' OK');

} elsif (exists $ptrArgs->{'get-content-list'}) {

  $::dumpDir = &getDumpDir("$workDir/pma");

  my $root = &makeContentList($::dumpDir,$dumpFile);

  if (ref($root)=~/HASH/) {
	my $outFh = &openOutput($ptrArgs->{'output'});

	&printXml($root,$outFh);

	&closeOutput();
  }

} else {
  &printHelp();
}

exit $returnCode;

#
# end main
#
#==============================================================
#
# subroutines
#

sub printAgentStatus {
  my $dumpFile = shift;
  my $statusFile = shift;
  my $outFh = shift || *STDOUT;
  my $rootName = shift || 'agentstatus';

  my ($root,$util,$out,$item,$engineRoot,$rpm,$sql,$ptrRow);

  $root = &makeXmlNode($rootName);

  $rpm = $bin{'rpm'} || 'rpm';
  $ensimVersion = `$rpm -q webppliance-version`;

  unless ($ensimVersion =~ /^webppliance-version-([^-]+)-[^-]+$/ && defined $bin{'iconv'}) {
	if (defined $bin{'iconv'}) {
	  $item = &makeXmlNode('wrong-platform', '');
	} else {
	  $item = &makeXmlNode('wrong-platform', 'no iconv found on the source host');
	}

	$root->{'ADDCHILD'}->($item);
	&printXml($root, $outFh, $rootName . '.dtd');
	return;
  }

  $item = &makeXmlNode('dumpfile', $dumpFile);
  $root->{'ADDCHILD'}->($item);
  $item = &makeXmlNode('statusfile', $statusFile);
  $root->{'ADDCHILD'}->($item);

  &printXml($root, $outFh, $rootName . '.dtd');
}

sub getEnsimXDump {
  my ($dumpDir, $ptrAccounts, $ptrDomains, $rootName) = @_;
  $rootName = 'PleskXDump' unless $rootName;

  my $root = &makeXmlNode($rootName);

  if (ref($ptrDomains) =~ /ARRAY/ && (@{$ptrDomains} > 0)) {

	my $list = join(',', map {"'$_'"} @{$ptrDomains});
	my $sql = "SELECT site_id FROM siteinfo WHERE domain in ($list)";

	$wrapDbh->{'EXECUTE'}->($sql);

	my $ptrRow;
	my %domainIds;
	while ($ptrRow = $wrapDbh->{'FETCHROW'}->()) {
	  $domainIds{(@{$ptrRow})[0]} = 1;
	}

	$wrapDbh->{'FINISH'}->();

	createSelectedDomains($root, \%domainIds);

  } elsif (ref($ptrAccounts) =~ /ARRAY/ && (@{$ptrAccounts} > 0)) {
	my %accountIds;

	my @accounts = @{$ptrAccounts};

	if (grep {$_ eq $defaultDomainsAccount} @accounts) {
	  $accountIds{0} = 1;
	}

	@accounts = grep {$_ ne $defaultDomainsAccount} @accounts;

	my $list = join(',', map {"'$_'"} @accounts);
	my $sql = "SELECT reseller_id FROM reseller_info WHERE username in ($list)";

	$wrapDbh->{'EXECUTE'}->($sql);

	my $ptrRow;
	while ($ptrRow = $wrapDbh->{'FETCHROW'}->()) {
	  $accountIds{(@{$ptrRow})[0]} = 1;
	}

	$wrapDbh->{'FINISH'}->();

	createSelectedAccounts($root, \%accountIds);

  } else {
	createAllAccounts($root);
  }

  return $root;
}

sub createAllAccounts {
  my ($parent) = @_;

  my $sql = "SELECT DISTINCT reseller_info.reseller_id, username " .
	"FROM reseller_info LEFT JOIN reseller " .
	  "ON reseller_info.reseller_id = reseller.reseller_id ".
		"WHERE (reseller_info.reseller_id <> 0) OR (site_id IS NOT NULL)";

  $wrapDbh->{'EXECUTE'}->($sql)
	or die "Cannot retrieve resellers list from DB";

  my %accounts;
  my $ptrRow;
  while ($ptrRow = $wrapDbh->{'FETCHROW'}->()) {
	my @accountInfo = @$ptrRow;
	$accounts{$accountInfo[0]} = 1;
  }

  createSelectedAccounts($parent, \%accounts);
}

sub createSelectedAccounts {
  my ($parent, $accountsPtr) = @_;
  my %accounts = %{$accountsPtr};

  my $account;
  foreach $account (keys %accounts) {
	&createAccountNode($parent, $account);
  }
}

sub createSelectedDomains {
  my ($parent, $domainsPtr) = @_;
  my %domains = %{$domainsPtr};

  my $account = makeXmlNode('reseller');
  createPasswordNode($account, '', 'empty');
  createPinfoNode($account, '', undef, undef);

  my $domain;
  foreach $domain (keys %domains) {
	&createDomain($account, $domain);
  }

  $parent->{'ADDCHILD'}->($account);
}

sub createAccountNode {
  my ($parent, $accountId) = @_;

  &printToLog("Creating account $accountId");

  my $sql = "SELECT reseller_id, username, email, password, fullname, enabled,".
	"total_threshold, enabled_bandwidth, total_quota, enabled_diskquota,".
	  "total_ipbased, total_namebased, enabled_ipinfo_nb, enabled_ipinfo_ip,".
		"total_maxusers, enabled_users ".
		  "FROM reseller_info ".
			"NATURAL LEFT JOIN reseller_bandwidth ".
			  "NATURAL LEFT JOIN reseller_diskquota ".
				"NATURAL LEFT JOIN reseller_users ".
				  "NATURAL LEFT JOIN reseller_ipinfo ".
					"WHERE (reseller_id = '$accountId')";

  my $ptrHash;
  ($wrapDbh->{'EXECUTE'}->($sql) && ($ptrHash = $wrapDbh->{'FETCHHASH'}->()))
	or die "Cannot retrieve info for account $accountId";
  my %info = %$ptrHash;

  $wrapDbh->{'FINISH'}->();

  my $item = &makeXmlNode('reseller');

  &createPasswordNode($item, $info{'password'}, 'encrypted');

  my $username = $info{'username'};
  my $fullname = $info{'fullname'};
  if ($username eq "") {
	$username = $defaultDomainsAccount;
	$fullname = $defaultDomainsFullname;
  }

  &createPinfoNode($item, $username, $fullname, $info{'email'});

  unless (exists $ptrArgs->{'no-content'}) {
	&createLimitNode($item, 'reseller', 'traffic-quota',
					 $info{'total_threshold'},
					 $info{'enabled_bandwidth'} == 0 || $info{'total_threshold'} == 0);

	&createLimitNode($item, 'reseller', 'disk-quota',
					 $info{'total_quota'},
					 $info{'enabled_diskquota'} == 0 || $info{'total_quota'} == 0);

	&createLimitNode($item, 'reseller', 'max-users',
					 $info{'total_maxusers'}, $info{'reseller_id'} eq "0");
	&createLimitNode($item, 'reseller', 'ip-based-domains',
					 $info{'total_ipbased'}, $info{'reseller_id'} eq "0");
	&createLimitNode($item, 'reseller', 'name-based-domains',
					 $info{'total_namebased'}, $info{'reseller_id'} eq "0");

	&createPermissionNode($item, 'reseller', 'create-users',
						  $info{'enabled_users'});
	&createPermissionNode($item, 'reseller', 'create-ipbased',
						  $info{'enabled_ipinfo_ip'});
	&createPermissionNode($item, 'reseller', 'create-namebased',
						  $info{'enabled_ipinfo_nb'});

	if ($info{'enabled'} == 0) {
	  my $disabledNode = makeXmlNode('disabled');
	  $item->{'ADDCHILD'}->($disabledNode);
	}
  }

  createAllDomainsForAccount($item, $accountId);

  $parent->{'ADDCHILD'}->($item);
}

sub createAllDomainsForAccount {
  my ($parent, $accountId) = @_;

  my $sql = "SELECT site_id FROM reseller WHERE reseller_id = '$accountId'";

  $wrapDbh->{'EXECUTE'}->($sql);

  my $ptrRow;
  while ($ptrRow = $wrapDbh->{'FETCHROW'}->()) {
	my $domainId = (@$ptrRow)[0];

	createDomain($parent, $domainId);
  }
  $wrapDbh->{'FINISH'}->();
}

sub createDomain {
  my ($parent, $domainId) = @_;

  &printToLog("Creating account $domainId");

  my $item = &makeXmlNode('domain');
  $parent->{'ADDCHILD'}->($item);

  #
  # MAIN CONFIGURATION FILES
  #
  my %domainInfo       = %{readConfigFile($domainId, 'siteinfo')};
  my %ipinfo           = %{readConfigFile($domainId, 'ipinfo')};

  $item->{'ATTRIBUTE'}->('name', $domainInfo{'domain'});

  #
  # PARSE USERS INFO
  #
  my $passwdInfo = parsePasswdShadow($domainId);

  my $domadmNode = &makeXmlNode('domain-admin');
  createPasswordNode($domadmNode,
					 $passwdInfo->{$domainInfo{'admin_user'}}->{'password'},
					 'encrypted');

  createPinfoNode($domadmNode, $domainInfo{'admin_user'},
				  undef, $domainInfo{'email'});
  $item->{'ADDCHILD'}->($domadmNode);

  #
  # HOSTING TYPE & IP
  #
  my $htNode = &makeXmlNode('hosting-type');
  my $domainIpNode;

  if ($ipinfo{'namebased'} eq '1') {
	$htNode->{'ATTRIBUTE'}->('type', 'name-based');
	$ipinfo{'nbaddrs'} =~ /\[\'(.+)\'\]/;
	$domainIpNode = &makeXmlNode('domain-ip', $1);
  } else {
	$htNode->{'ATTRIBUTE'}->('type', 'ip-based');
	$ipinfo{'ipaddrs'} =~ /\[\'(.+)\'\]/;
	$domainIpNode = &makeXmlNode('domain-ip', $1);
  }
  $item->{'ADDCHILD'}->($htNode);
  $item->{'ADDCHILD'}->($domainIpNode);

  # NO MORE INFORMATION REQUIRED IN '--no-content' MODE
  if (exists $ptrArgs->{'no-content'}) {
	return;
  }

  #
  # ADDITIONAL CONFIGURATION FILES
  #
  my %aliasesInfo      = %{readConfigFile($domainId, 'aliases')};
  my %diskquotaInfo    = %{readConfigFile($domainId, 'diskquota')};
  my %bandwidthInfo    = %{readConfigFile($domainId, 'bandwidth')};
  my %analogInfo       = %{readConfigFile($domainId, 'analog')};
  my %webalizerInfo    = %{readConfigFile($domainId, 'webalizer')};
  my %usersInfo        = %{readConfigFile($domainId, 'users')};
  my %subdomainInfo    = %{readConfigFile($domainId, 'subdomain')};
  my %mivamerchantInfo = %{readConfigFile($domainId, 'mivamerchant')};
  my %filesInfo        = %{readConfigFile($domainId, 'files')};
  my %vhbackupInfo     = %{readConfigFile($domainId, 'vhbackup')};
  my %develenvInfo     = %{readConfigFile($domainId, 'develenv')};
  my %sshInfo          = %{readConfigFile($domainId, 'ssh')};
  my %telnetInfo       = %{readConfigFile($domainId, 'telnet')};
  my %sslInfo          = %{readConfigFile($domainId, 'openssl')};

  #
  # CONTENT
  #
  my $domainName = $domainInfo{'domain'};

  my $docroot = getFstPath($domainId) . "/home/" .
	$domainInfo{'admin_user'} . "/mainwebsite_html";
  my $cgi = getFstPath($domainId) . "/home/" .
	$domainInfo{'admin_user'} . "/mainwebsite_cgi";

  if (-d $docroot) {
	my $filename = makeDumpFile("$dumpDir/$domainName\_docroot",
								$docroot);
	$item->{'ATTRIBUTE'}->('cid_docroot', $filename);
  }

  if (-d $cgi) {
	my $filename = makeDumpFile("$dumpDir/$domainName\_cgi",
								$cgi);
	$item->{'ATTRIBUTE'}->('cid_cgi', $filename);
  }


  #
  # ALIASES
  #
  if ($aliasesInfo{'enabled'} == '1') {
	$aliasesInfo{'aliases'} =~ /\[(.+)\]/;
	my @aliases = map { /^\'(.*)\'$/; $1 } split /\s*,\s*/, $1;

	my $alias;
	foreach $alias (@aliases) {
	  printToLog($domainName);
	  printToLog($alias);
	  if ($alias =~ /^.*\Q$domainName\E$/) {
		my $aliasNode = &makeXmlNode('domain-alias', $alias);
		$item->{'ADDCHILD'}->($aliasNode);
	  }
	}
  }

  #
  # MONITORING
  #
  my $monitoringNode = &makeXmlNode('monitoring');
  $monitoringNode->{'ATTRIBUTE'}->('cycle-start', $bandwidthInfo{'rollover'});

  createPermissionNode($monitoringNode, 'monitoring',
					   'analog', $analogInfo{'enabled'});

  createPermissionNode($monitoringNode, 'monitoring',
					   'webalizer', $webalizerInfo{'enabled'});

  $item->{'ADDCHILD'}->($monitoringNode);

  #
  # DOMAIN LIMITS
  #

  createLimitNode($item, 'domain', 'disk-quota',
				  $diskquotaInfo{'quota'}*$megabyte,
				  $diskquotaInfo{'enabled'} == 0);

  createLimitNode($item, 'domain', 'traffic-quota',
				  $bandwidthInfo{'threshold'},
				  $bandwidthInfo{'enabled'} == 0);

  createLimitNode($item, 'domain', 'users',
				  $usersInfo{'maxusers'}, !defined($usersInfo{'maxusers'}));

  createLimitNode($item, 'domain', 'max-subdomain',
				  $subdomainInfo{'max'}, !defined($subdomainInfo{'max'}));

  #
  # DOMAIN PERMISSIONS
  #
  createPermissionNode($item, 'domain',
					   'create-subdomains', $subdomainInfo{'enabled'});

  createPermissionNode($item, 'domain',
					   'e-commerce', $mivamerchantInfo{'enabled'});

  createPermissionNode($item, 'domain',
					   'file-manager', $filesInfo{'enabled'});

  createPermissionNode($item, 'domain',
					   'backup-utils', $vhbackupInfo{'enabled'});

  createPermissionNode($item, 'domain',
					   'devtools', $develenvInfo{'enabled'});

  createPermissionNode($item, 'domain',
					   'ssh', $sshInfo{'enabled'});

  createPermissionNode($item, 'domain',
					   'telnet', $telnetInfo{'enabled'});

  createPermissionNode($item, 'domain',
					   'ssl', $sslInfo{'enabled'});

  #
  # DOMAIN USERS
  #
  my %mailusers = %{createDomainUsersNode($domainId, $item, $passwdInfo)};

  #
  # WEB
  #
  createWebNode($domainId, $item);

  #
  # SUBDOMAINS
  #
  createSubdomains($domainId, $domainInfo{'domain'}, $item);

  #
  # MAILLISTS
  #
  my $aliasesToIgnore = createMaillists($domainId, $domainInfo{'domain'}, $item);

  #
  # EMAIL
  #
  createMailNode($domainId, $domainInfo{'domain'}, $item, \%mailusers, $aliasesToIgnore);

  #
  # LOGROTATION
  #
  createLogrotationNode($domainId, $item, $domainInfo{'email'});

  #
  # FTP
  #
  createFtpNode($domainId, $item);

  #
  # Protected Directories
  #
  createPdirNodes($domainId, $item);

  #
  # DB
  #
  createDbNode($domainId, $item);

  #
  # CERTIFICATE
  #
  createSslNode($domainId, $item, \%sslInfo);
}


sub getUserQuota {
  my ($uid) = @_;

  my ($bc, $bs, $bh, $bt, $ic, $is, $ih, $it) = Quota::query($quotaDev, $uid, 0);

  if (defined $bs) {
	return $bs * 1024;
  }

  return undef;
}

sub parsePasswdShadow {
  my ($domainId) = @_;

  my $passwdReader = makeFileParser(getFstPath($domainId) . "/etc/passwd");
  my $shadowReader = makeFileParser(getFstPath($domainId) . "/etc/shadow");

  my (%passwd, %shadow);

  if (defined($passwdReader)) {
	%passwd = %{$passwdReader->{'PARSE'}->('KEYSEPARATOR' => ':',
										   'VALUESEPARATOR' => ':')};
  } else {
	return undef;
  }

  if (defined($shadowReader)) {
	%shadow = %{$shadowReader->{'PARSE'}->('KEYSEPARATOR' => ':',
										   'VALUESEPARATOR' => ':')};
  }

  my %userParams;

  my ($username, $paramsPtr);
  while (($username, $paramsPtr) = each %passwd) {
	my @params = @{$paramsPtr};
	my $fullname = $params[3]; #see the 'passwd' format reference
	my $id = $params[1]; #see the 'passwd' format reference
	my $password;

	if (%shadow && defined($shadow{$username})) {
	  my @shadowParams = @{$shadow{$username}};
	  $password = $shadowParams[0]; #see the 'shadow' format reference
	}

	$userParams{$username} = { 'fullname' => $fullname, 'password' => $password, 'id' => $id };
  }

  return \%userParams;
}

sub createDomainUsersNode
{
	my ($domainId, $parent, $usersInfo) = @_;
	my %domainInfo  = %{readConfigFile($domainId, 'siteinfo')};
	my @pamlistlist = ('imap', 'proftpd', 'smtp_relay', 'ssh', 'telnet');

	my %pamlists;
	my $pamlist;

	foreach $pamlist (@pamlistlist) {
		$pamlists{$pamlist} = {};

		my $pamlistfile;
		open($pamlistfile, getFstPath($domainId) . "/etc/" . $pamlist . ".pamlist");
		if (defined($pamlistfile)) {
			binmode $pamlistfile;
			while (my $line = <$pamlistfile>) {
				chomp $line;
				$pamlists{$pamlist}{$line} = 1;
			}
			close($pamlistfile);
		}
	}


  my %cgiPermissions = %{retrieveCgiPermissions($domainId)};

  my %domusers;

  my $username;
  foreach $username (keys %{$usersInfo}) {
	if (defined($ignoredUsers{$username})) {# || $username eq $domainInfo{'admin_user'}) {
	  next;
	}

	my $domainName = $domainInfo{'domain'};
	my $password = $usersInfo->{$username}->{'password'};

	if ($username ne $domainInfo{'admin_user'}) {

	  my $fullname = $usersInfo->{$username}->{'fullname'};
	  my $domuser = makeXmlNode('domain-user');
	  my $contentRoot = getFstPath($domainId) . "/home/" . $username;

	  if (-d $contentRoot) {
		my $fileName = makeDumpFile("$dumpDir/$domainName\_$username",
									$contentRoot);
		if ($fileName) {
		  $domuser->{'ATTRIBUTE'}->('cid', $fileName);
		}
	  }


	  createPinfoNode($domuser, $username, $fullname, undef);
	  if (defined($password)) {
		createPasswordNode($domuser, $password, 'encrypted');
	  }

	  #LIMITS
	  my $userQuota = getUserQuota($usersInfo->{$username}->{'id'});

	  createLimitNode($domuser, 'domain-user', 'disk-quota',
					  $userQuota,	!defined($userQuota) || $userQuota eq "0");

	  #PERMISSIONS
	  createPermissionNode($domuser, 'domain-user', 'ftp',
						   defined($pamlists{'proftpd'}{$username}));

	  createPermissionNode($domuser, 'domain-user', 'ssh',
						   defined($pamlists{'ssh'}{$username}));

	  createPermissionNode($domuser, 'domain-user', 'telnet',
						   defined($pamlists{'telnet'}{$username}));

	  createPermissionNode($domuser, 'domain-user', 'cgi',
						   defined($cgiPermissions{$username}));

	  $parent->{'ADDCHILD'}->($domuser);
	}

	my $mailbox = getFstPath($domainId) . "/var/spool/mail/" . $username;

	my $fileName;
	if (-f $mailbox) {
	  printToLog("Packing $username\@$domainName mail");
	  $fileName = makeDumpFile("$dumpDir/$username\@$domainName\_mbox",
							   getFstPath($domainId) . "/var/spool/mail",
							   $username);
	}

	$domusers{$username} = { 'password' => $password, 'cid' => $fileName };
  }

  return \%domusers;
}

sub retrieveCgiPermissions {
  my ($domainId, $username) = @_;

  my %cgiAllowed = ();
  my (%db, $user);

  tie %db, "DB_File", "/etc/virtualhosting/cgi/site" . $domainId . ".cgi_owner_map",
	O_RDONLY, 0666, $DB_HASH;

  foreach $user (values %db) {
	$cgiAllowed{$user} = "1";
  }

  untie %db;

  return \%cgiAllowed;
}

sub createWebNode {
  my ($domainId, $parent) = @_;

  #
  # CONFIGURATION FILES
  #
  my %apacheInfo       = %{readConfigFile($domainId, 'apache')};
  my %cgiInfo          = %{readConfigFile($domainId, 'cgi')};
  my %mod_perlInfo     = %{readConfigFile($domainId, 'mod_perl')};
  my %ssiInfo          = %{readConfigFile($domainId, 'ssi')};
  my %weblogsInfo      = %{readConfigFile($domainId, 'weblogs')};
  my %tomcat4Info      = %{readConfigFile($domainId, 'tomcat4')};
  my %frontpageInfo    = %{readConfigFile($domainId, 'frontpage')};


  my $item = &makeXmlNode('web');
  $item->{'ATTRIBUTE'}->('web-server-name', $apacheInfo{'webserver'});
  $item->{'ATTRIBUTE'}->('cgi-content-root', $cgiInfo{'scriptalias'});
  $item->{'ATTRIBUTE'}->('perl-content-root', $mod_perlInfo{'alias'});

  #
  # PERMISSIONS
  #
  createPermissionNode($item, 'web', 'web', $apacheInfo{'enabled'});
  createPermissionNode($item, 'web', 'cgi', $cgiInfo{'enabled'});
  createPermissionNode($item, 'web', 'ssi', $ssiInfo{'enabled'});
  createPermissionNode($item, 'web', 'mod_perl', $mod_perlInfo{'enabled'});
  createPermissionNode($item, 'web', 'weblogs', $weblogsInfo{'enabled'});
  createPermissionNode($item, 'web', 'tomcat', $tomcat4Info{'enabled'});
  createPermissionNode($item, 'web', 'frontpage', $frontpageInfo{'enabled'});

  #
  # WEBUSERS
  #
  createWebUsersNode($domainId, $item);

  $parent->{'ADDCHILD'}->($item);
}

sub createWebUsersNode {
  my ($domainId, $parent) = @_;

}

sub createSubdomains {
  my ($domainId, $domain, $parent) = @_;

  my ($subdomainsfile, %subdomains);
  open($subdomainsfile, "/etc/virtualhosting/subdomain/site" . $domainId . ".map");
  binmode $subdomainsfile;
  while (<$subdomainsfile>) {
	unless(/^www/) {
	  my @info = split;
	  my @domparts = split /\./, $info[0];
	  $subdomains{$domparts[0]} = $info[1];
	}
  }

  my ($subdomainscgifile, %subdomainscgi);
  open($subdomainscgifile, "/etc/virtualhosting/subdomain/site" . $domainId . ".cgisubmap");
  binmode $subdomainscgifile;
  while (<$subdomainscgifile>) {
	unless(/^www/) {
	  my @info = split ":";
	  $subdomainscgi{$info[0]} = 1;
	}
  }

  my ($subdomain, $location);
  while (($subdomain, $location) = each %subdomains) {

	my $node = makeXmlNode('subdomain');
	$node->{'ATTRIBUTE'}->('name', $subdomain);

	if (-d $location . "/html") {
	  my $htmlcontent = makeDumpFile("$dumpDir/$subdomain.${domain}_htdocs",
									 $location . "/html", ('.'), (), '-h');

	  if (defined $htmlcontent) {
		$node->{'ATTRIBUTE'}->('cid_docroot', $htmlcontent);
	  }
	}

	if (-d $location . "/cgi-bin") {
	  my $cgicontent = makeDumpFile("$dumpDir/$subdomain.${domain}_cgibin",
									$location . "/cgi-bin", ('.'), (), '-h');

	  if (defined $cgicontent) {
		$node->{'ATTRIBUTE'}->('cid_cgi', $cgicontent);
	  }
	}

	if (defined $subdomainscgi{$subdomain}) {
	  my $sdpn = makeXmlNode('subdomain-permission');
	  $sdpn->{'ATTRIBUTE'}->('name', 'cgi');
	  $node->{'ADDCHILD'}->($sdpn);
	}

	$parent->{'ADDCHILD'}->($node);
  }
}

sub createMaillists {
  my ($domainId, $domain, $parent) = @_;

  my %aliasesToIgnore;

  open(LISTLIST, "find " . getFstPath($domainId) . "/var/lib/majordomo/lists -type f -name '*.config' -mindepth 1 -maxdepth 1 |");
  binmode LISTLIST;
  while(<LISTLIST>) {
	chomp;
	my $configFileName = $_;

	/([^\/]+)\.config/;
	my $listName = $1;

	/^(.+)\.config$/;
	my $usersFileName = $1;

	my $admin_passwd = "";

	open(LISTCONFIG, $configFileName);
	binmode LISTCONFIG;
	while (<LISTCONFIG>) {
	  if (/admin_passwd\s=\s(.*)$/) {
		$admin_passwd = $1;
		last;
	  }
	}
	close LISTCONFIG;

	my $item = makeXmlNode('maillist');
	$item->{'ATTRIBUTE'}->('name', $listName);
	$item->{'ATTRIBUTE'}->('state', 'on');

	my $listOwner = 'owner-not-found';
	my $aliasGatherer = sub {
	  my ($alias, $redirects, $localDelivery, $responders) = @_;

	  if ($alias eq $listName . '-owner') {
		if (@{$localDelivery}) {
		  $listOwner = $localDelivery->[0] . '@' . $domain;
		} elsif (@{$redirects}) {
		  $listOwner = $redirects->[0];
		}
		return 0;
	  }

	  return 1;
	};
	traverseMailAliases($domainId, $aliasGatherer);

	$item->{'ADDCHILD'}->(makeXmlNode('owner', $listOwner));
	createPasswordNode($item, $admin_passwd, 'plain');

	open(LISTMEMBERS, $usersFileName);
	binmode LISTMEMEBERS;
	while (<LISTMEMBERS>) {
	  chomp;
	  $item->{'ADDCHILD'}->(makeXmlNode('recipient', $_));
	}

	$parent->{'ADDCHILD'}->($item);

	$aliasesToIgnore{'owner-' . $listName} = 1;
	$aliasesToIgnore{$listName . '-owner'} = 1;
	$aliasesToIgnore{$listName . '-request'} = 1;
	$aliasesToIgnore{$listName . '-approval'} = 1;
  }
  close LISTLIST;

  return \%aliasesToIgnore;
}

sub createMailNode {
  my ($domainId, $domain, $parent, $domusersRef, $aliasesToIgnore) = @_;
  my %domusers = %{$domusersRef};


  #
  # CONFIGURATION FILES
  #
  my %sqmailInfo       = %{readConfigFile($domainId, 'sqmail')};
  my %sendmailInfo     = %{readConfigFile($domainId, 'sendmail')};
  my %majordomoInfo    = %{readConfigFile($domainId, 'majordomo')};
  my %vacationInfo     = %{readConfigFile($domainId, 'vacation')};

  unless ($sendmailInfo{'enabled'}) {
	return;
  }

  my $item = &makeXmlNode('email');
  $item->{'ATTRIBUTE'}->('mail-server-name', $sendmailInfo{'mailserver'});

  #
  # PERMISSIONS
  #
  createPermissionNode($item, 'email', 'majordomo',
					   $majordomoInfo{'enabled'});
  createPermissionNode($item, 'email', 'autoresponder',
					   $vacationInfo{'enabled'});
  createPermissionNode($item, 'email', 'webmail',
					   $sqmailInfo{'enabled'});

  #
  # MAILALIASES & DOMAIN USERS
  #

  my %singleAliases;
  my %complexAliases;

  my $aliasGatherer = sub {
	my ($alias, $redirects, $localDelivery, $responders) = @_;

	if (defined $domusers{$alias}) {
	  return 1;
	}

	if (defined $aliasesToIgnore->{$alias}) {
	  return 1;
	}

	if (@{$redirects} == 0 && @{$localDelivery} == 0 && @{$responders} == 0) {
	  return 1;
	}

	if (@{$redirects} == 0 && @{$localDelivery} == 1 && @{$responders} == 0) {
	  unless (defined $singleAliases{$localDelivery->[0]}) {
		$singleAliases{$localDelivery->[0]} = [];
	  }
	  push @{$singleAliases{$localDelivery->[0]}}, $alias;
	} else {
	  $complexAliases{$alias} = { 'redirects' => $redirects,
								  'localDelivery' => $localDelivery,
								  'responders' => $responders };
	}
	return 1;
  };
  traverseMailAliases($domainId, $aliasGatherer);

  my ($domuser, $p);
  while (($domuser, $p) = each(%domusers)) {
	createDomuserMailNode($item, $domain, $domuser, $p, $singleAliases{$domuser});
  }


  my ($alias, $complexAlias);
  while (($alias, $complexAlias) = each %complexAliases) {
	createComplexAliasMailNode($item, $domain, $alias, $complexAlias);
  }

  #
  # SPAMFILTERS
  #
  createSpamFiltersNode($domainId, $item);

  $parent->{'ADDCHILD'}->($item);
}

sub createDomuserMailNode {
  my ($parent, $domain, $domuser, $paramsPtr,  $aliasesPtr) = @_;
  my %params = %{$paramsPtr};
  my $password = $params{'password'};

  my $mailuserNode = makeXmlNode('mailuser');
  $mailuserNode->{'ATTRIBUTE'}->('name', $domuser);
  $mailuserNode->{'ATTRIBUTE'}->('mailgroup-enabled', 'false');
  if (defined($password)) {
	createPasswordNode($mailuserNode, $password, 'encrypted');
  }

  my $mailboxNode = makeXmlNode('mailbox');

  if (defined $params{'cid'}) {
	$mailboxNode->{'ATTRIBUTE'}->('cid', $params{'cid'});
	$mailboxNode->{'ATTRIBUTE'}->('type', 'mbox');
  }

  $mailuserNode->{'ADDCHILD'}->($mailboxNode);

  if (defined($aliasesPtr)) {
	my $alias;
	foreach $alias (@{$aliasesPtr}) {
	  my $aliasNode = makeXmlNode('alias', $alias);
	  $mailuserNode->{'ADDCHILD'}->($aliasNode);
	}
  }

  $parent->{'ADDCHILD'}->($mailuserNode);
}

sub createComplexAliasMailNode {
  my ($parent, $domain, $alias, $complexAliasPtr) = @_;
  my %complexAlias = %{$complexAliasPtr};

  my $item = makeXmlNode('mailuser');
  $item->{'ATTRIBUTE'}->('name', $alias);

  if (@{$complexAlias{'localDelivery'}} || @{$complexAlias{'redirects'}}) {
	$item->{'ATTRIBUTE'}->('mailgroup-enabled', 'true');
  } else {
	$item->{'ATTRIBUTE'}->('mailgroup-enabled', 'false');
  }
  my $redirect;
  foreach $redirect (@{$complexAlias{'redirects'}}) {
	my $redirectNode = makeXmlNode('mailgroup-member', $redirect);
	$item->{'ADDCHILD'}->($redirectNode);
  }

  my $localDelivery;
  foreach $localDelivery (@{$complexAlias{'localDelivery'}}) {
	my $ldNode = makeXmlNode('mailgroup-member', $localDelivery . "@" . $domain);
	$item->{'ADDCHILD'}->($ldNode);
  }

  my ($nresp, $responder);
  $nresp = 1;
  foreach $responder (@{$complexAlias{'responders'}}) {
	my $respondernode = makeXmlNode('autoresponder_ext');
	$respondernode->{'ATTRIBUTE'}->('name', 'responder' . $nresp++);
	$respondernode->{'ATTRIBUTE'}->('subject', $responder->{'subject'});
	$respondernode->{'ADDCHILD'}->(makeXmlNode('text', $responder->{'body'}));
	$item->{'ADDCHILD'}->($respondernode);
  }

  $parent->{'ADDCHILD'}->($item);
}

sub traverseMailAliases {
  my ($domainId, $executor) = @_;

  my $configReader = makeFileParser(getFstPath($domainId). "/etc/aliases");
  my %aliases = %{$configReader->{'PARSE'}->('KEYSEPARATOR' => ':',
											 'COMMENT' => '#',
											 'VALUESEPARATOR' => ',')};

  my ($alias, $ptrActions);
  while (($alias, $ptrActions) = each %aliases) {
	my @actions = @{$ptrActions};
	my (@redirects, @localDelivery, @responders);

	@actions = map { chomp; s/^\s*(.*)$/$1/; $_; } @actions;
	foreach (@actions) {
	  if (/^"\|responder\.sh\s+\S+\s+(\S+)"$/) {
		open INPUT, getFstPath($domainId) . $1;
		binmode INPUT;
		my $subject = <INPUT>;
		my $body;
		while (<INPUT>) {
		  $body .= $_;
		}
		close(INPUT);

		$subject = $wrapBase64->{'ENCODE'}->($subject);
		$body = $wrapBase64->{'ENCODE'}->($body);

		chomp $subject;
		chomp $body;

		push @responders, { 'subject' => $subject, 'body' => $body };
	  } elsif (/^((\w|\.|-)+)$/) {
		push @localDelivery, $1;
	  } elsif (/^((\w|@|\.|-)+)$/) {
		push @redirects, $1;
	  } elsif (/^"\|\/usr\/lib\/majordomo\/wrapper$"/) {
		&printToLog("Majordomo action ignored: " . $_);
	  } elsif (/^:include:/) {
		&printToLog("Majordomo action ignored: " . $_);
	  } else {
		&printToLog("Unknown action detected: " . $_);
	  }
	}
	if ($executor->($alias, \@redirects, \@localDelivery, \@responders) == 0) {
	  last;
	}
  }
}

sub createSpamFiltersNode {
  my ($domainId, $parent) = @_;

  open(INPUT, getFstPath($domainId) . "/etc/mail/access");
  binmode INPUT;

  my %aliases;
  while (<INPUT>) {
	if (/^(.*)\s+REJECT$/) {
	  my $item = makeXmlNode('spam-filter', $1);
	  $parent->{'ADDCHILD'}->($item);
	}
  }
}

sub createLogrotationNode {
  my ($domainId, $parent, $adminemail) = @_;

  open(LOGROT, getFstPath($domainId) . "/etc/logrotate.conf") or return;
  binmode LOGROT;

  my $mult = 1;
  my $maxnumber;
  my $maxsize;
  my $period = 'monthly';

  while (<LOGROT>) {
	chomp;
	if (/^rotate\s+(\d+)*$/) {
	  $maxnumber = $1;
	} elsif (/^size\s+(\d+)(.?)$/) {
	  if ($2 eq '') {
		$mult = 1;
	  } elsif ($2 eq 'M') {
		$mult = $megabyte;
	  } elsif ($2 eq 'k') {
		$mult = $kilobyte;
	  }
	  $maxsize = $1;
	} elsif (/^(daily|weekly|monthly)$/) {
	  $period = $1;
	}
  }

  if (not(defined($maxnumber)) or not(defined($maxsize))) {
	printToLog("Log rotation configuration is not complete.");
	return;
  }

  my $logrotationnode = makeXmlNode('logrotation');
  $logrotationnode->{'ATTRIBUTE'}->('enabled', 'true');
  $logrotationnode->{'ATTRIBUTE'}->('max-number-of-logfiles', $maxnumber);
  $logrotationnode->{'ATTRIBUTE'}->('compress', 'true');
  $logrotationnode->{'ATTRIBUTE'}->('email', $adminemail) if defined($adminemail);

  if($maxsize != 0) {
	my $logrotationsizenode = makeXmlNode('logrotation-maxsize', ($maxsize * $mult));
	$logrotationnode->{'ADDCHILD'}->($logrotationsizenode);
  } else {
	my $logrotationperiodnode = makeXmlNode('logrotation-period');
	$logrotationperiodnode->{'ATTRIBUTE'}->('period', $period);
	$logrotationnode->{'ADDCHILD'}->($logrotationperiodnode);
  }
  $parent->{'ADDCHILD'}->($logrotationnode);
}

sub createFtpNode {
  my ($domainId, $parent) = @_;

  #
  # CONFIGURATION FILES
  #
  my %anonftpInfo      = %{readConfigFile($domainId, 'anonftp')};
  my %proftpdInfo      = %{readConfigFile($domainId, 'proftpd')};
  my %domainInfo       = %{readConfigFile($domainId, 'siteinfo')};

  my $domainName = $domainInfo{'domain'};

  unless ($proftpdInfo{'enabled'}) {
	return;
  }

  my $item = &makeXmlNode('ftp');
  $item->{'ATTRIBUTE'}->('ftp-server-name', $proftpdInfo{'ftpserver'});

  createPermissionNode($item, 'ftp', 'anon-ftp', $anonftpInfo{'enabled'});

  my $anonftpPub = getFstPath($domainId) . "/var/ftp/pub";
  if (-d $anonftpPub) {
	my $fileName = makeDumpFile("$dumpDir/$domainName\_\_\_anonftp",
								$anonftpPub);
	if ($fileName) {
	  $item->{'ATTRIBUTE'}->('cid', $fileName);
	}
  }

  my $anonftpIncoming = getFstPath($domainId) . "/var/ftp/incoming";
  if (-d $anonftpIncoming) {
	my $fileName = makeDumpFile("$dumpDir/$domainName\_\_\_anonftp\_incoming",
								$anonftpIncoming);
	if ($fileName) {
	  $item->{'ATTRIBUTE'}->('cid_incoming', $fileName);
	}
  }

  $parent->{'ADDCHILD'}->($item);
}

sub createPdirNodes {
  my ($domainId, $parent) = @_;

  my %htusers;
  my %htgroups;

  my $htpasswdReader =
	makeSafeFileParser(getFstPath($domainId) . "/var/www/.htpasswd");

  if (defined($htpasswdReader)) {
	%htusers = %{$htpasswdReader->{'PARSE'}->('KEYSEPARATOR'  => ':')};
  }

  my $htgroupReader = makeSafeFileParser(getFstPath($domainId) .
										 "/var/www/.htgroup");
  if (defined($htgroupReader)) {
	%htgroups = %{$htgroupReader->{'PARSE'}->('KEYSEPARATOR' => ':')};

	my $group;
	foreach $group (keys %htgroups) {
	  my @members = split(/\s+/, $htgroups{$group});
	  $htgroups{$group} = \@members;
	}
  }

  my $accesses;
  open($accesses, "find '" . getFstPath($domainId) . "/var/www/html' -name .htaccess -type f |");
  binmode $accesses;

  while (<$accesses>) {
	chomp $_;

	my $reldirname = getRelDir($domainId, $_);
	my $pdirnode = makeXmlNode('pdir');
	$pdirnode->{'ATTRIBUTE'}->('name', $reldirname);

	my %pdirusers;

	open ACCESS, $_;
	binmode ACCESS;
	while (<ACCESS>) {
	  chomp $_;
	  if (/require\s+group\s+(.*)$/) {
		my @groups = split /\s+/, $1;
		my $group;
		foreach $group (@groups) {
		  if (defined($htgroups{$group})) {
			my $htuser;
			foreach $htuser (@{$htgroups{$group}}) {
			  if (defined($htusers{$htuser})) {
				$pdirusers{$htuser} = $htusers{$htuser};
			  } else {
				$pdirusers{$htuser} = '';
			  }
			}
		  }
		}
	  } elsif (/AuthName \"(.*)\"$/) {
		$pdirnode->{'ATTRIBUTE'}->('title', $1);
	  }
	}
	close ACCESS;

	if (%pdirusers) {
	  my ($user, $password);
	  while (($user, $password) = each %pdirusers) {
		my $htusernode = makeXmlNode('pduser');
		$htusernode->{'ATTRIBUTE'}->('name', $user);
		if ($password eq '') {
		  createPasswordNode($htusernode, '', 'empty');
		} else {
		  createPasswordNode($htusernode, $password, 'encrypted');
		}
		$pdirnode->{'ADDCHILD'}->($htusernode);
	  }
	  $parent->{'ADDCHILD'}->($pdirnode);
	}
  }
  close $accesses;
}

sub getRelDir {
  my ($domainId, $name) = @_;
  my $prefix = getFstPath($domainId) . "/var/www/html";
  if ($name =~ /^\Q$prefix\E(.*)\.htaccess$/) {
	my $name = $1;
	if ($name =~ /^(.*)\/$/ and $name ne '/') {
	  return $1;
	}
	return $name;
  }

  die "Cannot remove '$prefix' from the '$name'";
}

sub createDbNode {
  my ($domainId, $parent) = @_;
  my ($ptrRow, $ptrHash);

  #
  # CONFIGURATION FILES
  #
  my %mysqlInfo = %{readConfigFile($domainId, 'mysql')};

  my $dbPassword = getMysqlPassword();

  my $wrapUserMysql = getDbConnect(\%bin,'mysql', 'DBI:mysql:mysql:localhost',
								   'root', $dbPassword, 'mysql', 'localhost');

  my $sql = "SHOW DATABASES LIKE '" . $mysqlInfo{'dbaseprefix'} . "%'";

  $wrapUserMysql->{'EXECUTE'}->($sql);

  my @userDatabases;
  while ($ptrRow = $wrapUserMysql->{'FETCHROW'}->()) {
	push @userDatabases, (@{$ptrRow})[0];
  }
  $wrapUserMysql->{'FINISH'}->();

  my ($dbName);
  foreach $dbName (@userDatabases) {

	my $dbNode = makeXmlNode('database');
	$dbNode->{'ATTRIBUTE'}->('name', $dbName);
	$dbNode->{'ATTRIBUTE'}->('type', 'mysql');

	#skip bases with '.'
	if ($dbName =~ /\./) {
	  next;
	}

	my $db = getDbConnect(\%bin, 'mysql', "DBI:mysql:$dbName:localhost",
						  'root', $dbPassword, $dbName, 'localhost');

	if (defined($db)) {
	  my $dbDump = makeDumpDb($dbName, 'mysql', 'root',
							  $dbPassword, 'localhost');
	  $dbNode->{'ATTRIBUTE'}->('cid', $dbDump);

	  addDbUsers($dbNode, $dbName, $wrapUserMysql);

	  $parent->{'ADDCHILD'}->($dbNode);
	}
  }
}

sub createSslNode {
  my ($domainId, $parent, $sslInfoPtr) = @_;

  if ($sslInfoPtr->{'enabled'} eq '0') {
	return;
  }

  my $sslNode = makeXmlNode('certificate');
  my $content;
  my $confDir = getFstPath($domainId) . "/etc/httpd/conf";

  $content = readFile($confDir . "/ssl.csr/server.csr");
  if ($content) {
	$sslNode->{'ADDCHILD'}->(makeXmlNode('signing-request', $content));
  }

  $content = readFile($confDir . "/ssl.crt/server.crt");
  if ($content) {
	$sslNode->{'ADDCHILD'}->(makeXmlNode('certificate-data', $content));
  }

  $content = readFile($confDir . "/ssl.key/server.key");
  if ($content) {
	$sslNode->{'ADDCHILD'}->(makeXmlNode('private-key', $content));
  }

  $parent->{'ADDCHILD'}->($sslNode);
}

sub readConfigFile {
  my ($domainId, $fileName) = @_;

  my $configReader = makeSafeFileParser("/home/virtual/site" . $domainId .
										"/info/current/" . $fileName);

  if (defined $configReader) {
	return $configReader->{'PARSE'}->('KEYSEPARATOR'  => '=',
									  'COMMENT' => '[DEFAULT]');
  } else {
	return { "enabled" => 0 };
  }
}

sub createPasswordNode {
  my ($parent, $password, $type) = @_;

  my $item = &makeXmlNode('password', $password);
  $item->{'ATTRIBUTE'}->('type', $type);
  $parent->{'ADDCHILD'}->($item);
}

sub createPinfoNode {
  my ($parent, $username, $fullname, $email) = @_;

  my $item = &makeXmlNode('pinfo');
  $item->{'ATTRIBUTE'}->('username', $username);
  $item->{'ATTRIBUTE'}->('fullname', $fullname) if defined($fullname);
  $item->{'ATTRIBUTE'}->('e-mail', $email) if defined($email);
  $parent->{'ADDCHILD'}->($item);
}

sub createLimitNode {
  my ($parent, $prefix, $name, $value, $unlimited) = @_;

  if ($unlimited) {
	return;
  }

  my $item = &makeXmlNode($prefix . "-limit", $value);
  $item->{'ATTRIBUTE'}->('name', $name);
  $parent->{'ADDCHILD'}->($item);
}

sub createPermissionNode {
  my ($parent, $prefix, $name, $create) = @_;

  if ($create) {
	my $item = &makeXmlNode($prefix . "-permission");
	$item->{'ATTRIBUTE'}->('name', $name);
	$parent->{'ADDCHILD'}->($item);
  }
}

sub getFstPath {
  my ($domainId) = @_;
  return "/home/virtual/site" . $domainId . "/fst";
}

sub getMysqlPassword {

  my $python = $bin{'python'} || 'python';
  my $script =<< "EOS"
import os
import sys
import rotor
mysqlrotkey = ('\%c\%c\%c\%c\%c\%c\%c' % (chr(3), chr(12), chr(234), chr(2), '3', 'e', chr(222)))
def read_mysqlpass():
    if os.path.exists('/etc/appliance/.vpf'):
        f = open('/etc/appliance/.vpf', 'r')
        file = f.read()
        f.close()
        if (len(file) >= 1):
            rotpass = file
            rt = rotor.newrotor(mysqlrotkey, 11)
            rt.decrypt(rotpass)
            return rt.decryptmore(rotpass)
    return ''
sys.stdout.write(str(read_mysqlpass()))
EOS
;
	unless (open(PY, "$python -c \"$script\" |")) {
		return '';
	}
	binmode PY;

	my $passwd = <PY>;
	close(PY);
	chomp $passwd;

	return $passwd;
}

sub printHelp {

  print <<"HELP";

Agent of migration from 'Ensim X' to 'Plesk 7'

Usage:
$0 <options>

Options:
-s |--get-status           get status of the agent
-dc|--dump-accounts=<list> a coma separated list of resellers to dump
-dd|--dump-domains=<list>  a coma separated list of customers to dump
-da|--dump-all             make a full dump

-lc|--get-content-list     get list of content files
-nc|--no-content           do not make content files
-nz|--no-compress          do not compress content files

-o |--output=<path>        redirect output to file

-h |--help                 this help
-d |--debug=<level>        level of debug messages
-v |--version

HELP
}
