#!/usr/bin/perl

use vars qw( %bin $dumpDir $ptrArgs $pwd $workDir);
use strict;

use lib "/usr/sausalito/perl";

my $VERSION='1.0.0';


#
# parse command line
#
my %arg_opts = ('--help|-h'=>'',
				'--debug|-d'=>'i?',
				'--get-status|-s'=>'',
				'--dump-accounts|-dc'=>'s',
				'--dump-domains|-dd'=>'s',
				'--dump-all|-da'=>'',
				'--get-content-list|-lc'=>'',
				'--no-content|-nc'=>'',
				'--no-compress|-nz'=>'',
				'--output|-o'=>'s',
				'--status-file|-sf'=>'s',
				'--raw|-r'=>'',
				'--version|-v'=>'',
	       );

%bin = ('tar'=>'tar','split'=>'split','cp'=>'cp','rm'=>'rm','echo'=>'echo',
		'mysql'=>'mysql','mysqldump'=>'mysqldump',
		'psql'=>'psql','uptime'=>'uptime','uname'=>'uname',
		'cat'=>'cat','pwd'=>'pwd','md5sum'=>'md5sum',
		'grep'=>'grep','find'=>'find','gzip'=>'gzip','wc'=>'wc',
		'locale'=>'locale', 'iconv'=>'iconv');

my $megabyte = 1024*1024;

my %domain_permissions = ('shell' => 'Shell', 'cgi' => 'CGI', 'ssi' => 'SSI',
						  'frontpage' => 'FP', 'apop' => 'APOP', 'php' => 'PHP',
						  'jsp' => 'Java', 'ssl' => 'SSL');

my %user_permissions = ('shell' => 'Shell',
						'frontpage' => 'FP',
						'apop' => 'APOP' );

my $returnCode = 0;


#==============================================================
#
# common subroutines
#

sub printToLog {
  my ($msg, $noeol) = @_;
  if ($msg) {
	print $msg;
  }
  unless ($noeol) {
	print "\n";
  }
}

sub printToError {
  my ($msg, $noeol, $errCode) = @_;
  if ($msg) {
	print STDERR $msg;
  }
  unless ($noeol) {
	print STDERR "\n";
  }
  if (defined($errCode)) {
    $returnCode = $errCode; ## translate an error code to the return code
  } else {
    $returnCode = 127;
  }
}

#BEGINNOTINRAQ2

sub makeFileReader
{
	my ($filename, $doNotFail) = @_;
	local (*INPUT);

	unless (-T $filename) {

		if (defined($doNotFail)) {
			&printToLog("Error: file '$filename' is not a text file");
		} else {
			&printToError("Error: file '$filename' is not a text file");
		}

		return undef;
	}

	unless (open(INPUT,"<$filename")) {

		binmode INPUT;
		if (defined($doNotFail)) {
			&printToLog("Error: unable to open file '$filename' to read");
		} else {
			&printToError("Error: unable to open file '$filename' to read");
		}

		return undef;
	}

	my $input = *INPUT;

	my $this = {
		'READ' => sub {
				my $a = <$input>;
				return $a;
			},

		'CLOSE' => sub { close($input); }

	};

	return $this;
}

sub makeArrayReader {
  my (@arr) = @_;
  my $size= 0;
  my $count = scalar(@arr);

  my $this={'READ'=>sub {
			  if($count == $size) {
				return undef;
			  }

			  return $arr[$count++];
			},
			'CLOSE'=>sub {
			}
		   };
  return $this;
}

sub makeParser {
  my $reader;
  my ($this,$ptrRows,$keySep,$valSep,$valKeySep,$comment,
      $typeRows,$maxIndex,$rowIndex);
  {
	$reader = shift;
	unless(defined($reader)) {
	  return undef;
	}

    my %args=(@_);
    $keySep = $args{'KEYSEPARATOR'};
    $valSep = $args{'VALUESEPARATOR'};
    $valKeySep = $args{'KEYVALUESEPARATOR'};
  }
  $this={'READER'=>sub {
		   if(@_){
			 $reader=shift;
		   }
		   return $reader;
		 },
		 'FILE'=>sub {
		   if(@_){
			 $reader = makeFileReader(@_);
		   }
		   return undef;
		 },
		 'KEYSEPARATOR'=>sub{
		   if(@_){
			 $keySep=shift;
		   }
		   return $keySep;
		 },
		 'VALUESEPARATOR'=>sub{
		   if(@_){
			 $valSep=shift;
		   }
		   return $valSep;
		 },
		 'KEYVALUESEPARATOR'=>sub{
		   if(@_){
			 $valKeySep=shift;
		   }
		   return $valKeySep;
		 },
		 'COMMENT'=>sub{
		   if(@_){
			 $comment=shift;
		   }
		   return $comment;
		 },
		 'PARSE'=>sub{
		   my %args=(@_);

		   $keySep=$args{'KEYSEPARATOR'} if exists $args{'KEYSEPARATOR'};
		   $valSep=$args{'VALUESEPARATOR'} if exists $args{'VALUESEPARATOR'};
		   $valKeySep=$args{'KEYVALUESEPARATOR'} if exists $args{'KEYVALUESEPARATOR'};
		   $comment=$args{'COMMENT'} if exists $args{'COMMENT'};

		   $ptrRows = &parseFile('reader'=>$reader,'keysep'=>$keySep,
								 'valsep'=>$valSep,'keyvalsep'=>$valKeySep,
								 'comment'=>$comment);
		   $typeRows = ref($ptrRows);
		   if($typeRows=~/ARRAY/){
			 $rowIndex=-1;
			 $maxIndex=scalar(@{$ptrRows})-1;
		   }
		   return $ptrRows;
		 },
		 'ROW'=>sub{
		   if($typeRows=~/HASH/){
			 my $key=shift;
			 return $ptrRows->{$key};
		   }elsif($typeRows=~/ARRAY/){
			 if($rowIndex<$maxIndex){
			   $rowIndex++;
			   return $ptrRows->[$rowIndex];
			 }else{
			   return undef;
			 }
		   }else{
			 return undef;
		   }
		 },
		};

  return $this;
}

sub makeFileParser {
  my $filename = shift;
  return makeParser(makeFileReader($filename), @_);
}

sub makeSafeFileParser {
  my $filename = shift;
  return makeParser(makeFileReader($filename, 'doNotFail'), @_);
}

sub makeStringParser {
  my $content = shift;
  return makeParser(makeArrayReader($content), @_);
}

sub parseFile {
  my %arg = (@_);
  my $reader = $arg{'reader'};
  my $keySep = $arg{'keysep'};
  my $valSep = $arg{'valsep'};
  my $valKeySep = $arg{'keyvalsep'};
  my $comment = $arg{'comment'};

  my ($ret,$key,$value,$reKey,$reValue,@values,$keyBlank,$valBlank,
     $vkey,$vvalue,$reKeyValue,$reComment);

  if($keySep){
    $ret = {};
    if(ref($keySep)=~/regexp/i || ($keySep eq ' ')){
      $reKey = $keySep;
    }else{
      $reKey = qr/$keySep/;
    }
  }else{
    $ret = [];
  }
  if($valSep){
    if(ref($valSep)=~/regexp/i || ($valSep eq ' ')){
      $reValue = $valSep;
    }else{
      $reValue = qr/$valSep/;
    }
  }
  if($valKeySep){
    if(ref($valKeySep)=~/regexp/i || ($valKeySep eq ' ')){
      $reKeyValue = $valKeySep;
    }else{
      $reKeyValue = qr/$valKeySep/;
    }
  }
  if($comment){
    if(ref($comment)=~/regexp/i){
      $reComment = $comment;
    }else{
      $reComment = qr/$comment/;
    }
  }

  while($_ = $reader->{'READ'}->()){
    chomp;
    next unless $_;
    if($comment && /$reComment/){  ## commentar
	  next;
    }
    if ($keySep){
      ($key,$value)=split($reKey,$_,2);
      if($key){
		$key=~s/^\s+//;
		$key=~s/\s+$//;
		if($valSep){

		  if ($valKeySep){
			$ret->{$key}={};
			foreach $value (split($reValue,$value)){
			  ($vkey,$vvalue)=split($reKeyValue,$value);
			  $vkey=~s/^\s+//;
			  $vkey=~s/\s+$//;
			  $ret->{$key}->{$value}=$vvalue;
			}
		  }else{
			push @{$ret->{$key}},split($reValue,$value);
		  }
		}else{
		  $value=~s/^\s+//;
		  $value=~s/\s+$//;
		  $ret->{$key}=$value;
		}
      }
    }else{
      if($valSep){
		push @{$ret},[split($valSep,$_)];
      }else{
		push @{$ret},$_;
      }
    }
  }

  $reader->{'CLOSE'}->();

  return $ret;
}

#ENDNOTINRAQ2

sub makePasswordNode {
  my $value= shift;
  my $type = shift||'encrypted';

  my $item = &makeXmlNode('password');
  $item->{'CONTENT'}->($value);  ## html encode
  $item->{'ATTRIBUTE'}->('type',$type);
  return $item;
}

sub mangleName {
  my $name = shift;

  if($name =~ /^\d/) {
    $name = 'user' . $name;
  }
  $name =~ s/\./_/g;

  return $name;
}

sub makeSysUserNode {
  my ($name, $password, $type) = @_;

  unless(defined($name) && defined($password) && defined($type)) {
	printToError("makeSysUserNode: Username, password or password type is not set.");
  }

  my $sysUserNode = makeXmlNode('sysuser');
  $sysUserNode->{'ATTRIBUTE'}->('name', mangleName($name));
  $sysUserNode->{'ADDCHILD'}->(makePasswordNode($password, $type));

  return $sysUserNode;
}

# Returns home directory for the system account $user
sub getHomeDir {
	my ($user) = @_;
	my $home;
	my @pw = getpwnam($user);
	if (-1 != $#pw) {
		$home = $pw[7];
	}
	unless ($home) {
		# how it is by default
		$home = '/home/$account';
	}
	return $home;
}

# Returns password for system account $user
sub getSystemPassword {
	my ($user) = @_;
	my $password;
	my @pw = getpwnam($user);
	if (-1 != $#pw) {
		$password = $pw[1];
	}
	unless ($password) {
		$password = '';
	}
	return $password;
}

sub createUserNameFromDomain {
  my ($name) = @_;

  if ($name =~ /^\d/) {
	$name = 'u' . $name;
  }

  $name =~ s/\./_/g;
  $name =~ s/^www\.//g;
  $name =~ s/[aoeui]//g if length($name) > 16;
  $name = substr($name, 0, 16) if length($name) > 16;

  return $name;
}

my $compress = 1;

sub setCompress {
  ($compress) = @_;
}

sub makeDumpDb
{
  my ($dbName, $dbType, $dbUser, $dbPasswd, $dbHost, $ptrDomainBin, $variables, $domainSocket) = @_;
  $dbType ||= 'mysql';

  my ($dumpFile,$dumpSql,$dumpName,$cmd,$gzip,$split,$mysql,$psql, $retCode,$out);

  $dumpName = "$dbName.$dbType.sql";
  $gzip = $compress == 1 ? ($bin{'gzip'} || 'gzip') : undef;

  if (-x $gzip) {
	$dumpName .= ".gz";
  }
  $dumpFile="$dumpDir/$dumpName";

  printToLog(">>Checking database presence");

  if ($dbType eq 'mysql') {
	$mysql = $bin{'mysql'} || 'mysql';
	if (-x $mysql) {
	  $cmd  = "$mysql -u '$dbUser'";
	  $cmd .= " -h $dbHost" if $dbHost;
	  $cmd .= " -p'$dbPasswd'" if $dbPasswd;
	  $cmd .= " -e '' '$dbName' 2>&1";

	  $out = `$cmd`;
	  chomp $out;
	  if ($out) {
		$retCode = $?>>8;
		printToLog("warning: mysql datatbase '$dbName' not found. mysql has returned '$retCode'");
		return "";
	  }
	}
  } elsif ($dbType eq 'Pg') {
	$psql = $bin{'psql'} || 'psql';
	if (-x $psql) {
	  $cmd  = "export PGUSER='$dbUser'";
	  $cmd .= " PGPASSWORD='$dbPasswd'" if $dbPasswd;
	  $cmd .= "; $psql '$dbName' ";
	  $cmd .= " -h $dbHost" if $dbHost;
	  $cmd .= " -c '' 2>&1";

	  $out = `$cmd`;
	  chomp $out;
	  if ($out) {
		$retCode = $?>>8;
		printToLog("warning: PostgreSQL datatbase '$dbName' not found. psql has returned '$retCode'");
		return "";
	  }
	}
  }

  printToLog("..done");

  unless (exists($ptrArgs->{'no-content'})) {
	$gzip = $bin{'gzip'} || 'gzip';
	$split = $bin{'split'} || 'split';

	if ($dbType eq 'mysql') {
	  my ($prefx, $postfix) = ("", "");
	  my $commonprefix = "LD_LIBRARY_PATH=$variables " if $variables;

	  my $commonopts;
	  $commonopts = " --socket='$domainSocket'" if $domainSocket;
	  $commonopts .= " -h $dbHost" if $dbHost;
	  $commonopts .= " -p'$dbPasswd'" if $dbPasswd;
	  $commonopts .= " -u '$dbUser' '$dbName'";

	  my $dumpcmd;
	  if ($ptrDomainBin) {
		$dumpcmd = $ptrDomainBin->{'mysqldump'};
	  } else {
		$dumpcmd = $bin{'mysqldump'} || 'mysqldump';
	  }

	  my $dumpopts = " --quick --quote-names --add-drop-table";

	  my $mysqlversion = `$commonprefix $bin{'mysql'} $commonopts --version`;
	  if ($mysqlversion =~ /Distrib (\d+)\.(\d+)\./) {
		if ($1 > 4 or ($1 == 4 and $2 > 0)) {
		  $dumpopts .= " --compatible=mysql323";
		}
	  }

	  $cmd = $commonprefix . $dumpcmd . $commonopts . $dumpopts;

	} elsif ($dbType eq 'Pg') {
	  $cmd  = "export PGUSER='$dbUser'";
	  $cmd .= " PGPASSWORD='$dbPasswd'" if($dbPasswd);
	  $cmd .= ' ; ' . ($bin{'pg_dump'} || 'pg_dump') . ' -d -O';
	  $cmd .= " -h '$dbHost'" if $dbHost;
	  $cmd .= " '$dbName'";
	}

	if ($compress and $gzip and -x $gzip) {
	  $cmd .= " | $gzip";
	}
	if (-x $split) {
	  $cmd .= " | $split -b 2040m - $dumpFile.";
	} else {
	  $cmd .= " > $dumpFile";
	}

	local $SIG{'PIPE'} = 'DEFAULT' if $SIG{'PIPE'} eq 'IGNORE';

	unless (system($cmd) == 0) {
	  printToError("Error execute '$cmd'");
	  return undef;
	}
  }
  return $dumpName;
}

sub addDbUsers {
  my ($dbNode, $dbName, $wrapUserMysql) = @_;

  printToLog("Dupming $dbName database users");

  $dbName =~ s/_/\\_/g;

  my (@users, $dbUser, $ptrRow);
  my $sql = "SELECT DISTINCT User FROM db WHERE Db = '$dbName'";
  $wrapUserMysql->{'EXECUTE'}->($sql);
  while ($ptrRow = $wrapUserMysql->{'FETCHROW'}->()) {
		push @users, (@{$ptrRow})[0];
  }
  $wrapUserMysql->{'FINISH'}->();

  foreach $dbUser (@users) {
		addDbUser($dbNode, $dbUser, $wrapUserMysql);
  }
}

sub addDbUser {
  my ($dbNode, $user, $wrapUserMysql, $passwordNode) = @_;

  &printToLog("Dumping mysql user '$user'");

	my ( $item );

  if ( ref( $wrapUserMysql ) =~ /HASH/ &&
			 ref( $wrapUserMysql->{'EXECUTE'} ) =~ /CODE/ ) {

    my ($sql,$ptrRow,@accesshosts,$accesshost);

    $sql = "SELECT password,host FROM user WHERE user='$user'";

    if ( $wrapUserMysql->{'EXECUTE'}->( $sql ) ) {

			$item = makeXmlNode( 'dbuser' );
			$item->{'ATTRIBUTE'}->('name', $user);

      while ( $ptrRow = $wrapUserMysql->{'FETCHROW'}->() ) {
				unless ( ref($passwordNode) =~ /HASH/ ) {
					$passwordNode = makePasswordNode( $ptrRow->[0] );
				}
				push @accesshosts, $ptrRow->[1];
      }

			if ( $passwordNode ) {
				$item->{'ADDCHILD'}->( $passwordNode );
			}

			foreach $accesshost ( @accesshosts ) {
				$item->{'ADDCHILD'}->( makeXmlNode( 'accesshost', $accesshost ) );
			}

			$dbNode->{'ADDCHILD'}->( $item );

    }

    $wrapUserMysql->{'FINISH'}->();
  }

	return $item;
}

my $lastDumpId = 0;

sub generateDumpId {
  return $$ . "-" . $lastDumpId++;
}

#
# sub makeDumpFile - make a dump of dir tree (use 'tar')
#
# Arguments:
#          $destFile - name of destination file.
#                      If the file has not '.tgz' or '.tar.gz' extension
#                      '.tgz' will be added
#          $scrDir   - path to source dir
#
#          $ptrNames - pointer to an array with names of files and dirs to
#                      include into dump
#                      can be a single name
#                      by default '.' - include all entries
#
#          $ptrExcludes - pointer to an array with names of files and dirs to
#                      exclude from dump
#                      can be a single name
#
#          $extraParams - extra params to run tar
#                         ('-h' - don't dump symlinks; dump the files they point to)
#
sub makeDumpFile {
  my ($destFile, $srcDir, $ptrNames,$ptrExclude,$extraParams)=@_;

  my @destPath = split(/\//, $destFile);
  if (length($destPath[$#destPath]) > 32) {
	$destPath[$#destPath] = generateDumpId();
  }
  $destFile = join('/', @destPath);

  my $tar = $bin{'tar'}||'tar';
  my $split = $bin{'split'}||'split';
  my $gzip = $compress == 1 ? ($bin{'gzip'} || 'gzip') : undef;

  my ($cmd,$ret,@names,@exclude);

  if(ref($ptrNames)=~/ARRAY/){
    @names = @{$ptrNames};
  }else{
    @names=($ptrNames) if $ptrNames;
  }

  if(ref($ptrExclude)=~/ARRAY/){
    @exclude = @{$ptrExclude};
  }else{
    @exclude=($ptrExclude) if $ptrExclude;
  }

  unless(-x $tar ){
    &printToError("Error: '$tar' is not executable");
    return undef;
  }

#
# check extension
#
  if(-x $gzip){
    unless($destFile=~/\.tgz$/ ||$destFile=~/\.tar\.gz$/){
      $destFile .= '.tar.gz';
    }
  }else{
    unless($destFile=~/\.tar$/){
      $destFile .= '.tar';
    }
  }
#
# end check extension
#

  unless( exists $ptrArgs->{'no-content'}){

    $cmd=$tar;
    if($srcDir){
      if( -d $srcDir){
		$cmd.=" -C $srcDir -f - -c --ignore-failed-read ";
      }else{
		return undef;
      }
    }


    if($extraParams){
      $cmd .= $extraParams.' ';
    }

    if(@names){
      if($srcDir){
		$cmd .= join( ' ', map {s/^\Q$srcDir\E\///;$_} @names); ## extract domain name
      }else{
		$cmd .= join( ' ', @names );
      }
    }else{
      $cmd .= '.';  ## dump all files & dirs
    }
    if(@exclude){
      $cmd .= ' '.join( ' ', map {s/^\Q$srcDir\E\///;'--exclude '.$_} @exclude); ## extract domain name
    }
    if(-x $gzip){
      $cmd .= " | $gzip";
    }
    if(-x $split){
      $cmd .= " | $split -b 2040m - $destFile.";
    }else{
      $cmd .= " >$destFile";
    }

		local $SIG{'PIPE'} = 'DEFAULT' if $SIG{'PIPE'} eq 'IGNORE';

    unless(system($cmd) == 0){
      &printToError("Error execute command: '$cmd': $!");
      $destFile='';
    }
  }

  $destFile=~s/^\Q$dumpDir\E\///;  ### remove dir name
  return $destFile;
}

sub getDumpDir {
  my ($subdir,$clean) = @_;
  my ($dumpdir);

  if($subdir=~/^\//){
    $dumpdir=$subdir;
  }else{
    my $util = $bin{'pwd'}||'pwd';
    $dumpdir = $pwd||`$util`;
    chomp $dumpdir;
    if($subdir){
      $dumpdir.='/'.$subdir;
    }
  }
  if(-d $dumpdir && $clean){
    unless(system("rm -rf $dumpdir")==0){
      &printToError("Error: can not clean temporary dir '$dumpdir': $!");
    }
  }
  unless(-d $dumpdir){
    mkdir ($dumpdir,0755) or
      die "Error: can not create temporary dir '$dumpdir': $!\n";
  }
  return $dumpdir;
}

sub printXml {
  my ($root,$outFh,$dtd) = @_;
  unless($outFh){
    $outFh = *STDOUT;
  }

  unless(ref($root)=~/HASH/ && ref($root->{'PRINT'})=~/CODE/){
    print $outFh "Error: a xml root is not found\n";
    &printToError("Error: sub printXml: bad arguments");
    return 0;
  }
  print $outFh "<?xml version='1.0' encoding=\"UTF-8\"?>\n";

  my  $doctype = $root->{'NAME'}->();
  if($dtd){
    print $outFh "<!DOCTYPE $doctype SYSTEM '$dtd'>\n";
  }
  $root->{'PRINT'}->($outFh,'  ');
  return 1;
}

#
# class for handle with XML
#
sub makeXmlNode {
  my ($this,$parent,%attributes,@content);
  my ($name,$plain)=@_;

  $plain = toUtf8(encodeHtml($plain));

  unless($name||$plain){
    return undef;
  }
  $this = {
	   'ATTRIBUTE' => sub {
	     my $key = shift;
	     if(@_){
	       my $value=shift;
	       if(@_){
					 $attributes{$key} = $value;
	       }else{
					 $attributes{$key} = toUtf8(normalizeAttribute($value));
	       }
	     }
	     return $attributes{$key};
	   },
	   'PARENT' => sub {
	     if(@_){
	       my $arg = shift;
	       if(ref($arg)=~/HASH/){
					 $parent=shift;
	       }
	     }
	     return $parent;
	   },
	   'NAME' => sub {
	     if(@_){
	       $name = shift;
	     }
	     return $name;
	   },
	   'CONTENT' => sub {
	     if(@_){
	       $plain = toUtf8(encodeHtml(shift));
	     }
	     return $plain;
	   },
	   'ADDCHILD'=> sub {
	     my $kind = shift;
	     if(
					(ref($kind)=~/HASH/) &&
					(ref($kind->{'PARENT'})=~/CODE/)
	       ) {
	       push @content,$kind;
	       $kind->{'PARENT'}->($this);
	     }
	     return $#content;
	   },
	   'KIND'=> sub {
	     my $index = shift;
			 my ($ret);
			 if ( $index =~ /^\d+$/ ) {
				 if($index >= 0 && $index <= $#content ) {
					 $ret =  $content[$index];
				 }
			 } else {
				 foreach my $item (@content){
					 if ( $item->{'NAME'}->() eq $index ) {
						 $ret = $item;
						 last;
					 }
				 }
			 }
			 return $ret;
	   },
	   'KINDSREF'=> sub {
			 my $name = shift;
			 my $ret = \@content;
			 if ( $name ) {
				 $ret = [];
				 foreach my $item (@content) {
					 if ( $item->{'NAME'}->() eq $name ) {
						 push @{$ret}, $item;
					 }
				 }
			 }
	     return $ret;
	   },
	   'ATTRIBUTESREF' => sub {
	     return \%attributes;
	   },
	   'PRINT'=>sub {
	     my $fh = shift || *STDOUT;
	     my $addPrefix = shift || '';
	     my $prefix = shift || '';
	     if($name){
	       my ($out,$item);
	       if(@content || defined($plain)){

					 $out = $prefix . &getTag($name,\%attributes,'start');
					 print $fh $out;
					 if(@content){
						 print $fh "\n";
						 foreach $item (@content){
							 if(ref($item)=~/HASH/ && ref($item->{'PRINT'})=~/CODE/){
								 $item->{'PRINT'}->($fh,$addPrefix,$prefix.$addPrefix);
							 }
						 }
					 }
					 if(defined($plain)){
						 print $fh $plain;
						 if(@content){
							 print $fh "\n$prefix";
						 }
					 }else{
						 if($prefix){
							 print $fh $prefix;
						 }
					 }
					 $out = &getTag($name,\%attributes,'stop') . "\n";
					 print $fh $out;

	       }else{
					 $out = $prefix . &getTag($name,\%attributes) . "\n";
					 print $fh $out;
	       }
	     }else{
	       if(defined($plain)){
					 print $fh $plain;
	       }
	     }
	   }
	 };

  return $this;
}

sub getTag {
  my ($name,$ptrAttr,$mode) = @_;
  my ($out,$key,$value);
  $mode||='';
  if ($mode eq 'stop'){
    $out="</$name>";
  }else{
    $out="<$name";
    while (($key,$value)=each(%{$ptrAttr})){
      $out .= ' '.$key;
      if (defined($value)){
		$out .= '="'.$value.'"';
      }
    }
    if($mode eq 'start'){
      $out .= ">";
    }else{
      $out .= "/>";
    }
  }
  return $out;
}

#
# end class for handle with XML
#


#
# class for handle with database
#
sub getDbConnect {
  my $ptrBin = shift;
  my $myDbType = shift;
  my $myDbAddress = shift ;
  my $myDbUser = shift ;
  my $myDbPasswd = shift ;
  my $myDbDB = shift ;
  my $myDbHost = shift ;

  #added by Honig for Sphera
  my $myDbSocket = shift;
  my $variables = shift;

  my ($dbh,$this,$sth);
  my ($shell,$args);
  my $echo = $ptrBin->{'echo'}||'echo';
  my ($sql,@result,$resultIndex,$ptrResult);

  unless(defined($myDbPasswd) && defined($myDbUser)){
    if($myDbType eq 'mysql'){
      foreach my $myCnf ('/etc/my.cnf',$ENV{'HOME'}.'/.my.cnf'){
		next unless (-T $myCnf);
		next unless open(CNF,"<$myCnf");
		binmode CNF;
		my($key,$value);
		my $state=0;
		while (<CNF>){
		  next unless $_;
		  next if (/^#/);
		  if($state){
			if(/^\s*\[/){
			  last;
			}
			if(/^\s*(\S+)\s*=\s*(\S+)/){
			  $key=$1;
			  $value=$2;
			  if($key =~ /^user$/){
				$myDbUser = $value;
				# sometimes strings are quoted
				$myDbUser =~ s/^('|")(.*)\1$/$2/;
			  }elsif($key =~ /^pass$/){
				$myDbPasswd = $value;
				# sometimes strings are quoted
				$myDbPasswd =~ s/^('|")(.*)\1$/$2/;
			  }
			}
		  }else{
			if(/^\s*\[client\]/){
			  $state=1;
			}
		  }
		}
		close (CNF);
      }
    }elsif($myDbType eq 'Pg'){
      $myDbPasswd ||= $ENV{'PGPASSWORD'};
      $myDbUser ||= $ENV{'PGUSER'};
    }
  }

  if (eval "require DBI"){
    if (eval "require DBD::$myDbType"){
      if($myDbAddress){
		if($myDbUser && $myDbPasswd){
		  $dbh = DBI->connect($myDbAddress,$myDbUser,$myDbPasswd) or
			die "Error: unable to connect to database: $DBI::errstr\n";
		}elsif($myDbUser){
		  $dbh = DBI->connect($myDbAddress,$myDbUser) or
			die "Error: unable to connect to database: $DBI::errstr\n";
		}else{
		  $dbh = DBI->connect($myDbAddress) or
			die "Error: unable to connect to database: $DBI::errstr\n";
		}
      }
    }
  }
#  unless($dbh){
#    if ($myDbType eq 'Pg'){
#      if(eval "require Pg"){
#        $myDbAddress="user=$myDbUser dbname=$myDbDB";
#        if($myDbPasswd){
#          $myDbAddress.=" password=$myDbPasswd";
#        }
#        $dbh = Pg::connectdb($myDbAddress);
#      }
#    }
# }
	unless ($dbh) {

		if ($myDbType eq 'mysql') {
			$shell = $ptrBin->{'mysql'} || 'mysql';
			$args  = " --socket='$myDbSocket'" if ($myDbSocket);
			if ($myDbPasswd) {
				$args .= " -B -u '$myDbUser' -p'$myDbPasswd' '$myDbDB' ";
			} else {
				$args .= " -B -u $myDbUser $myDbDB ";
			}
			$shell = "LD_LIBRARY_PATH=$variables $shell" if ($variables);
    }elsif($myDbType eq 'Pg'){
      $shell = $ptrBin->{'psql'}||'psql';
	  if($myDbPasswd) {
		$shell = "PGPASSWORD=$myDbPasswd $shell";
	  }
	  if($myDbUser) {
		$shell = "PGUSER=$myDbUser $shell";
	  }

      $args = " -d $myDbDB -c ";
    }
#
# check connection to Confixx databnase
#
    if($myDbDB=~/confixx/){
      $sql = "select hostname from admin";
      @result = `$echo "$sql"| $shell $args`;
      unless($result[0]=~/^\w+$/){
		$args = ' -h $myDbHost'.$args;
		@result = `$echo "$sql"| $shell $args`;
		unless($result[0]=~/^\s*hostname/){
		  die "Error: Can not connect to Confixx database (unable to execute sql: $sql)\n@result\n";
		}
      }
    }
#
# end check connection to Confixx databnase
#
    $shell .= $args;
  }
  $this = {
	   'RECONNECT'=>sub{
	     if($dbh){
	       $dbh->disconnect();
		   if($myDbPasswd){
			 $dbh = DBI->connect($myDbAddress,$myDbUser,$myDbPasswd) or
			   die "Error: unable to connect to database: $DBI::errstr\n";
		   }else{
			 $dbh = DBI->connect($myDbAddress,$myDbUser) or
			   die "Error: unable to connect to database: $DBI::errstr\n";
		   }
	       return $dbh;
	     }else{
	       return 1;
	     }
	   },
	   'EXECUTE'=> sub {
	     $sql = shift;
	     if($dbh){
	       $sth = &executeDbh($dbh,$sql);
	       if($sth){
		 return $sth->rows;
	       }else{
		 return 0;
	       }
	     }else{
	       $resultIndex=0;
	       &executeDbShell($shell,\@result,$echo,$sql,$myDbType);
	       return scalar(@result);
	     }

	   },
	   'FETCHROW'=> sub {
	     if($sth){
	       if(@result = $sth->fetchrow){
		 return \@result;
	       }else{
		 return undef;
	       }
	     }else{
	       $resultIndex++;
	       if($resultIndex>$#result){
		 return undef;
	       }else{
		 return $result[$resultIndex];
	       }
	     }
	   },
	   'FETCHHASH'=>sub {
	     if($sth){
	       $ptrResult = $sth->fetchrow_hashref();
	     }else{
	       $resultIndex++;
	       if($resultIndex>$#result){
		 $ptrResult = undef;
	       }else{
			 my ($i,$key);
			 $ptrResult = {};

			 for($i=0; $i<@{$result[0]}; $i++){
			   $ptrResult->{$result[0]->[$i]}=$result[$resultIndex]->[$i];
			 }
	       }
	     }
	     return $ptrResult;
	   },
	   'FINISH'=>sub {
	     if($sth){
	       $sth->finish();
	     }
	     @result=();
	     $ptrResult = undef;
	   },
	   'PRINTERROR'=>sub {
	     if($dbh){
	       if(@_){
			 $dbh->{'PrintError'}=shift;
	       }
	       return $dbh->{'PrintError'};
	     }else{
	       return 0;
	     }
	   },
	   'PING'=>sub{
	     if($dbh){
	       return $dbh->ping();
	     }else{
	       return 1;
	     }
	   },
	   'DISCONNECT'=>sub{
	     if($dbh){
	       return $dbh->disconnect();
	     }else{
	       return 1;
	     }
	   },
	   'USER'=>sub{
	     return $myDbUser;
	   },
	   'PASSWORD'=>sub{
	     return $myDbPasswd;
	   },
	   'HOST'=>sub{
	     return $myDbHost;
	   },
	   'TYPE'=>sub{
	     return $myDbType;
	   },
	  };

  return $this;
}

sub executeDbh {
  my $dbh = shift;
  unless(ref($dbh)=~/DBI/){
    &printToError ("Error: DBI object is not found");
  }
  my $sql = shift;

  my $sth = $dbh->prepare($sql) or
    &printToError ("Error: prepare sql: $sql\n$DBI::errstr");
  if($sth->execute){
    return $sth;
  }else{
    &printToError ("Error: exesute sql: $sql\n$DBI::errstr");
  }
}

sub executeDbShell {
  my ($shell,$ptrRes,$echo,$sql, $type) = @_;
  @{$ptrRes}=();
  my (@out,@row,$ret);
  $ret=0;
  if($type eq 'Pg'){
    @out = `$shell "$sql"`;
  }else{
    @out = `$echo "$sql"|$shell`;
  }
  if ($out[0]=~/^error/i){
    &printToErrror ("Error: execute sql: $sql\n@out");
  }else{
    if($type eq 'Pg'){
      foreach my $line (@out){
	chomp $line;
	next if ($line=~/^-[-+]*$/); # skip delimiter
	if($line=~/^\(\s*\d+\s+\w+\)$/){
	    last;  ## (6 rows)
	}
	$line=~s/^\s+//;
	$line=~s/\s+$//;
	@row = split(/\s*\|\s*/,$line);
	push @{$ptrRes}, [@row];
	$ret++;
      }
    }else{
      if($out[0]=~/^\+[-+]*$/){
	foreach my $line (@out){
	  chomp $line;
	  next if ($line=~/^\+[-+]*$/); # skip delimiter
	  $line=~s/^\|\s*//;
	  $line=~s/\s+$//;
	  @row = split(/\s*\|\s*/,$line);
	  @row = map { $_ eq 'NULL' ? undef : $_ } @row;
	  push @{$ptrRes}, [@row];
	  $ret++;
	}
      }else{  ## batch mode
	foreach my $line (@out){
	  chomp $line;
	  @row = split(/\t/,$line);
	  @row = map { $_ eq 'NULL' ? undef : $_ } @row;
	  push @{$ptrRes}, [@row];
	  $ret++;
	}
      }
    }
  }
  return $ret;
}

#
# end class for handle with database
#

sub makeMIMEBase64 {
  my ($useModule,$this);
  if (eval "require MIME::Base64"){
    $useModule = 1;
  }else{
    $useModule = 0;
  }

  $this={
	 'ENCODE'=>sub{
	   my $src = shift;
	   my ($encoded);
	   if($useModule){
	     $encoded = MIME::Base64::encode($src);
		 chomp $encoded;
	   }else{
	     $encoded = &encodeBase64($src);
	   }
	   return $encoded;
	 },
	 'USEMODULE'=>sub{
	   return $useModule;
	 },
	};

  return $this;
}

#
# url decode (certificates in Plesk DB encoded by the url encoding)
#

sub urlDecode {
    my $url = $_[0];
    $url =~ tr/+/ /;
    $url =~ s/%([a-fA-F0-9]{2,2})/chr(hex($1))/eg;
    return $url;
}

sub encodeHtml {
  my(@ret,$src);
  if ($src=shift){
    if(ref $src){
      if(${$src}=~/[&"'><]/){
		${$src}=~s/&/&amp;/sg;
		${$src}=~s/>/&gt;/sg;
		${$src}=~s/</&lt;/sg;
		${$src}=~s/\'/&#39;/sg;
		${$src}=~s/\"/&quot;/sg;
      }
      return ${$src};
    }else{
      if($src=~/[&"'><]/){
		$src=~s/&/&amp;/sg;
		$src=~s/>/&gt;/sg;
		$src=~s/</&lt;/sg;
		$src=~s/\'/&#39;/sg;
		$src=~s/\"/&quot;/sg;
      }
    }
  }
  return $src;
}

# Normalization according to the XML 1.0 (Third Edition),
# clauses 2.3 "Common Syntactic Constructs"
# and 3.3.3 "Attribute-value normalization"

sub normalizeAttribute {
  $_ = $_[0];
  s/\r/&#xD;/g;
  s/\n/&#xA;/g;
  s/\t/&#x9;/g;
  s/\</&lt;/g;
  s/\&/&amp;/g;
  s/\"/&quot;/g;
  s/\'/&#39;/g;
  return $_;
}

############################
#
#	Non-ascii characters handling
{
	# static
	my $charset;

	sub setCharset {
	  $charset = shift;
	}

	# try to determine the local charmap
	sub _determineCharset()
	{
	  my $charset;
	  my $locale = $bin{'locale'} || 'locale';

	  if (exists $ENV{'LANG'}) {
		my @pieces = split /\./, $ENV{'LANG'};
		if (defined $pieces[1]) {
		  $charset = $pieces[1];
		  chomp $charset;
		}
	  }
	  if (not $charset && open(CM, "$locale charmap |")) {
	  	binmode CM;
		$charset = <CM>;
		close(CM);
		chomp $charset;
	  }
	  if (not $charset) {
		$charset = 'ISO-8859-1';
	  }

	  return $charset;
	}

	sub toUtf8
	{
		my ($subj, $myCharset) = @_;

		$charset = _determineCharset() if (not $charset);

		$myCharset = $charset unless ( $myCharset );

		if ( (not $subj) ||
				 (not $bin{'iconv'}) ||
				 ( $myCharset =~ /UTF-?8/i )
			 ) {
			return $subj;
		}

		if ($subj =~ /^[\x00-\x7f]*$/) { return $subj; }

		my $converted;
		my $escaped = $subj;
		$escaped =~ s/'/'"'"'/g;

		if (open(IC, "echo '$escaped' | " . $bin{'iconv'} . " -f '$myCharset' -t 'UTF-8' |")) {
			binmode IC;
			while (my $string = <IC>) { $converted .= $string; }
			close(IC);
			chomp $converted;
			if ($converted) {
			  return $converted;
			}
		}

		return "";
	}
}

#
# Based on the b64.c by Bob Trower (base64.sf.net)
#

my @cb64 = ('A', 'B', 'C', 'D', 'E', 'F', 'G', 'H',
			'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P',
			'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X',
			'Y', 'Z', 'a', 'b', 'c', 'd', 'e', 'f',
			'g', 'h', 'i', 'j',	'k', 'l', 'm', 'n',
			'o', 'p', 'q', 'r', 's', 't', 'u', 'v',
			'w', 'x', 'y', 'z', '0', '1', '2', '3',
			'4', '5', '6', '7', '8', '9', '+', '/');

@cb64 = map { ord } @cb64;

sub encodeBlock {
  my ($len, @in) = @_;
  my @res;
  push @res, $cb64[$in[0] >> 2];
  push @res, $cb64[(($in[0] & 0x3)<<4)|(($in[1]&0xf0)>>4)];
  push @res, $len>1 ? $cb64[(($in[1]&0xf)<<2)|(($in[2]&0xc0)>>6)] : ord('=');
  push @res, $len>2 ? $cb64[$in[2]&0x3f] : ord('=');
  return @res;
}

sub encodeBase64 {
  my @in = unpack "C*", $_[0];

  my @out;
  while (scalar(@in)) {
	my @block = (0, 0, 0);
	my $length = 0;
	while (scalar(@in) && $length < 3) {
	  $block[$length++] = shift(@in);
	}
	push @out, encodeBlock($length, @block);
  }
  return pack "C*", @out;
}

sub getBins {
  my ($ptrBin,$path) = @_;
  my (%searchPath,$key,$prefix,$bin,$fullPath);
  my ($notFound,$i,@prefixes);

  $notFound=0;

  $i=0;
  foreach $prefix (split(/:/,$path)){
    next unless $prefix;
    unless(exists($searchPath{$prefix})){
      $searchPath{$prefix}=$i++;
      push @prefixes,$prefix;
    }
  }

  foreach $key (keys %{$ptrBin}){
	$ptrBin->{$key} = undef;
    foreach $prefix (@prefixes){
      $prefix.='/' unless $prefix=~/\/$/;
      $bin = $key;
      $bin =~ s/^\///;
      $fullPath = $prefix.$bin;
      if (-x $fullPath){
		$ptrBin->{$key}=$fullPath;
		last;
      }
    }
    $notFound++;
  }
  return $notFound;
}

sub getArguments {
  my ($ptrArgv,$ptrOpts) = @_;

  my (@keys,$firstKey,%prepKeys,$key,$value,$ptrArr,$arg,$state);
  my (%retHash,$pat,$found,$used,@rest,$fullArg,$prevKey);

  while (($key,$value)=each(%{$ptrOpts})){
    @keys = split(/\|/,$key);
    $firstKey = $keys[0];
    $firstKey =~s/^-*//;
    push @{$prepKeys{$firstKey}},[@keys];
    push @{$prepKeys{$firstKey}},$value;
  }

  $state =0;
  foreach $arg (@{$ptrArgv}){

#    print "arg: $arg\n";

    $used = 0;
    if($state==1){
      if  ($arg =~ /^-/){
	$state=0;
      }else{
	$used = 1;
	$retHash{$prevKey}=$arg;
	next;
      }
    }
    if ($state == 2){
      $retHash{$prevKey}=$arg;
      $state = 0;
      $used = 1;
    }else{
      $fullArg = $arg;
      if ($arg =~ /^(-\S+)=(\S+)/){
	$arg = $1;
	$value = $2;
      }else{
	$value = undef;
      }
      foreach $key (keys %prepKeys){

	$ptrArr = $prepKeys{$key};

#	print "key: $key, ",join('|',@{$ptrArr->[0]}),"\n";


	$found = 0;
	foreach $pat (@{$ptrArr->[0]}){
	  if ($pat eq $arg){
	    $found = 1;
	    last;
	  }
	}
	if($found){
	  $used = 1;
	  if(defined($value)){
	    $retHash{$key}=$value;
	  }else{
	    if($ptrArr->[1]){
	      if($ptrArr->[1] =~ /\?$/){
		$state = 1;
	      }else{
		$state = 2;
	      }
	      $prevKey = $key;
	    }else{
	      $retHash{$key}=undef;
	    }
	  }
	  last;
	}
      }
    }
    unless($used){
      if($value){
	push @rest,$fullArg;
      }else{
	push @rest,$arg;
      }
    }
  }
  @{$ptrArgv}=@rest;
  return \%retHash;

}

sub makeDumpStatus($) {
  my $statusFile = shift;
  my($pid,$currAccount,$currDomain,$leftAccounts,$leftDomains,$this);
  $pid=$$;
  $this = {'FILE'=>sub {
	     if(@_){
	       $statusFile = shift;
	     }
	     return $statusFile;
	   },
	   'PID'=>sub {
	     if(@_){
	       $pid = shift;
	     }
	     return $pid;
	   },
	   'ACCOUNT'=>sub{
	     if(@_){
	       $currAccount = shift;
	       if($leftAccounts){
		 $leftAccounts--;
	       }
	     }
	     return $currAccount;
	   },
	   'DOMAIN'=>sub{
	     if(@_){
	       $currDomain = shift;
	       if($leftDomains){
		 $leftDomains--;
	       }
	     }
	     return $currDomain;
	   },
	   'COUNTS'=>sub{
	     if(@_){
	       my $newValue = shift;
	       if (defined $newValue){
		 $leftAccounts=$newValue;
	       }
	       $newValue = shift;
	       if (defined $newValue){
		 $leftDomains=$newValue;
	       }
	     }
	     return ($leftAccounts,$leftDomains);
	   },
	   'PRINT'=>sub{
	     if(@_){
	       $statusFile=shift;
	     }
	     open(STATUS,">$statusFile.tmp");

		 binmode STATUS;
	     my $root = &makeXmlNode('agent-dump-status');
	     my $item = &makeXmlNode('pid',$pid);
	     $root->{'ADDCHILD'}->($item);
	     $item = &makeXmlNode('account',$currAccount||'');
	     $root->{'ADDCHILD'}->($item);
	     $item = &makeXmlNode('domain',$currDomain||'');
	     $root->{'ADDCHILD'}->($item);
	     $item = &makeXmlNode('accounts-left',
				  defined($leftAccounts)?$leftAccounts:'');
	     $root->{'ADDCHILD'}->($item);
	     $item = &makeXmlNode('domains-left',
				  defined($leftDomains)?$leftDomains:'');
	     $root->{'ADDCHILD'}->($item);

	     &printXml($root,*STATUS,'agent-dump-status.dtd');

	     close(STATUS);

		 rename("$statusFile.tmp", $statusFile);
	   }

	  };
  return $this;
}

sub makeContentList {
  my ($dirDump,$xmlDump) = @_;
  my ($name,$fullPath,%files,%fromXML,$size,@cids,$root,$item,$key,$cmd,
      $gzip,$wc,$md5sum);

  if(opendir(DUMP,$dirDump)){
    while ($name=readdir(DUMP)){
      $fullPath="$dirDump/$name";
      unless(-f $fullPath && $name=~/\.\w\w$/){
	next;
      }
      $key = $fullPath;
      $key =~ s/\.\w\w$//;
      $size=(stat($fullPath))[7];
      if(defined $size){
	push @{$files{$key}},[$fullPath,$size];
      }
    }
    closedir(DUMP);
  }

	if (-T $xmlDump) {
		if (open(DUMP, "<$xmlDump")) {
			binmode DUMP;
			while (<DUMP>) {
				if (@cids = /cid(?:_\w+)?=\"([^"]+)\"/g) {
					foreach $fullPath (@cids) {
						unless ($fullPath =~ /^\//) {
							$fullPath = "$dirDump/$fullPath";
						}
						$fromXML{$fullPath} = undef;
					}
				}
			}
			close(DUMP);
		}

		foreach $key (keys %files) {
			unless (exists($fromXML{$key})) {
				delete $files{$key};
			}
		}
	}

  $gzip = $bin{'gzip'}||'gzip';
  $wc = $bin{'wc'}||'wc';
  $md5sum = $bin{'md5sum'}||'md5sum';

  $root=&makeXmlNode('contentlist');

  my($compressed,$uncompressed,$list,$count,$ptrList,@sortedList,
     $ptrItem,$out);
  my($totalUncomp,$totalSize)=(0,0);
  foreach $key (keys %files){

#
# sort by extension  .aa .ab ...
#
    @sortedList =  sort {&getExt($a->[0]) cmp &getExt($b->[0])} @{$files{$key}};

#
# output list of files
#
    $compressed=0;
    foreach $ptrItem (@sortedList){
      ($fullPath,$size)=@{$ptrItem};
      $item=&makeXmlNode('file',$fullPath);
      $item->{'ATTRIBUTE'}->('size',$size,'noencode');
      $root->{'ADDCHILD'}->($item);

#
# get md5sum
#
      if(-x $md5sum){
	$out=`$md5sum $fullPath`;
	if($out=~/\s*([0-9a-fA-F]+)\s+/){
	  $item->{'ATTRIBUTE'}->('md5sum',$1,'noencode');
	}
      }

      $compressed += $size;

    }

    $uncompressed = undef;
#
# get unpacked size of the archive
#
    if ($key=~/\.tgz$|\.gz$/){
      if(-x $gzip){
	if(scalar(@sortedList)==1){
	  ($fullPath,$size)=@{$sortedList[0]};
	  ($uncompressed,$compressed)=&getZipSize($fullPath,$gzip);
	}else{
	  $list = join(' ',map {$_->[0]} @sortedList);
	  if(-x $wc){
	    $cmd = "cat $list | $gzip | $wc -c";
	    $uncompressed = `$cmd`;
	    chomp $uncompressed;
	    if ($uncompressed =~ /^\s*(\d+)\s*$/){
	      $uncompressed = $1;
	    }else{
	      $uncompressed = undef;
	    }
	  }else{
	    $cmd = "cat $list >$key";
	    if(system($cmd)==0){
	      ($uncompressed,$compressed)=&getZipSize($key,$gzip);
	      unlink $key;
	    }
	  }
	}
      }
    }
#
# calc total size
#
    $totalSize += $compressed;
    $totalUncomp += $uncompressed if $uncompressed;
  }

  $root->{'ATTRIBUTE'}->('size',$totalSize,'noencode');
  if ($totalUncomp && $totalUncomp != $totalSize){
    $root->{'ATTRIBUTE'}->('unpacked',$totalUncomp,'noencode');
  }

  return $root;
}

sub getExt {
  my $path = shift;
  if($path=~/\.([^.]*)$/){
    return $1;
  }
  return $path;
}

sub getZipSize {
  my ($fullPath,$gzip) = @_;
  $gzip ||= 'gzip';
  my $cmd = "$gzip -l $fullPath";
  my ($uncompressed,$compressed,@out);
  if (open(OUTPUT,"$cmd|")){
    ($compressed,$uncompressed)=(0,0);
    while(<OUTPUT>){
      @out = split(' ',$_);
      if($out[0]=~/^\d+$/ && $out[1]=~/^\d+$/){
	$uncompressed+=$out[1];
	$compressed+=$out[0];
      }
    }
    close (OUTPUT);
  }
  return $uncompressed,$compressed;
}

my $outputIsFile;

sub openOutput
{
	my ($outPath) = @_;
	my ($outFh);

	if ($outPath) {
		open(OUTPUT, ">>$outPath") or die "Error: can not open '$outPath' for write: $!\n";
		binmode OUTPUT;
		$outFh = *OUTPUT;
		$outputIsFile = 1;
	} else {
		$outFh = *STDOUT;
		$outputIsFile = 0;
	}

	return $outFh;
}

sub closeOutput{
  if($outputIsFile){
    close (OUTPUT) or
      &printToError("Error: unable to close OUTPUT: $!\n");
    $outputIsFile=0;
  }
}

#=========================================================================
#
# dtd-specific (but common) subroutines
#
sub createPasswordNode {
  my ($parent, $password, $type) = @_;

  my $item = makeXmlNode('password', $password);
  $item->{'ATTRIBUTE'}->('type', $type);
  $parent->{'ADDCHILD'}->($item);
}

sub createSysuserNode {
  my ($parent, $name, $password, $type) = @_;

  my $item = makeXmlNode('sysuser');
  $item->{'ATTRIBUTE'}->('name', $name);
  createPasswordNode($item, $password, $type);
  $parent->{'ADDCHILD'}->($item);
}

sub readFile {
  my ($fileName) = @_;

  unless (-f $fileName) {
	return undef;
  }

  open(INPUT, "< $fileName");
  binmode INPUT;
  my $content = join('', <INPUT>);
  close(INPUT);

  return $content;
}

1;

#
# end common subroutines
#
#=========================================================================


$ptrArgs = &getArguments (\@ARGV,\%arg_opts);

#
# search binaries
#

&getBins(\%bin,'/bin:/usr/bin:/usr/local/bin:/sbin:/usr/sbin:/usr/local/sbin:'.$ENV{'PATH'});

my ($outPath,$outFh,$outHas,$dumpFile,$statusFile,$app);

$app = $0;
if ($app=~/(.+)\/([^\/]+)$/){
  $workDir = $1;
  $app = $2;
}
$app=~s/\.pl$//;

$pwd=`$bin{'pwd'}`||'.';

chop $pwd;

if($workDir eq '.'){
  $workDir = $pwd;
}elsif($workDir !~ /^\//){
  $workDir = $pwd.'/'.$workDir;
}

$dumpFile = $workDir.'/RaQ550.xml';
$statusFile = $workDir.'/RaQ550.status';
$dumpDir = $workDir.'/pma';

my $objDumpStatus = &makeDumpStatus($ptrArgs->{'status-file'}||$statusFile);

if(!$ptrArgs->{'output'} &&
   (exists $ptrArgs->{'dump-all'} ||
    exists $ptrArgs->{'dump-accounts'}||
    exists $ptrArgs->{'dump-domains'}
   )
  ){
  $ptrArgs->{'output'}=$dumpFile;
}

if ($outPath = $ptrArgs->{'output'}) {

	open(OUTPUT, ">$outPath") or die "Error: can not open '$outPath' for write: $!\n";
	binmode OUTPUT;
	$dumpFile = $outPath;
	close(OUTPUT);
}

if(exists $ptrArgs->{'version'}){
  print $outFh $VERSION,"\n";
  exit 0;
}

#
# get MIME Base64 encoder
#
my $wrapBase64 = makeMIMEBase64();

my $rootName = 'RaQ550dump';
my $dtdName = 'RaQ550.dtd';

#
# Open connection to the Sausalito CCE (Cobalt's server engine interface)
#

my $cce;
unless((eval "require CCE") and $cce = new CCE) {
  $cce = undef;
} else {
  $cce->connectuds();
}

#
# RAW DUMP (DEBUG)
#

if(exists $ptrArgs->{'raw'}) {
  my (undef, @classes) = $cce->classes();

  my ($class, $oid, $namespace, $prop, $value);
  foreach $class (@classes) {
	print "Class " . $class . "\n";

	my @oids = $cce->find($class);
	foreach $oid (@oids) {
	  print $oid . ":\n";
	  my (undef, @namespaces) = $cce->names($oid);
	  push @namespaces, '';

	  foreach $namespace (@namespaces) {
		print $namespace . ": ";

		my (undef, $object) = $cce->get($oid, $namespace);

		while(($prop, $value) = each %{$object}) {
		  print $prop.'->'.$value.'; ';
		}
		print "\n";
	  }
	}
  }
  exit(0);
}

my %users;
my %domainToUsers;
my %domainToMaillists;
my %domainToAdmins;
my %adminToDomains;

my @allMaillists;

#---------------------------------------
# Cache required server data structures
#---------------------------------------

my @allDomains = $cce->find('Vsite');

# Create domain->users, domain->domain-admins relations

my @userlist = $cce->find('User');
my $user;
foreach $user (@userlist) {
  my (undef, $object) = $cce->get($user, '');
  $users{$object->{'name'}} = $user;

  unless (defined $domainToUsers{$object->{'site'}}) {
    $domainToUsers{$object->{'site'}} = [];
  }
  push @{$domainToUsers{$object->{'site'}}}, $user;

  if ($object->{'capLevels'} =~ /siteAdmin/) {
	unless ($domainToAdmins{$object->{'site'}}) {
	  $domainToAdmins{$object->{'site'}} = [];
	}
	push @{$domainToAdmins{$object->{'site'}}}, $object->{'name'};
  }
}

# Create lexically-first-domain-admin -> domains relation
my $domain;
foreach $domain (keys %domainToAdmins) {
  my @admins = sort @{$domainToAdmins{$domain}};
  if (@admins > 0) {
	$domainToAdmins{$domain} = $admins[0];
  } else {
	$domainToAdmins{$domain} = undef;
  }
}

# Create domain -> maillists relation

@allMaillists = $cce->find('MailList');
my $list;
foreach $list (@allMaillists) {
  my (undef, $object) = $cce->get($list, '');
  unless (defined $domainToMaillists{$object->{'site'}}) {
    $domainToMaillists{$object->{'site'}} = [];
  }
  push @{$domainToMaillists{$object->{'site'}}}, $list;
}

###########

my $defaultCobaltDomains = "Default Cobalt Domains";
my $defaultCobaltDomainsSysuser = "defaultcobaltdomains";

if(exists $ptrArgs->{'get-status'}){
  my $outFh = &openOutput($ptrArgs->{'output'});
  &printAgentStatus($outPath=$ptrArgs->{'output'}||$dumpFile,
		    $objDumpStatus->{'FILE'}->(),
		    $outFh);
  &closeOutput();
  exit 0;
}

unless (defined $cce) {
  die "Cannot connect to the RaQ Sausalito CCE";
}

if (exists ($ptrArgs->{'dump-all'}) ||
   exists ($ptrArgs->{'dump-accounts'}) ||
   exists ($ptrArgs->{'dump-domains'})) {

  my ($root,@accounts,@domains,$ptrAccounts,$ptrDomains, $value);
  if (exists $ptrArgs->{'no-content'}) {
    $::dumpDir = &getDumpDir("$workDir/pma");
  } else {
    $::dumpDir = &getDumpDir("$workDir/pma",'clean');
  }

  if (exists $ptrArgs->{'no-compress'}) {
	setCompress(0);
  }

  &printToLog("Work dir: $workDir");
  &printToLog("Dump file: $dumpFile");
  &printToLog("Status file: ".$objDumpStatus->{'FILE'}->());


  if ($value = $ptrArgs->{'dump-accounts'}){
  	if ($value eq "-") {
		$value = <STDIN>;
		chomp $value;
	}
    @accounts = split(/\s*,\s*/,$value);
    $ptrAccounts = \@accounts;
  }

  if ($value = $ptrArgs->{'dump-domains'}){
  	if ($value eq "-") {
		$value = <STDIN>;
		chomp $value;
	}
    @domains = split(/\s*,\s*/,$value);
    $ptrDomains = \@domains;
  }
#
# generate a xml dump
#

  $root = &getRaQ550Dump($dumpDir, $ptrAccounts, $ptrDomains, $rootName);

#
# print dump to output
#
  &printToLog('Print dump file ...',1);
  if(ref($root)=~/HASH/ && ref($root->{'PRINT'})=~/CODE/){
    my $outFh = &openOutput($ptrArgs->{'output'});

    &printXml($root,$outFh,$dtdName);

    &closeOutput();

  }
  &printToLog(' OK');

}elsif(exists $ptrArgs->{'get-content-list'}){

  $::dumpDir = &getDumpDir("$workDir/pma");

  my $root = &makeContentList($::dumpDir,$dumpFile);

  if(ref($root)=~/HASH/){
    my $outFh = &openOutput($ptrArgs->{'output'});

    &printXml($root,$outFh);

    &closeOutput();

  }

}else{
  &printHelp();
}

exit $returnCode;

#######################################################

sub printAgentStatus {
  my $dumpFile = shift;
  my $statusFile = shift;
  my $outFh = shift || *STDOUT;
  my $rootName = shift || 'agentstatus';

  my ($root,$util,$out,$item,$engineRoot);

  $root = &makeXmlNode($rootName);

  unless(defined($cce) && defined $bin{'iconv'}) {
	if (defined $bin{'iconv'}) {
	  $item = &makeXmlNode('wrong-platform', '');
	} else {
	  $item = &makeXmlNode('wrong-platform', 'no iconv found on the source host');
	}

    $root->{'ADDCHILD'}->($item);
    &printXml($root,$outFh,$rootName.'.dtd');
    exit;
  }

  $item = &makeXmlNode('dumpfile',$dumpFile);
  $root->{'ADDCHILD'}->($item);
  $item = &makeXmlNode('statusfile',$statusFile);
  $root->{'ADDCHILD'}->($item);

  &printXml($root,$outFh,$rootName.'.dtd');
}

sub getRaQ550Dump() {
  my ($dumpDir, $ptrAccounts, $ptrDomains, $rootName) = @_;

  my $root = makeXmlNode($rootName);

  if (ref($ptrDomains) =~ /ARRAY/) {
    createDomains($root, $ptrDomains);
  } elsif (ref($ptrAccounts) =~ /ARRAY/) {
	createAccounts($root, $ptrAccounts);
  } else {
    createAllAccounts($root);
  }

  return $root;
}

sub createAccounts {
  my ($parent, $ptrAccounts) = @_;

  my $account;
  foreach $account (@{$ptrAccounts}) {
	if ($account eq $defaultCobaltDomainsSysuser) {
	  createDomainsByOid($parent, \@allDomains);
	}
  }
}

sub createAllAccounts {
  my ($parent) = @_;

  createDomainsByOid($parent, \@allDomains)
}

sub createDummyClient {
  my $dummyclient = makeXmlNode('client');
  $dummyclient->{'ATTRIBUTE'}->('disk-quota', '-1');
  $dummyclient->{'ATTRIBUTE'}->('enabled', 'true');
  $dummyclient->{'ATTRIBUTE'}->('fullname', $defaultCobaltDomains);
  createSysuserNode($dummyclient, $defaultCobaltDomainsSysuser, '', 'empty');

  return $dummyclient;
}

sub createDomains {
  my ($parent, $ptrDomains) = @_;

  my $dummyclient = createDummyClient();

  my ($domainName);
  foreach $domainName (@{$ptrDomains}) {
    createDomain($dummyclient, $domainName);
  }

  $parent->{'ADDCHILD'}->($dummyclient);
}

sub createDomainsByOid {
  my ($parent, $ptrOids) = @_;

  my $dummyclient = createDummyClient();

  my $oid;
  foreach $oid (@{$ptrOids}) {
    createDomainByOid($dummyclient, $oid);
  }

  $parent->{'ADDCHILD'}->($dummyclient);
}

sub getDomainName {
  my ($props) = @_;

  if($props->{'hostname'} eq 'www') {
	return $props->{'domain'};
  }
  return $props->{'fqdn'};
}

sub createDomain {
  my ($parent, $domainName) = @_;
  my @il = $cce->find('Vsite', {'fqdn' => $domainName});
  if(@il == 0) {
	@il = $cce->find('Vsite', {'domain' => $domainName, 'hostname' => 'www'});
	if(@il == 0) {
	  printToLog("No $domainName found");
	  return;
	}
  }
  createDomainByOid($parent, $il[0]);
}

sub createDomainByOid {
  my ($parent, $oid) = @_;

  my ($ok, $props) = $cce->get($oid, '');
  unless($ok) {
	printToLog("Unable to retrieve $oid domain data");
	return;
  }

  my $item = makeXmlNode('domain');
  $parent->{'ADDCHILD'}->($item);

  $item->{'ATTRIBUTE'}->('name', getDomainName($props));
  $item->{'ATTRIBUTE'}->('ip', $props->{'ipaddr'});

  # No more info required in 'no-content' mode
  if (exists $ptrArgs->{'no-content'}) {
	return;
  }

  $item->{'ATTRIBUTE'}->('cid',
			 makeDumpFile("$dumpDir/$oid-www", $props->{'basedir'}."/web"), '.', {}, '-h');
  if($props->{'suspend'} eq '1') {
    $item->{'ATTRIBUTE'}->('state', 'false');
  } else {
    $item->{'ATTRIBUTE'}->('state', 'true');
  }

  if ($domainToAdmins{$props->{'name'}}) {
	$item->{'ATTRIBUTE'}->('admin-user-name',
						   normalizeSysuserName($domainToAdmins{$props->{'name'}}));
  };

  createDnsRecords($item, $props);

  my (undef, $diskprops) = $cce->get($oid, 'Disk');

  my $disk_quota = adaptLimit($diskprops, 'quota', sub { $_[0] * $megabyte });

  my $disklimitnode = makeXmlNode('limit', $disk_quota);
  $disklimitnode->{'ATTRIBUTE'}->('name', 'disk-quota');
  $item->{'ADDCHILD'}->($disklimitnode);

  createPermissionNodes($item, \%domain_permissions, $oid);

  createAnonFtp($item, $oid, $props->{'basedir'});

  createDomainUsers($item, $props->{'name'}, $props->{'basedir'}, getDomainName($props));

  createMaillists($item, $props->{'name'}, getDomainName($props));
}

sub createDnsRecords {
  my ($parent, $props) = @_;

  my @mx_aliases = $cce->scalar_to_array($props->{'mailAliases'});
  my @ptr_aliases = $cce->scalar_to_array($props->{'webAliases'});

  my $alias;
  foreach $alias (@mx_aliases) {
    $parent->{'ADDCHILD'}->(makeXmlNode('email-server-alias', $alias));
  }
  foreach $alias (@ptr_aliases) {
    $parent->{'ADDCHILD'}->(makeXmlNode('web-server-alias', $alias));
  }
}

sub createAnonFtp {
  my ($parent, $oid, $basedir) = @_;

  printToLog(".anonftp");

  my (undef, $anonftpprops) = $cce->get($oid, 'AnonFtp');

  if($anonftpprops->{'enabled'} eq '0') {
    return;
  }

  my $item = makeXmlNode('anonftp');

  my $disklimit = makeXmlNode('limit', adaptLimit($anonftpprops, 'quota', sub {$_[0]*$megabyte}));
  $disklimit->{'ATTRIBUTE'}->('name', 'disk-quota');
  $item->{'ADDCHILD'}->($disklimit);

  my $accountlimit = makeXmlNode('limit', adaptLimit($anonftpprops, 'maxConnections'));
  $accountlimit->{'ATTRIBUTE'}->('name', 'account');
  $item->{'ADDCHILD'}->($accountlimit);

  my $anonftpdump = makeDumpFile("$dumpDir/$oid-anonftp", $basedir . "/ftp", ".", "incoming");

  if(defined $anonftpdump) {
    $item->{'ATTRIBUTE'}->('cid', $anonftpdump);

    my $anonftpincomingdump = makeDumpFile("$dumpDir/$oid-anonftpincoming", $basedir . "/ftp/incoming");
    if(defined $anonftpincomingdump) {
      $item->{'ATTRIBUTE'}->('cid_incoming', $anonftpincomingdump);
    }
  }

  $parent->{'ADDCHILD'}->($item);
}

sub createDomainUsers {
  my ($parent, $site, $basedir, $domainName) = @_;

  printToLog(".users");

  my $userid;
  foreach $userid (@{$domainToUsers{$site}}) {
    createDomainUser($parent, $userid, $basedir, $domainName);
  }
}

sub createDomainUser {
  my ($parent, $userid, $basedir, $domainName) = @_;

  my (undef, $props) = $cce->get($userid, '');
  my (undef, $diskprops) = $cce->get($userid, 'Disk');

  printToLog("..user " . $props->{'name'});

  my $item = makeXmlNode('user');
  $item->{'ATTRIBUTE'}->('fullname', $props->{'fullName'});
  $item->{'ATTRIBUTE'}->('disk-quota', adaptLimit($diskprops, 'quota', sub {$_[0]*$megabyte}));

  if($props->{'enabled'} eq '1') {
    $item->{'ATTRIBUTE'}->('state', 'true');
  } else {
    $item->{'ATTRIBUTE'}->('state', 'false');
  }

  createSysuser($item, $props);
  createPermissionNodes($item, \%user_permissions, $userid);
  createMail($item, $userid, $domainName, $basedir);

  $parent->{'ADDCHILD'}->($item);
}

sub createMail {
  my ($parent, $userid, $domainName, $basedir) = @_;
  my (undef, $props) = $cce->get($userid, '');
  my (undef, $mailprops) = $cce->get($userid, 'Email');

  my $item = makeXmlNode('mail');
  $item->{'ATTRIBUTE'}->('mailname', $props->{'name'});

  createForwardNode($item, $mailprops, $domainName);
  createAliasesNode($item, $mailprops);
  createAutoresponderNode($item, $mailprops);
  createMailboxNode($item, $props->{'name'}, $userid, $basedir);

  $parent->{'ADDCHILD'}->($item);
}

sub createForwardNode {
  my ($parent, $mailprops, $domainName) = @_;

  if($mailprops->{'forwardEnable'} eq '1') {
	my @forwards = $cce->scalar_to_array($mailprops->{'forwardEmail'});

	return unless @forwards;

	@forwards = map { $_ !~ /@/ ? $_ . "@" . $domainName : $_ } @forwards;

	if(@forwards == 1) {
	  $parent->{'ADDCHILD'}->(makeXmlNode('forward', $forwards[0]));
	} else {
	  my $item = makeXmlNode('mailgroup');
	  my $forward;
	  foreach $forward (@forwards) {
		$item->{'ADDCHILD'}->(makeXmlNode('forward', $forward));
	  }
	  $parent->{'ADDCHILD'}->($item);
	}
  }
}

sub createAliasesNode {
  my ($parent, $mailprops) = @_;

  my @aliases = $cce->scalar_to_array($mailprops->{'aliases'});
  my $alias;
  foreach $alias (@aliases) {
	$parent->{'ADDCHILD'}->(makeXmlNode('alias', $alias));
  }
}

sub createAutoresponderNode {
  my ($parent, $mailprops) = @_;

  my $text = $wrapBase64->{'ENCODE'}->($mailprops->{'vacationMsg'});
  my $item = makeXmlNode('autoresponder', $text);

  $item->{'ATTRIBUTE'}->('state',
						 $mailprops->{'vacationOn'} eq '1' ? 'true' : 'false');

  $parent->{'ADDCHILD'}->($item);
}

sub createMailboxNode {
  my ($parent, $name, $userid, $basedir) = @_;

  my $mboxdir = $basedir."/users/".$name;

  my $item = makeXmlNode('mailbox');
  $item->{'ATTRIBUTE'}->('type', 'mbox');
  if(-r $mboxdir."/mbox") {
	my $cid = makeDumpFile("$dumpDir/${userid}-mbox", $mboxdir, 'mbox', [], '-h');
	$item->{'ATTRIBUTE'}->('cid', $cid);
  } else {
	`touch /tmp/mbox`;
	my $cid = makeDumpFile("$dumpDir/${userid}-mbox", "/tmp", "mbox", [], '-h');
	$item->{'ATTRIBUTE'}->('cid', $cid);
	`rm /tmp/mbox`;
  }
  $parent->{'ADDCHILD'}->($item);
}

sub createMaillists {
  my ($parent, $site, $domain_name) = @_;

  printToLog(".maillists");

  my $listid;
  foreach $listid (@{$domainToMaillists{$site}}) {
    createMaillistByOid($parent, $listid, $domain_name);
  }
}

sub createMaillistByOid {
  my ($parent, $oid, $domain_name) = @_;

  my (undef, $props) = $cce->get($oid, '');

  printToLog("..maillist " . $props->{'name'});

  my $mlnode = makeXmlNode('maillist');
  $mlnode->{'ATTRIBUTE'}->('owner', $props->{'moderator'}.'@'.$domain_name);
  if ($props->{'apassword'} eq '') {
	$mlnode->{'ADDCHILD'}->(makePasswordNode('', 'empty'));
  } else {
	$mlnode->{'ADDCHILD'}->(makePasswordNode($props->{'apassword'}, 'plain'));
  }
  $mlnode->{'ATTRIBUTE'}->('name', $props->{'name'});
  $mlnode->{'ATTRIBUTE'}->('post-policy', $props->{'postPolicy'});
  $mlnode->{'ATTRIBUTE'}->('subscribe-policy', $props->{'subPolicy'});

  my @local_recipients = $cce->scalar_to_array($props->{'local_recips'});
  my @remote_recipients = $cce->scalar_to_array($props->{'remote_recips'});

  my $recip;
  foreach $recip (@local_recipients) {
    my $recipient = makeXmlNode('recipient', $recip . '@' . $domain_name);
    $mlnode->{'ADDCHILD'}->($recipient);
  }
  foreach $recip (@remote_recipients) {
    my $recipient = makeXmlNode('recipient', $recip);
    $mlnode->{'ADDCHILD'}->($recipient);
  }
  $parent->{'ADDCHILD'}->($mlnode);
}

sub createPermissionNodes {
  my ($parent, $permissionsPtr, $oid) = @_;

  my ($limit, $namespace);
  while (($limit, $namespace) = each(%{$permissionsPtr})) {
    my (undef, $params) = $cce->get($oid, $namespace);
    if($params->{'enabled'} eq '1') {
      my $permissionNode = makeXmlNode('permission');
      $permissionNode->{'ATTRIBUTE'}->('name', $limit);;
      $parent->{'ADDCHILD'}->($permissionNode);
    }
  }
}

sub normalizeSysuserName {
  my ($name) = @_;
  $name =~ s/\./_/g;
  return $name;
}

sub createSysuser {
  my ($parent, $props) = @_;

  my $name = normalizeSysuserName($props->{'name'});
  my $passwdtype = "" ne $props->{'md5_password'} ? 'encrypted' : "" ne $props->{'password'} ? 'plain' : 'empty';
  my $passwd = $props->{'md5_password'} || $props->{'password'};
  createSysuserNode($parent, $name, $passwd, $passwdtype);
}

sub adaptLimit {
  my ($props, $name, $adapter) = @_;

  my $value = $props->{$name};
  if("" eq $value) {
    $value = -1;
  } elsif (defined $adapter) {
    $value = &$adapter($value);
  }
  return $value;
}

sub printHelp {

  print <<"HELP";

Agent of migration from 'RaQ550' to 'Plesk 7'

Usage:
  $0 <options>

Options:

  -s |--get-status           get status of the agent
  -dc|--dump-accounts=<list> a coma separated list of resellers to dump
  -dd|--dump-domains=<list>  a coma separated list of customers to dump
  -da|--dump-all             make a full dump

  -lc|--get-content-list     get list of content files
  -nc|--no-content           do not make content files
  -nz|--no-compress          do not compress content files

  -o |--output=<path>        redirect output to file

  -h |--help                 this help
  -d |--debug=<level>        level of debug messages
  -v |--version

HELP
}

