#!/usr/bin/perl

# use for dump of user's mysql databases
use vars qw($mysqlUserUser $mysqlUserServer $mysqlUserPw);

use vars qw( %bin $dumpDir $ptrArgs $pwd $workDir);
use strict;

my $VERSION='1.0.1';

my $mysqlUserUser='root';
my $mysqlUserPw=undef;
my $mysqlUserServer='localhost';
my $DEFAULT_RESELLER='root';

my %mapLimits = ('max_addon'=>'MAXADDON',
				 'max_parked'=>'MAXPARK',
				 'max_ftpusers'=>'MAXFTP',
				 'max_maillists'=>'MAXLST',
				 'max_box'=>'MAXPOP',
				 'max_db'=>'MAXSQL',
				 'max_subdom'=>'MAXSUB',
				);


#
# parse command line
#
my %arg_opts = ('--help|-h'=>'',
				'--config|-c'=>'s',
				'--debug|-d'=>'i?',
				'--get-status|-s'=>'',
				'--dump-accounts|-dc'=>'s',
				'--dump-domains|-dd'=>'s',
				'--dump-all|-da'=>'',
				'--get-content-list|-lc'=>'',
				'--no-content|-nc'=>'',
				'--no-compress|-nz'=>'',
				'--output|-o'=>'s',
				'--status-file|-sf'=>'s',
				'--version|-v'=>'',
			   );

%bin = ('tar'=>'tar','split'=>'split','cp'=>'cp','rm'=>'rm','echo'=>'echo',
		'mysql'=>'mysql','mysqldump'=>'mysqldump',
		'psql'=>'psql','uptime'=>'uptime','uname'=>'uname',
		'cat'=>'cat','pwd'=>'pwd','md5sum'=>'md5sum',
		'grep'=>'grep','find'=>'find','gzip'=>'gzip','wc'=>'wc',
		'locale'=>'locale', 'iconv'=>'iconv');

my $returnCode = 0;

my $noContent = 0;

#
# global variables
#
my ($ptrRowResellers,%resellers,@default_reseller_account,
	$ptrQuota,$ptrResellerInfo, %servers_ips);
my $ptrUserdomainsHash;
my %trueUserDomains;
my %anonftps;
#
# end global variables
#



#==============================================================
#
# common subroutines
#

sub printToLog {
  my ($msg, $noeol) = @_;
  if ($msg) {
	print $msg;
  }
  unless ($noeol) {
	print "\n";
  }
}

sub printToError {
  my ($msg, $noeol, $errCode) = @_;
  if ($msg) {
	print STDERR $msg;
  }
  unless ($noeol) {
	print STDERR "\n";
  }
  if (defined($errCode)) {
    $returnCode = $errCode; ## translate an error code to the return code
  } else {
    $returnCode = 127;
  }
}

#BEGINNOTINRAQ2

sub makeFileReader
{
	my ($filename, $doNotFail) = @_;
	local (*INPUT);

	unless (-T $filename) {

		if (defined($doNotFail)) {
			&printToLog("Error: file '$filename' is not a text file");
		} else {
			&printToError("Error: file '$filename' is not a text file");
		}

		return undef;
	}

	unless (open(INPUT,"<$filename")) {

		binmode INPUT;
		if (defined($doNotFail)) {
			&printToLog("Error: unable to open file '$filename' to read");
		} else {
			&printToError("Error: unable to open file '$filename' to read");
		}

		return undef;
	}

	my $input = *INPUT;

	my $this = {
		'READ' => sub {
				my $a = <$input>;
				return $a;
			},

		'CLOSE' => sub { close($input); }

	};

	return $this;
}

sub makeArrayReader {
  my (@arr) = @_;
  my $size= 0;
  my $count = scalar(@arr);

  my $this={'READ'=>sub {
			  if($count == $size) {
				return undef;
			  }

			  return $arr[$count++];
			},
			'CLOSE'=>sub {
			}
		   };
  return $this;
}

sub makeParser {
  my $reader;
  my ($this,$ptrRows,$keySep,$valSep,$valKeySep,$comment,
      $typeRows,$maxIndex,$rowIndex);
  {
	$reader = shift;
	unless(defined($reader)) {
	  return undef;
	}

    my %args=(@_);
    $keySep = $args{'KEYSEPARATOR'};
    $valSep = $args{'VALUESEPARATOR'};
    $valKeySep = $args{'KEYVALUESEPARATOR'};
  }
  $this={'READER'=>sub {
		   if(@_){
			 $reader=shift;
		   }
		   return $reader;
		 },
		 'FILE'=>sub {
		   if(@_){
			 $reader = makeFileReader(@_);
		   }
		   return undef;
		 },
		 'KEYSEPARATOR'=>sub{
		   if(@_){
			 $keySep=shift;
		   }
		   return $keySep;
		 },
		 'VALUESEPARATOR'=>sub{
		   if(@_){
			 $valSep=shift;
		   }
		   return $valSep;
		 },
		 'KEYVALUESEPARATOR'=>sub{
		   if(@_){
			 $valKeySep=shift;
		   }
		   return $valKeySep;
		 },
		 'COMMENT'=>sub{
		   if(@_){
			 $comment=shift;
		   }
		   return $comment;
		 },
		 'PARSE'=>sub{
		   my %args=(@_);

		   $keySep=$args{'KEYSEPARATOR'} if exists $args{'KEYSEPARATOR'};
		   $valSep=$args{'VALUESEPARATOR'} if exists $args{'VALUESEPARATOR'};
		   $valKeySep=$args{'KEYVALUESEPARATOR'} if exists $args{'KEYVALUESEPARATOR'};
		   $comment=$args{'COMMENT'} if exists $args{'COMMENT'};

		   $ptrRows = &parseFile('reader'=>$reader,'keysep'=>$keySep,
								 'valsep'=>$valSep,'keyvalsep'=>$valKeySep,
								 'comment'=>$comment);
		   $typeRows = ref($ptrRows);
		   if($typeRows=~/ARRAY/){
			 $rowIndex=-1;
			 $maxIndex=scalar(@{$ptrRows})-1;
		   }
		   return $ptrRows;
		 },
		 'ROW'=>sub{
		   if($typeRows=~/HASH/){
			 my $key=shift;
			 return $ptrRows->{$key};
		   }elsif($typeRows=~/ARRAY/){
			 if($rowIndex<$maxIndex){
			   $rowIndex++;
			   return $ptrRows->[$rowIndex];
			 }else{
			   return undef;
			 }
		   }else{
			 return undef;
		   }
		 },
		};

  return $this;
}

sub makeFileParser {
  my $filename = shift;
  return makeParser(makeFileReader($filename), @_);
}

sub makeSafeFileParser {
  my $filename = shift;
  return makeParser(makeFileReader($filename, 'doNotFail'), @_);
}

sub makeStringParser {
  my $content = shift;
  return makeParser(makeArrayReader($content), @_);
}

sub parseFile {
  my %arg = (@_);
  my $reader = $arg{'reader'};
  my $keySep = $arg{'keysep'};
  my $valSep = $arg{'valsep'};
  my $valKeySep = $arg{'keyvalsep'};
  my $comment = $arg{'comment'};

  my ($ret,$key,$value,$reKey,$reValue,@values,$keyBlank,$valBlank,
     $vkey,$vvalue,$reKeyValue,$reComment);

  if($keySep){
    $ret = {};
    if(ref($keySep)=~/regexp/i || ($keySep eq ' ')){
      $reKey = $keySep;
    }else{
      $reKey = qr/$keySep/;
    }
  }else{
    $ret = [];
  }
  if($valSep){
    if(ref($valSep)=~/regexp/i || ($valSep eq ' ')){
      $reValue = $valSep;
    }else{
      $reValue = qr/$valSep/;
    }
  }
  if($valKeySep){
    if(ref($valKeySep)=~/regexp/i || ($valKeySep eq ' ')){
      $reKeyValue = $valKeySep;
    }else{
      $reKeyValue = qr/$valKeySep/;
    }
  }
  if($comment){
    if(ref($comment)=~/regexp/i){
      $reComment = $comment;
    }else{
      $reComment = qr/$comment/;
    }
  }

  while($_ = $reader->{'READ'}->()){
    chomp;
    next unless $_;
    if($comment && /$reComment/){  ## commentar
	  next;
    }
    if ($keySep){
      ($key,$value)=split($reKey,$_,2);
      if($key){
		$key=~s/^\s+//;
		$key=~s/\s+$//;
		if($valSep){

		  if ($valKeySep){
			$ret->{$key}={};
			foreach $value (split($reValue,$value)){
			  ($vkey,$vvalue)=split($reKeyValue,$value);
			  $vkey=~s/^\s+//;
			  $vkey=~s/\s+$//;
			  $ret->{$key}->{$value}=$vvalue;
			}
		  }else{
			push @{$ret->{$key}},split($reValue,$value);
		  }
		}else{
		  $value=~s/^\s+//;
		  $value=~s/\s+$//;
		  $ret->{$key}=$value;
		}
      }
    }else{
      if($valSep){
		push @{$ret},[split($valSep,$_)];
      }else{
		push @{$ret},$_;
      }
    }
  }

  $reader->{'CLOSE'}->();

  return $ret;
}

#ENDNOTINRAQ2

sub makePasswordNode {
  my $value= shift;
  my $type = shift||'encrypted';

  my $item = &makeXmlNode('password');
  $item->{'CONTENT'}->($value);  ## html encode
  $item->{'ATTRIBUTE'}->('type',$type);
  return $item;
}

sub mangleName {
  my $name = shift;

  if($name =~ /^\d/) {
    $name = 'user' . $name;
  }
  $name =~ s/\./_/g;

  return $name;
}

sub makeSysUserNode {
  my ($name, $password, $type) = @_;

  unless(defined($name) && defined($password) && defined($type)) {
	printToError("makeSysUserNode: Username, password or password type is not set.");
  }

  my $sysUserNode = makeXmlNode('sysuser');
  $sysUserNode->{'ATTRIBUTE'}->('name', mangleName($name));
  $sysUserNode->{'ADDCHILD'}->(makePasswordNode($password, $type));

  return $sysUserNode;
}

# Returns home directory for the system account $user
sub getHomeDir {
	my ($user) = @_;
	my $home;
	my @pw = getpwnam($user);
	if (-1 != $#pw) {
		$home = $pw[7];
	}
	unless ($home) {
		# how it is by default
		$home = '/home/$account';
	}
	return $home;
}

# Returns password for system account $user
sub getSystemPassword {
	my ($user) = @_;
	my $password;
	my @pw = getpwnam($user);
	if (-1 != $#pw) {
		$password = $pw[1];
	}
	unless ($password) {
		$password = '';
	}
	return $password;
}

sub createUserNameFromDomain {
  my ($name) = @_;

  if ($name =~ /^\d/) {
	$name = 'u' . $name;
  }

  $name =~ s/\./_/g;
  $name =~ s/^www\.//g;
  $name =~ s/[aoeui]//g if length($name) > 16;
  $name = substr($name, 0, 16) if length($name) > 16;

  return $name;
}

my $compress = 1;

sub setCompress {
  ($compress) = @_;
}

sub makeDumpDb
{
  my ($dbName, $dbType, $dbUser, $dbPasswd, $dbHost, $ptrDomainBin, $variables, $domainSocket) = @_;
  $dbType ||= 'mysql';

  my ($dumpFile,$dumpSql,$dumpName,$cmd,$gzip,$split,$mysql,$psql, $retCode,$out);

  $dumpName = "$dbName.$dbType.sql";
  $gzip = $compress == 1 ? ($bin{'gzip'} || 'gzip') : undef;

  if (-x $gzip) {
	$dumpName .= ".gz";
  }
  $dumpFile="$dumpDir/$dumpName";

  printToLog(">>Checking database presence");

  if ($dbType eq 'mysql') {
	$mysql = $bin{'mysql'} || 'mysql';
	if (-x $mysql) {
	  $cmd  = "$mysql -u '$dbUser'";
	  $cmd .= " -h $dbHost" if $dbHost;
	  $cmd .= " -p'$dbPasswd'" if $dbPasswd;
	  $cmd .= " -e '' '$dbName' 2>&1";

	  $out = `$cmd`;
	  chomp $out;
	  if ($out) {
		$retCode = $?>>8;
		printToLog("warning: mysql datatbase '$dbName' not found. mysql has returned '$retCode'");
		return "";
	  }
	}
  } elsif ($dbType eq 'Pg') {
	$psql = $bin{'psql'} || 'psql';
	if (-x $psql) {
	  $cmd  = "export PGUSER='$dbUser'";
	  $cmd .= " PGPASSWORD='$dbPasswd'" if $dbPasswd;
	  $cmd .= "; $psql '$dbName' ";
	  $cmd .= " -h $dbHost" if $dbHost;
	  $cmd .= " -c '' 2>&1";

	  $out = `$cmd`;
	  chomp $out;
	  if ($out) {
		$retCode = $?>>8;
		printToLog("warning: PostgreSQL datatbase '$dbName' not found. psql has returned '$retCode'");
		return "";
	  }
	}
  }

  printToLog("..done");

  unless (exists($ptrArgs->{'no-content'})) {
	$gzip = $bin{'gzip'} || 'gzip';
	$split = $bin{'split'} || 'split';

	if ($dbType eq 'mysql') {
	  my ($prefx, $postfix) = ("", "");
	  my $commonprefix = "LD_LIBRARY_PATH=$variables " if $variables;

	  my $commonopts;
	  $commonopts = " --socket='$domainSocket'" if $domainSocket;
	  $commonopts .= " -h $dbHost" if $dbHost;
	  $commonopts .= " -p'$dbPasswd'" if $dbPasswd;
	  $commonopts .= " -u '$dbUser' '$dbName'";

	  my $dumpcmd;
	  if ($ptrDomainBin) {
		$dumpcmd = $ptrDomainBin->{'mysqldump'};
	  } else {
		$dumpcmd = $bin{'mysqldump'} || 'mysqldump';
	  }

	  my $dumpopts = " --quick --quote-names --add-drop-table";

	  my $mysqlversion = `$commonprefix $bin{'mysql'} $commonopts --version`;
	  if ($mysqlversion =~ /Distrib (\d+)\.(\d+)\./) {
		if ($1 > 4 or ($1 == 4 and $2 > 0)) {
		  $dumpopts .= " --compatible=mysql323";
		}
	  }

	  $cmd = $commonprefix . $dumpcmd . $commonopts . $dumpopts;

	} elsif ($dbType eq 'Pg') {
	  $cmd  = "export PGUSER='$dbUser'";
	  $cmd .= " PGPASSWORD='$dbPasswd'" if($dbPasswd);
	  $cmd .= ' ; ' . ($bin{'pg_dump'} || 'pg_dump') . ' -d -O';
	  $cmd .= " -h '$dbHost'" if $dbHost;
	  $cmd .= " '$dbName'";
	}

	if ($compress and $gzip and -x $gzip) {
	  $cmd .= " | $gzip";
	}
	if (-x $split) {
	  $cmd .= " | $split -b 2040m - $dumpFile.";
	} else {
	  $cmd .= " > $dumpFile";
	}

	local $SIG{'PIPE'} = 'DEFAULT' if $SIG{'PIPE'} eq 'IGNORE';

	unless (system($cmd) == 0) {
	  printToError("Error execute '$cmd'");
	  return undef;
	}
  }
  return $dumpName;
}

sub addDbUsers {
  my ($dbNode, $dbName, $wrapUserMysql) = @_;

  printToLog("Dupming $dbName database users");

  $dbName =~ s/_/\\_/g;

  my (@users, $dbUser, $ptrRow);
  my $sql = "SELECT DISTINCT User FROM db WHERE Db = '$dbName'";
  $wrapUserMysql->{'EXECUTE'}->($sql);
  while ($ptrRow = $wrapUserMysql->{'FETCHROW'}->()) {
		push @users, (@{$ptrRow})[0];
  }
  $wrapUserMysql->{'FINISH'}->();

  foreach $dbUser (@users) {
		addDbUser($dbNode, $dbUser, $wrapUserMysql);
  }
}

sub addDbUser {
  my ($dbNode, $user, $wrapUserMysql, $passwordNode) = @_;

  &printToLog("Dumping mysql user '$user'");

	my ( $item );

  if ( ref( $wrapUserMysql ) =~ /HASH/ &&
			 ref( $wrapUserMysql->{'EXECUTE'} ) =~ /CODE/ ) {

    my ($sql,$ptrRow,@accesshosts,$accesshost);

    $sql = "SELECT password,host FROM user WHERE user='$user'";

    if ( $wrapUserMysql->{'EXECUTE'}->( $sql ) ) {

			$item = makeXmlNode( 'dbuser' );
			$item->{'ATTRIBUTE'}->('name', $user);

      while ( $ptrRow = $wrapUserMysql->{'FETCHROW'}->() ) {
				unless ( ref($passwordNode) =~ /HASH/ ) {
					$passwordNode = makePasswordNode( $ptrRow->[0] );
				}
				push @accesshosts, $ptrRow->[1];
      }

			if ( $passwordNode ) {
				$item->{'ADDCHILD'}->( $passwordNode );
			}

			foreach $accesshost ( @accesshosts ) {
				$item->{'ADDCHILD'}->( makeXmlNode( 'accesshost', $accesshost ) );
			}

			$dbNode->{'ADDCHILD'}->( $item );

    }

    $wrapUserMysql->{'FINISH'}->();
  }

	return $item;
}

my $lastDumpId = 0;

sub generateDumpId {
  return $$ . "-" . $lastDumpId++;
}

#
# sub makeDumpFile - make a dump of dir tree (use 'tar')
#
# Arguments:
#          $destFile - name of destination file.
#                      If the file has not '.tgz' or '.tar.gz' extension
#                      '.tgz' will be added
#          $scrDir   - path to source dir
#
#          $ptrNames - pointer to an array with names of files and dirs to
#                      include into dump
#                      can be a single name
#                      by default '.' - include all entries
#
#          $ptrExcludes - pointer to an array with names of files and dirs to
#                      exclude from dump
#                      can be a single name
#
#          $extraParams - extra params to run tar
#                         ('-h' - don't dump symlinks; dump the files they point to)
#
sub makeDumpFile {
  my ($destFile, $srcDir, $ptrNames,$ptrExclude,$extraParams)=@_;

  my @destPath = split(/\//, $destFile);
  if (length($destPath[$#destPath]) > 32) {
	$destPath[$#destPath] = generateDumpId();
  }
  $destFile = join('/', @destPath);

  my $tar = $bin{'tar'}||'tar';
  my $split = $bin{'split'}||'split';
  my $gzip = $compress == 1 ? ($bin{'gzip'} || 'gzip') : undef;

  my ($cmd,$ret,@names,@exclude);

  if(ref($ptrNames)=~/ARRAY/){
    @names = @{$ptrNames};
  }else{
    @names=($ptrNames) if $ptrNames;
  }

  if(ref($ptrExclude)=~/ARRAY/){
    @exclude = @{$ptrExclude};
  }else{
    @exclude=($ptrExclude) if $ptrExclude;
  }

  unless(-x $tar ){
    &printToError("Error: '$tar' is not executable");
    return undef;
  }

#
# check extension
#
  if(-x $gzip){
    unless($destFile=~/\.tgz$/ ||$destFile=~/\.tar\.gz$/){
      $destFile .= '.tar.gz';
    }
  }else{
    unless($destFile=~/\.tar$/){
      $destFile .= '.tar';
    }
  }
#
# end check extension
#

  unless( exists $ptrArgs->{'no-content'}){

    $cmd=$tar;
    if($srcDir){
      if( -d $srcDir){
		$cmd.=" -C $srcDir -f - -c --ignore-failed-read ";
      }else{
		return undef;
      }
    }


    if($extraParams){
      $cmd .= $extraParams.' ';
    }

    if(@names){
      if($srcDir){
		$cmd .= join( ' ', map {s/^\Q$srcDir\E\///;$_} @names); ## extract domain name
      }else{
		$cmd .= join( ' ', @names );
      }
    }else{
      $cmd .= '.';  ## dump all files & dirs
    }
    if(@exclude){
      $cmd .= ' '.join( ' ', map {s/^\Q$srcDir\E\///;'--exclude '.$_} @exclude); ## extract domain name
    }
    if(-x $gzip){
      $cmd .= " | $gzip";
    }
    if(-x $split){
      $cmd .= " | $split -b 2040m - $destFile.";
    }else{
      $cmd .= " >$destFile";
    }

		local $SIG{'PIPE'} = 'DEFAULT' if $SIG{'PIPE'} eq 'IGNORE';

    unless(system($cmd) == 0){
      &printToError("Error execute command: '$cmd': $!");
      $destFile='';
    }
  }

  $destFile=~s/^\Q$dumpDir\E\///;  ### remove dir name
  return $destFile;
}

sub getDumpDir {
  my ($subdir,$clean) = @_;
  my ($dumpdir);

  if($subdir=~/^\//){
    $dumpdir=$subdir;
  }else{
    my $util = $bin{'pwd'}||'pwd';
    $dumpdir = $pwd||`$util`;
    chomp $dumpdir;
    if($subdir){
      $dumpdir.='/'.$subdir;
    }
  }
  if(-d $dumpdir && $clean){
    unless(system("rm -rf $dumpdir")==0){
      &printToError("Error: can not clean temporary dir '$dumpdir': $!");
    }
  }
  unless(-d $dumpdir){
    mkdir ($dumpdir,0755) or
      die "Error: can not create temporary dir '$dumpdir': $!\n";
  }
  return $dumpdir;
}

sub printXml {
  my ($root,$outFh,$dtd) = @_;
  unless($outFh){
    $outFh = *STDOUT;
  }

  unless(ref($root)=~/HASH/ && ref($root->{'PRINT'})=~/CODE/){
    print $outFh "Error: a xml root is not found\n";
    &printToError("Error: sub printXml: bad arguments");
    return 0;
  }
  print $outFh "<?xml version='1.0' encoding=\"UTF-8\"?>\n";

  my  $doctype = $root->{'NAME'}->();
  if($dtd){
    print $outFh "<!DOCTYPE $doctype SYSTEM '$dtd'>\n";
  }
  $root->{'PRINT'}->($outFh,'  ');
  return 1;
}

#
# class for handle with XML
#
sub makeXmlNode {
  my ($this,$parent,%attributes,@content);
  my ($name,$plain)=@_;

  $plain = toUtf8(encodeHtml($plain));

  unless($name||$plain){
    return undef;
  }
  $this = {
	   'ATTRIBUTE' => sub {
	     my $key = shift;
	     if(@_){
	       my $value=shift;
	       if(@_){
					 $attributes{$key} = $value;
	       }else{
					 $attributes{$key} = toUtf8(normalizeAttribute($value));
	       }
	     }
	     return $attributes{$key};
	   },
	   'PARENT' => sub {
	     if(@_){
	       my $arg = shift;
	       if(ref($arg)=~/HASH/){
					 $parent=shift;
	       }
	     }
	     return $parent;
	   },
	   'NAME' => sub {
	     if(@_){
	       $name = shift;
	     }
	     return $name;
	   },
	   'CONTENT' => sub {
	     if(@_){
	       $plain = toUtf8(encodeHtml(shift));
	     }
	     return $plain;
	   },
	   'ADDCHILD'=> sub {
	     my $kind = shift;
	     if(
					(ref($kind)=~/HASH/) &&
					(ref($kind->{'PARENT'})=~/CODE/)
	       ) {
	       push @content,$kind;
	       $kind->{'PARENT'}->($this);
	     }
	     return $#content;
	   },
	   'KIND'=> sub {
	     my $index = shift;
			 my ($ret);
			 if ( $index =~ /^\d+$/ ) {
				 if($index >= 0 && $index <= $#content ) {
					 $ret =  $content[$index];
				 }
			 } else {
				 foreach my $item (@content){
					 if ( $item->{'NAME'}->() eq $index ) {
						 $ret = $item;
						 last;
					 }
				 }
			 }
			 return $ret;
	   },
	   'KINDSREF'=> sub {
			 my $name = shift;
			 my $ret = \@content;
			 if ( $name ) {
				 $ret = [];
				 foreach my $item (@content) {
					 if ( $item->{'NAME'}->() eq $name ) {
						 push @{$ret}, $item;
					 }
				 }
			 }
	     return $ret;
	   },
	   'ATTRIBUTESREF' => sub {
	     return \%attributes;
	   },
	   'PRINT'=>sub {
	     my $fh = shift || *STDOUT;
	     my $addPrefix = shift || '';
	     my $prefix = shift || '';
	     if($name){
	       my ($out,$item);
	       if(@content || defined($plain)){

					 $out = $prefix . &getTag($name,\%attributes,'start');
					 print $fh $out;
					 if(@content){
						 print $fh "\n";
						 foreach $item (@content){
							 if(ref($item)=~/HASH/ && ref($item->{'PRINT'})=~/CODE/){
								 $item->{'PRINT'}->($fh,$addPrefix,$prefix.$addPrefix);
							 }
						 }
					 }
					 if(defined($plain)){
						 print $fh $plain;
						 if(@content){
							 print $fh "\n$prefix";
						 }
					 }else{
						 if($prefix){
							 print $fh $prefix;
						 }
					 }
					 $out = &getTag($name,\%attributes,'stop') . "\n";
					 print $fh $out;

	       }else{
					 $out = $prefix . &getTag($name,\%attributes) . "\n";
					 print $fh $out;
	       }
	     }else{
	       if(defined($plain)){
					 print $fh $plain;
	       }
	     }
	   }
	 };

  return $this;
}

sub getTag {
  my ($name,$ptrAttr,$mode) = @_;
  my ($out,$key,$value);
  $mode||='';
  if ($mode eq 'stop'){
    $out="</$name>";
  }else{
    $out="<$name";
    while (($key,$value)=each(%{$ptrAttr})){
      $out .= ' '.$key;
      if (defined($value)){
		$out .= '="'.$value.'"';
      }
    }
    if($mode eq 'start'){
      $out .= ">";
    }else{
      $out .= "/>";
    }
  }
  return $out;
}

#
# end class for handle with XML
#


#
# class for handle with database
#
sub getDbConnect {
  my $ptrBin = shift;
  my $myDbType = shift;
  my $myDbAddress = shift ;
  my $myDbUser = shift ;
  my $myDbPasswd = shift ;
  my $myDbDB = shift ;
  my $myDbHost = shift ;

  #added by Honig for Sphera
  my $myDbSocket = shift;
  my $variables = shift;

  my ($dbh,$this,$sth);
  my ($shell,$args);
  my $echo = $ptrBin->{'echo'}||'echo';
  my ($sql,@result,$resultIndex,$ptrResult);

  unless(defined($myDbPasswd) && defined($myDbUser)){
    if($myDbType eq 'mysql'){
      foreach my $myCnf ('/etc/my.cnf',$ENV{'HOME'}.'/.my.cnf'){
		next unless (-T $myCnf);
		next unless open(CNF,"<$myCnf");
		binmode CNF;
		my($key,$value);
		my $state=0;
		while (<CNF>){
		  next unless $_;
		  next if (/^#/);
		  if($state){
			if(/^\s*\[/){
			  last;
			}
			if(/^\s*(\S+)\s*=\s*(\S+)/){
			  $key=$1;
			  $value=$2;
			  if($key =~ /^user$/){
				$myDbUser = $value;
				# sometimes strings are quoted
				$myDbUser =~ s/^('|")(.*)\1$/$2/;
			  }elsif($key =~ /^pass$/){
				$myDbPasswd = $value;
				# sometimes strings are quoted
				$myDbPasswd =~ s/^('|")(.*)\1$/$2/;
			  }
			}
		  }else{
			if(/^\s*\[client\]/){
			  $state=1;
			}
		  }
		}
		close (CNF);
      }
    }elsif($myDbType eq 'Pg'){
      $myDbPasswd ||= $ENV{'PGPASSWORD'};
      $myDbUser ||= $ENV{'PGUSER'};
    }
  }

  if (eval "require DBI"){
    if (eval "require DBD::$myDbType"){
      if($myDbAddress){
		if($myDbUser && $myDbPasswd){
		  $dbh = DBI->connect($myDbAddress,$myDbUser,$myDbPasswd) or
			die "Error: unable to connect to database: $DBI::errstr\n";
		}elsif($myDbUser){
		  $dbh = DBI->connect($myDbAddress,$myDbUser) or
			die "Error: unable to connect to database: $DBI::errstr\n";
		}else{
		  $dbh = DBI->connect($myDbAddress) or
			die "Error: unable to connect to database: $DBI::errstr\n";
		}
      }
    }
  }
#  unless($dbh){
#    if ($myDbType eq 'Pg'){
#      if(eval "require Pg"){
#        $myDbAddress="user=$myDbUser dbname=$myDbDB";
#        if($myDbPasswd){
#          $myDbAddress.=" password=$myDbPasswd";
#        }
#        $dbh = Pg::connectdb($myDbAddress);
#      }
#    }
# }
	unless ($dbh) {

		if ($myDbType eq 'mysql') {
			$shell = $ptrBin->{'mysql'} || 'mysql';
			$args  = " --socket='$myDbSocket'" if ($myDbSocket);
			if ($myDbPasswd) {
				$args .= " -B -u '$myDbUser' -p'$myDbPasswd' '$myDbDB' ";
			} else {
				$args .= " -B -u $myDbUser $myDbDB ";
			}
			$shell = "LD_LIBRARY_PATH=$variables $shell" if ($variables);
    }elsif($myDbType eq 'Pg'){
      $shell = $ptrBin->{'psql'}||'psql';
	  if($myDbPasswd) {
		$shell = "PGPASSWORD=$myDbPasswd $shell";
	  }
	  if($myDbUser) {
		$shell = "PGUSER=$myDbUser $shell";
	  }

      $args = " -d $myDbDB -c ";
    }
#
# check connection to Confixx databnase
#
    if($myDbDB=~/confixx/){
      $sql = "select hostname from admin";
      @result = `$echo "$sql"| $shell $args`;
      unless($result[0]=~/^\w+$/){
		$args = ' -h $myDbHost'.$args;
		@result = `$echo "$sql"| $shell $args`;
		unless($result[0]=~/^\s*hostname/){
		  die "Error: Can not connect to Confixx database (unable to execute sql: $sql)\n@result\n";
		}
      }
    }
#
# end check connection to Confixx databnase
#
    $shell .= $args;
  }
  $this = {
	   'RECONNECT'=>sub{
	     if($dbh){
	       $dbh->disconnect();
		   if($myDbPasswd){
			 $dbh = DBI->connect($myDbAddress,$myDbUser,$myDbPasswd) or
			   die "Error: unable to connect to database: $DBI::errstr\n";
		   }else{
			 $dbh = DBI->connect($myDbAddress,$myDbUser) or
			   die "Error: unable to connect to database: $DBI::errstr\n";
		   }
	       return $dbh;
	     }else{
	       return 1;
	     }
	   },
	   'EXECUTE'=> sub {
	     $sql = shift;
	     if($dbh){
	       $sth = &executeDbh($dbh,$sql);
	       if($sth){
		 return $sth->rows;
	       }else{
		 return 0;
	       }
	     }else{
	       $resultIndex=0;
	       &executeDbShell($shell,\@result,$echo,$sql,$myDbType);
	       return scalar(@result);
	     }

	   },
	   'FETCHROW'=> sub {
	     if($sth){
	       if(@result = $sth->fetchrow){
		 return \@result;
	       }else{
		 return undef;
	       }
	     }else{
	       $resultIndex++;
	       if($resultIndex>$#result){
		 return undef;
	       }else{
		 return $result[$resultIndex];
	       }
	     }
	   },
	   'FETCHHASH'=>sub {
	     if($sth){
	       $ptrResult = $sth->fetchrow_hashref();
	     }else{
	       $resultIndex++;
	       if($resultIndex>$#result){
		 $ptrResult = undef;
	       }else{
			 my ($i,$key);
			 $ptrResult = {};

			 for($i=0; $i<@{$result[0]}; $i++){
			   $ptrResult->{$result[0]->[$i]}=$result[$resultIndex]->[$i];
			 }
	       }
	     }
	     return $ptrResult;
	   },
	   'FINISH'=>sub {
	     if($sth){
	       $sth->finish();
	     }
	     @result=();
	     $ptrResult = undef;
	   },
	   'PRINTERROR'=>sub {
	     if($dbh){
	       if(@_){
			 $dbh->{'PrintError'}=shift;
	       }
	       return $dbh->{'PrintError'};
	     }else{
	       return 0;
	     }
	   },
	   'PING'=>sub{
	     if($dbh){
	       return $dbh->ping();
	     }else{
	       return 1;
	     }
	   },
	   'DISCONNECT'=>sub{
	     if($dbh){
	       return $dbh->disconnect();
	     }else{
	       return 1;
	     }
	   },
	   'USER'=>sub{
	     return $myDbUser;
	   },
	   'PASSWORD'=>sub{
	     return $myDbPasswd;
	   },
	   'HOST'=>sub{
	     return $myDbHost;
	   },
	   'TYPE'=>sub{
	     return $myDbType;
	   },
	  };

  return $this;
}

sub executeDbh {
  my $dbh = shift;
  unless(ref($dbh)=~/DBI/){
    &printToError ("Error: DBI object is not found");
  }
  my $sql = shift;

  my $sth = $dbh->prepare($sql) or
    &printToError ("Error: prepare sql: $sql\n$DBI::errstr");
  if($sth->execute){
    return $sth;
  }else{
    &printToError ("Error: exesute sql: $sql\n$DBI::errstr");
  }
}

sub executeDbShell {
  my ($shell,$ptrRes,$echo,$sql, $type) = @_;
  @{$ptrRes}=();
  my (@out,@row,$ret);
  $ret=0;
  if($type eq 'Pg'){
    @out = `$shell "$sql"`;
  }else{
    @out = `$echo "$sql"|$shell`;
  }
  if ($out[0]=~/^error/i){
    &printToErrror ("Error: execute sql: $sql\n@out");
  }else{
    if($type eq 'Pg'){
      foreach my $line (@out){
	chomp $line;
	next if ($line=~/^-[-+]*$/); # skip delimiter
	if($line=~/^\(\s*\d+\s+\w+\)$/){
	    last;  ## (6 rows)
	}
	$line=~s/^\s+//;
	$line=~s/\s+$//;
	@row = split(/\s*\|\s*/,$line);
	push @{$ptrRes}, [@row];
	$ret++;
      }
    }else{
      if($out[0]=~/^\+[-+]*$/){
	foreach my $line (@out){
	  chomp $line;
	  next if ($line=~/^\+[-+]*$/); # skip delimiter
	  $line=~s/^\|\s*//;
	  $line=~s/\s+$//;
	  @row = split(/\s*\|\s*/,$line);
	  @row = map { $_ eq 'NULL' ? undef : $_ } @row;
	  push @{$ptrRes}, [@row];
	  $ret++;
	}
      }else{  ## batch mode
	foreach my $line (@out){
	  chomp $line;
	  @row = split(/\t/,$line);
	  @row = map { $_ eq 'NULL' ? undef : $_ } @row;
	  push @{$ptrRes}, [@row];
	  $ret++;
	}
      }
    }
  }
  return $ret;
}

#
# end class for handle with database
#

sub makeMIMEBase64 {
  my ($useModule,$this);
  if (eval "require MIME::Base64"){
    $useModule = 1;
  }else{
    $useModule = 0;
  }

  $this={
	 'ENCODE'=>sub{
	   my $src = shift;
	   my ($encoded);
	   if($useModule){
	     $encoded = MIME::Base64::encode($src);
		 chomp $encoded;
	   }else{
	     $encoded = &encodeBase64($src);
	   }
	   return $encoded;
	 },
	 'USEMODULE'=>sub{
	   return $useModule;
	 },
	};

  return $this;
}

#
# url decode (certificates in Plesk DB encoded by the url encoding)
#

sub urlDecode {
    my $url = $_[0];
    $url =~ tr/+/ /;
    $url =~ s/%([a-fA-F0-9]{2,2})/chr(hex($1))/eg;
    return $url;
}

sub encodeHtml {
  my(@ret,$src);
  if ($src=shift){
    if(ref $src){
      if(${$src}=~/[&"'><]/){
		${$src}=~s/&/&amp;/sg;
		${$src}=~s/>/&gt;/sg;
		${$src}=~s/</&lt;/sg;
		${$src}=~s/\'/&#39;/sg;
		${$src}=~s/\"/&quot;/sg;
      }
      return ${$src};
    }else{
      if($src=~/[&"'><]/){
		$src=~s/&/&amp;/sg;
		$src=~s/>/&gt;/sg;
		$src=~s/</&lt;/sg;
		$src=~s/\'/&#39;/sg;
		$src=~s/\"/&quot;/sg;
      }
    }
  }
  return $src;
}

# Normalization according to the XML 1.0 (Third Edition),
# clauses 2.3 "Common Syntactic Constructs"
# and 3.3.3 "Attribute-value normalization"

sub normalizeAttribute {
  $_ = $_[0];
  s/\r/&#xD;/g;
  s/\n/&#xA;/g;
  s/\t/&#x9;/g;
  s/\</&lt;/g;
  s/\&/&amp;/g;
  s/\"/&quot;/g;
  s/\'/&#39;/g;
  return $_;
}

############################
#
#	Non-ascii characters handling
{
	# static
	my $charset;

	sub setCharset {
	  $charset = shift;
	}

	# try to determine the local charmap
	sub _determineCharset()
	{
	  my $charset;
	  my $locale = $bin{'locale'} || 'locale';

	  if (exists $ENV{'LANG'}) {
		my @pieces = split /\./, $ENV{'LANG'};
		if (defined $pieces[1]) {
		  $charset = $pieces[1];
		  chomp $charset;
		}
	  }
	  if (not $charset && open(CM, "$locale charmap |")) {
	  	binmode CM;
		$charset = <CM>;
		close(CM);
		chomp $charset;
	  }
	  if (not $charset) {
		$charset = 'ISO-8859-1';
	  }

	  return $charset;
	}

	sub toUtf8
	{
		my ($subj, $myCharset) = @_;

		$charset = _determineCharset() if (not $charset);

		$myCharset = $charset unless ( $myCharset );

		if ( (not $subj) ||
				 (not $bin{'iconv'}) ||
				 ( $myCharset =~ /UTF-?8/i )
			 ) {
			return $subj;
		}

		if ($subj =~ /^[\x00-\x7f]*$/) { return $subj; }

		my $converted;
		my $escaped = $subj;
		$escaped =~ s/'/'"'"'/g;

		if (open(IC, "echo '$escaped' | " . $bin{'iconv'} . " -f '$myCharset' -t 'UTF-8' |")) {
			binmode IC;
			while (my $string = <IC>) { $converted .= $string; }
			close(IC);
			chomp $converted;
			if ($converted) {
			  return $converted;
			}
		}

		return "";
	}
}

#
# Based on the b64.c by Bob Trower (base64.sf.net)
#

my @cb64 = ('A', 'B', 'C', 'D', 'E', 'F', 'G', 'H',
			'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P',
			'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X',
			'Y', 'Z', 'a', 'b', 'c', 'd', 'e', 'f',
			'g', 'h', 'i', 'j',	'k', 'l', 'm', 'n',
			'o', 'p', 'q', 'r', 's', 't', 'u', 'v',
			'w', 'x', 'y', 'z', '0', '1', '2', '3',
			'4', '5', '6', '7', '8', '9', '+', '/');

@cb64 = map { ord } @cb64;

sub encodeBlock {
  my ($len, @in) = @_;
  my @res;
  push @res, $cb64[$in[0] >> 2];
  push @res, $cb64[(($in[0] & 0x3)<<4)|(($in[1]&0xf0)>>4)];
  push @res, $len>1 ? $cb64[(($in[1]&0xf)<<2)|(($in[2]&0xc0)>>6)] : ord('=');
  push @res, $len>2 ? $cb64[$in[2]&0x3f] : ord('=');
  return @res;
}

sub encodeBase64 {
  my @in = unpack "C*", $_[0];

  my @out;
  while (scalar(@in)) {
	my @block = (0, 0, 0);
	my $length = 0;
	while (scalar(@in) && $length < 3) {
	  $block[$length++] = shift(@in);
	}
	push @out, encodeBlock($length, @block);
  }
  return pack "C*", @out;
}

sub getBins {
  my ($ptrBin,$path) = @_;
  my (%searchPath,$key,$prefix,$bin,$fullPath);
  my ($notFound,$i,@prefixes);

  $notFound=0;

  $i=0;
  foreach $prefix (split(/:/,$path)){
    next unless $prefix;
    unless(exists($searchPath{$prefix})){
      $searchPath{$prefix}=$i++;
      push @prefixes,$prefix;
    }
  }

  foreach $key (keys %{$ptrBin}){
	$ptrBin->{$key} = undef;
    foreach $prefix (@prefixes){
      $prefix.='/' unless $prefix=~/\/$/;
      $bin = $key;
      $bin =~ s/^\///;
      $fullPath = $prefix.$bin;
      if (-x $fullPath){
		$ptrBin->{$key}=$fullPath;
		last;
      }
    }
    $notFound++;
  }
  return $notFound;
}

sub getArguments {
  my ($ptrArgv,$ptrOpts) = @_;

  my (@keys,$firstKey,%prepKeys,$key,$value,$ptrArr,$arg,$state);
  my (%retHash,$pat,$found,$used,@rest,$fullArg,$prevKey);

  while (($key,$value)=each(%{$ptrOpts})){
    @keys = split(/\|/,$key);
    $firstKey = $keys[0];
    $firstKey =~s/^-*//;
    push @{$prepKeys{$firstKey}},[@keys];
    push @{$prepKeys{$firstKey}},$value;
  }

  $state =0;
  foreach $arg (@{$ptrArgv}){

#    print "arg: $arg\n";

    $used = 0;
    if($state==1){
      if  ($arg =~ /^-/){
	$state=0;
      }else{
	$used = 1;
	$retHash{$prevKey}=$arg;
	next;
      }
    }
    if ($state == 2){
      $retHash{$prevKey}=$arg;
      $state = 0;
      $used = 1;
    }else{
      $fullArg = $arg;
      if ($arg =~ /^(-\S+)=(\S+)/){
	$arg = $1;
	$value = $2;
      }else{
	$value = undef;
      }
      foreach $key (keys %prepKeys){

	$ptrArr = $prepKeys{$key};

#	print "key: $key, ",join('|',@{$ptrArr->[0]}),"\n";


	$found = 0;
	foreach $pat (@{$ptrArr->[0]}){
	  if ($pat eq $arg){
	    $found = 1;
	    last;
	  }
	}
	if($found){
	  $used = 1;
	  if(defined($value)){
	    $retHash{$key}=$value;
	  }else{
	    if($ptrArr->[1]){
	      if($ptrArr->[1] =~ /\?$/){
		$state = 1;
	      }else{
		$state = 2;
	      }
	      $prevKey = $key;
	    }else{
	      $retHash{$key}=undef;
	    }
	  }
	  last;
	}
      }
    }
    unless($used){
      if($value){
	push @rest,$fullArg;
      }else{
	push @rest,$arg;
      }
    }
  }
  @{$ptrArgv}=@rest;
  return \%retHash;

}

sub makeDumpStatus($) {
  my $statusFile = shift;
  my($pid,$currAccount,$currDomain,$leftAccounts,$leftDomains,$this);
  $pid=$$;
  $this = {'FILE'=>sub {
	     if(@_){
	       $statusFile = shift;
	     }
	     return $statusFile;
	   },
	   'PID'=>sub {
	     if(@_){
	       $pid = shift;
	     }
	     return $pid;
	   },
	   'ACCOUNT'=>sub{
	     if(@_){
	       $currAccount = shift;
	       if($leftAccounts){
		 $leftAccounts--;
	       }
	     }
	     return $currAccount;
	   },
	   'DOMAIN'=>sub{
	     if(@_){
	       $currDomain = shift;
	       if($leftDomains){
		 $leftDomains--;
	       }
	     }
	     return $currDomain;
	   },
	   'COUNTS'=>sub{
	     if(@_){
	       my $newValue = shift;
	       if (defined $newValue){
		 $leftAccounts=$newValue;
	       }
	       $newValue = shift;
	       if (defined $newValue){
		 $leftDomains=$newValue;
	       }
	     }
	     return ($leftAccounts,$leftDomains);
	   },
	   'PRINT'=>sub{
	     if(@_){
	       $statusFile=shift;
	     }
	     open(STATUS,">$statusFile.tmp");

		 binmode STATUS;
	     my $root = &makeXmlNode('agent-dump-status');
	     my $item = &makeXmlNode('pid',$pid);
	     $root->{'ADDCHILD'}->($item);
	     $item = &makeXmlNode('account',$currAccount||'');
	     $root->{'ADDCHILD'}->($item);
	     $item = &makeXmlNode('domain',$currDomain||'');
	     $root->{'ADDCHILD'}->($item);
	     $item = &makeXmlNode('accounts-left',
				  defined($leftAccounts)?$leftAccounts:'');
	     $root->{'ADDCHILD'}->($item);
	     $item = &makeXmlNode('domains-left',
				  defined($leftDomains)?$leftDomains:'');
	     $root->{'ADDCHILD'}->($item);

	     &printXml($root,*STATUS,'agent-dump-status.dtd');

	     close(STATUS);

		 rename("$statusFile.tmp", $statusFile);
	   }

	  };
  return $this;
}

sub makeContentList {
  my ($dirDump,$xmlDump) = @_;
  my ($name,$fullPath,%files,%fromXML,$size,@cids,$root,$item,$key,$cmd,
      $gzip,$wc,$md5sum);

  if(opendir(DUMP,$dirDump)){
    while ($name=readdir(DUMP)){
      $fullPath="$dirDump/$name";
      unless(-f $fullPath && $name=~/\.\w\w$/){
	next;
      }
      $key = $fullPath;
      $key =~ s/\.\w\w$//;
      $size=(stat($fullPath))[7];
      if(defined $size){
	push @{$files{$key}},[$fullPath,$size];
      }
    }
    closedir(DUMP);
  }

	if (-T $xmlDump) {
		if (open(DUMP, "<$xmlDump")) {
			binmode DUMP;
			while (<DUMP>) {
				if (@cids = /cid(?:_\w+)?=\"([^"]+)\"/g) {
					foreach $fullPath (@cids) {
						unless ($fullPath =~ /^\//) {
							$fullPath = "$dirDump/$fullPath";
						}
						$fromXML{$fullPath} = undef;
					}
				}
			}
			close(DUMP);
		}

		foreach $key (keys %files) {
			unless (exists($fromXML{$key})) {
				delete $files{$key};
			}
		}
	}

  $gzip = $bin{'gzip'}||'gzip';
  $wc = $bin{'wc'}||'wc';
  $md5sum = $bin{'md5sum'}||'md5sum';

  $root=&makeXmlNode('contentlist');

  my($compressed,$uncompressed,$list,$count,$ptrList,@sortedList,
     $ptrItem,$out);
  my($totalUncomp,$totalSize)=(0,0);
  foreach $key (keys %files){

#
# sort by extension  .aa .ab ...
#
    @sortedList =  sort {&getExt($a->[0]) cmp &getExt($b->[0])} @{$files{$key}};

#
# output list of files
#
    $compressed=0;
    foreach $ptrItem (@sortedList){
      ($fullPath,$size)=@{$ptrItem};
      $item=&makeXmlNode('file',$fullPath);
      $item->{'ATTRIBUTE'}->('size',$size,'noencode');
      $root->{'ADDCHILD'}->($item);

#
# get md5sum
#
      if(-x $md5sum){
	$out=`$md5sum $fullPath`;
	if($out=~/\s*([0-9a-fA-F]+)\s+/){
	  $item->{'ATTRIBUTE'}->('md5sum',$1,'noencode');
	}
      }

      $compressed += $size;

    }

    $uncompressed = undef;
#
# get unpacked size of the archive
#
    if ($key=~/\.tgz$|\.gz$/){
      if(-x $gzip){
	if(scalar(@sortedList)==1){
	  ($fullPath,$size)=@{$sortedList[0]};
	  ($uncompressed,$compressed)=&getZipSize($fullPath,$gzip);
	}else{
	  $list = join(' ',map {$_->[0]} @sortedList);
	  if(-x $wc){
	    $cmd = "cat $list | $gzip | $wc -c";
	    $uncompressed = `$cmd`;
	    chomp $uncompressed;
	    if ($uncompressed =~ /^\s*(\d+)\s*$/){
	      $uncompressed = $1;
	    }else{
	      $uncompressed = undef;
	    }
	  }else{
	    $cmd = "cat $list >$key";
	    if(system($cmd)==0){
	      ($uncompressed,$compressed)=&getZipSize($key,$gzip);
	      unlink $key;
	    }
	  }
	}
      }
    }
#
# calc total size
#
    $totalSize += $compressed;
    $totalUncomp += $uncompressed if $uncompressed;
  }

  $root->{'ATTRIBUTE'}->('size',$totalSize,'noencode');
  if ($totalUncomp && $totalUncomp != $totalSize){
    $root->{'ATTRIBUTE'}->('unpacked',$totalUncomp,'noencode');
  }

  return $root;
}

sub getExt {
  my $path = shift;
  if($path=~/\.([^.]*)$/){
    return $1;
  }
  return $path;
}

sub getZipSize {
  my ($fullPath,$gzip) = @_;
  $gzip ||= 'gzip';
  my $cmd = "$gzip -l $fullPath";
  my ($uncompressed,$compressed,@out);
  if (open(OUTPUT,"$cmd|")){
    ($compressed,$uncompressed)=(0,0);
    while(<OUTPUT>){
      @out = split(' ',$_);
      if($out[0]=~/^\d+$/ && $out[1]=~/^\d+$/){
	$uncompressed+=$out[1];
	$compressed+=$out[0];
      }
    }
    close (OUTPUT);
  }
  return $uncompressed,$compressed;
}

my $outputIsFile;

sub openOutput
{
	my ($outPath) = @_;
	my ($outFh);

	if ($outPath) {
		open(OUTPUT, ">>$outPath") or die "Error: can not open '$outPath' for write: $!\n";
		binmode OUTPUT;
		$outFh = *OUTPUT;
		$outputIsFile = 1;
	} else {
		$outFh = *STDOUT;
		$outputIsFile = 0;
	}

	return $outFh;
}

sub closeOutput{
  if($outputIsFile){
    close (OUTPUT) or
      &printToError("Error: unable to close OUTPUT: $!\n");
    $outputIsFile=0;
  }
}

#=========================================================================
#
# dtd-specific (but common) subroutines
#
sub createPasswordNode {
  my ($parent, $password, $type) = @_;

  my $item = makeXmlNode('password', $password);
  $item->{'ATTRIBUTE'}->('type', $type);
  $parent->{'ADDCHILD'}->($item);
}

sub createSysuserNode {
  my ($parent, $name, $password, $type) = @_;

  my $item = makeXmlNode('sysuser');
  $item->{'ATTRIBUTE'}->('name', $name);
  createPasswordNode($item, $password, $type);
  $parent->{'ADDCHILD'}->($item);
}

sub readFile {
  my ($fileName) = @_;

  unless (-f $fileName) {
	return undef;
  }

  open(INPUT, "< $fileName");
  binmode INPUT;
  my $content = join('', <INPUT>);
  close(INPUT);

  return $content;
}

1;

#
# end common subroutines
#
#=========================================================================


$ptrArgs = &getArguments (\@ARGV,\%arg_opts);

#
# search binaries
#

&getBins(\%bin,'/bin:/usr/bin:/usr/local/bin:/sbin:/usr/sbin:/usr/local/sbin:'.$ENV{'PATH'});

my ($outPath,$outFh,$outHas,$dumpFile,$statusFile,$app);

$app = $0;
if ($app=~/(.+)\/([^\/]+)$/) {
  $workDir = $1;
  $app = $2;
}
$app=~s/\.pl$//;

$pwd=`$bin{'pwd'}`||'.';

chop $pwd;

if ($workDir eq '.') {
  $workDir = $pwd;
} elsif ($workDir !~ /^\//) {
  $workDir = $pwd.'/'.$workDir;
}

$dumpFile = $workDir.'/cPanel9.xml';
$statusFile = $workDir.'/cPanel9.status';
$dumpDir = $workDir.'/pma';

my $objDumpStatus = &makeDumpStatus($ptrArgs->{'status-file'}||$statusFile);


if (!$ptrArgs->{'output'} &&
	(exists $ptrArgs->{'dump-all'} ||
	 exists $ptrArgs->{'dump-accounts'}||
	 exists $ptrArgs->{'dump-domains'}
	)
   ) {
  $ptrArgs->{'output'}=$dumpFile;
}

if ($outPath = $ptrArgs->{'output'}) {

  open(OUTPUT, ">$outPath") or die "Error: can not open '$outPath' for write: $!\n";
  binmode OUTPUT;
  $dumpFile = $outPath;
  close(OUTPUT);
}

if (exists $ptrArgs->{'version'}) {
  print $outFh $VERSION,"\n";
  exit 0;
}

my (%cPanelConfig);
my $rootName = 'cPanel9dump';
my $dtdName = 'cPanel9.dtd';

unless ($ptrArgs->{'config'} = &loadMainConfig($ptrArgs->{'config'},\%cPanelConfig)) {
  if (exists $ptrArgs->{'get-status'}) {

	my $outFh = &openOutput($ptrArgs->{'output'});

	&printAgentStatus($outPath=$ptrArgs->{'output'}||$dumpFile,
					  $objDumpStatus->{'FILE'}->(),
					  $outFh);

	&closeOutput();

	exit 0;
  } else {

	if (exists ($ptrArgs->{'dump-all'}) ||
		exists ($ptrArgs->{'dump-accounts'}) ||
		exists ($ptrArgs->{'dump-domains'})) {

	  my $root=&makeXmlNode($rootName);

	  if (ref($root)=~/HASH/ && ref($root->{'PRINT'})=~/CODE/) {
		my $outFh = &openOutput($ptrArgs->{'output'});
		&printXml($root,$outFh,$dtdName);
		&closeOutput();
	  }
	}

	die "Error: cPanel config file 'cpanel.conf' is not found\n";
  }
}
my $installDir = $cPanelConfig{'installDir'}||'/var/cpanel';

my $userMysqlDbAddress = "DBI:mysql:mysql:$mysqlUserServer";
my $wrapUserMysql;

#
# get MIME Base64 encoder
#
my $wrapBase64 = makeMIMEBase64();



if (exists $ptrArgs->{'get-status'}) {

  my $outFh = &openOutput($ptrArgs->{'output'});

  &printAgentStatus($outPath=$ptrArgs->{'output'}||$dumpFile,
					$objDumpStatus->{'FILE'}->(),
					$outFh);
  &closeOutput();

} elsif (exists ($ptrArgs->{'dump-all'}) ||
		 exists ($ptrArgs->{'dump-accounts'}) ||
		 exists ($ptrArgs->{'dump-domains'})) {

  my ($root,@accounts,@domains,$ptrAccounts,$ptrDomains, $value);
  if (exists $ptrArgs->{'no-content'}) {
	$::dumpDir = &getDumpDir("$workDir/pma");
	$noContent = 1;
  } else {
	$::dumpDir = &getDumpDir("$workDir/pma",'clean');
  }

  if (exists $ptrArgs->{'no-compress'}) {
	setCompress(0);
  }

  &printToLog("Work dir: $workDir");
  &printToLog("Dump file: $dumpFile");
  &printToLog("Status file: ".$objDumpStatus->{'FILE'}->());


  if ($value = $ptrArgs->{'dump-accounts'}) {
	if ($value eq "-") {
	  $value = <STDIN>;
	  chomp $value;
	}
	@accounts = split(/\s*,\s*/,$value);
	$ptrAccounts = \@accounts;
  }

  if ($value = $ptrArgs->{'dump-domains'}) {
	if ($value eq "-") {
	  $value = <STDIN>;
	  chomp $value;
	}
	@domains = split(/\s*,\s*/,$value);
	$ptrDomains = \@domains;
  }
  #
  # generate a xml dump
  #

  $root = &getcPanel9Dump($dumpDir, $ptrAccounts, $ptrDomains, $rootName);

  #
  # print dump to output
  #
  &printToLog('Print dump file ...',1);
  if (ref($root)=~/HASH/ && ref($root->{'PRINT'})=~/CODE/) {
	my $outFh = &openOutput($ptrArgs->{'output'});

	&printXml($root,$outFh,$dtdName);

	&closeOutput();
  }
  &printToLog(' OK');

} elsif (exists $ptrArgs->{'get-content-list'}) {

  $::dumpDir = &getDumpDir("$workDir/pma");

  my $root = &makeContentList($::dumpDir,$dumpFile);

  if (ref($root)=~/HASH/) {
	my $outFh = &openOutput($ptrArgs->{'output'});

	&printXml($root,$outFh);

	&closeOutput();
  }

} else {
  &printHelp();
}

exit $returnCode;

#######################################################
#
#  Get all client accounts, registered in the system
#
#######################################################
sub getAllAccounts {
  my @all_accounts;

  opendir(USRS, "$installDir/users");
  my @accts = readdir(USRS);
  closedir(USRS);
  chomp(@accts);

  # remove duplicate and possible wrong entries
  foreach my $account (@accts) {
	next if ($account =~ /\./ || grep(/^$account$/, @all_accounts));
	push @all_accounts, $account;
  }

  return @all_accounts;
}

##########################################
#
#   $ptrAccounts - accounts, that have been specified as a --dump-accounts
#   $ptrDomains  - domains, that have been specified as a --dump-domains
#
##########################################
sub getcPanel9Dump {

  my ($dumpDir, $ptrAccounts, $ptrDomains, $rootName) = @_;

  my $root = &makeXmlNode($rootName);

  my($reseller, $fileResellers, $ptrRows, $owner, $item, $account,
	 $ptrHash, $key, $value);

  my @all_accounts = getAllAccounts();

  # No accounts - nothing to dump
  return if (-1 == $#all_accounts);

  if ( -f "$installDir/resellers") {
	# Find all the resellers in the system
	$fileResellers = &makeFileParser("$installDir/resellers",
									 'KEYSEPARATOR' => ':',
									 'VALUESEPARATOR' => ',',
									 'KEYVALUESEPARATOR' => ' ');

	# All the resellers in the system (not including the default reseller)
	$ptrRowResellers = $fileResellers->{'PARSE'}->();
  }

  # Associative array of domain owners => {0 if only this domain should
  # be dumped and 1 if all domains should be dumped, domain}
  my %accounts;


  # If domains were specified through command line search their owning accounts
  if (ref($ptrDomains) =~ /ARRAY/) {

	foreach my $domain (@{$ptrDomains}) {
	  if ($owner = &getDomainOwner($domain)) {
		if (exists($accounts{$owner})) {
		  push @{$accounts{$owner}->[1]}, $domain;
		} else {
		  $accounts{$owner} = [0, [$domain]];
		}
	  }
	}
  }

  # If accounts were specified through command line - mark them for dumping
  if (ref($ptrAccounts) =~ /ARRAY/) {
	foreach my $account (@{$ptrAccounts}) {
	  $accounts{$account} = [1, undef];
	}
  }


  # Check that all found resellers are exists
  foreach my $reseller (keys %{$ptrRowResellers}) {
	unless(-T "$installDir/users/$reseller") {
	  delete $ptrRowResellers->{$reseller};
	  next;
	}
  }

  # If no accounts/domains were specified through command line $checkAccounts is 0
  my $checkAccounts = scalar(%accounts);


  # For each reseller in the system
  foreach $account (@all_accounts) {

	my $fileUser = makeFileParser("$installDir/users/$account");

	$ptrRows = $fileUser->{'PARSE'}->('KEYSEPARATOR' => '=');

	# If file is empty or wrong format
	unless (ref($ptrRows) =~ /HASH/) {
	  delete $ptrRowResellers->{$account};
	  next;
	}

	$owner = $ptrRows->{'OWNER'};

	# If account belongs to first-level reseller
	if ($owner && ($owner ne $account) &&
		($owner ne $DEFAULT_RESELLER) &&
		(exists $ptrRowResellers->{$owner})) {
	  push @{$resellers{$owner}}, $account; # customer
	  # If account was specified through the command line
	  if ($checkAccounts &&
		  exists ($accounts{$account}) &&
		  !exists ($accounts{$owner})
		 ) {					## account
		$accounts{$owner} = [0, undef];	# account drag reseller
	  }
	} elsif ($owner eq $account || $owner eq $DEFAULT_RESELLER) {
	  # If reseller belongs to default reseller

	  unless (exists $resellers{$account}) {
		$resellers{$account} = [];
	  }
	} else {
	  if (!$checkAccounts || exists($accounts{$account})) {
		push @default_reseller_account, $account;
	  }
	}
  }


  #
  # prepare --dump-accouns
  #
  if ($checkAccounts) {

	my (@accRes);
	foreach $reseller (keys %resellers) {
	  unless (exists($accounts{$reseller})) {
		delete $resellers{$reseller};
		next;
	  }

	  unless ($accounts{$reseller}->[0]) { ## not full
		@accRes = ();
		foreach $account (@{$resellers{$reseller}}) {
		  if (exists($accounts{$account})) {
			push @accRes,$account;
		  }
		}
		$resellers{$reseller}=[@accRes];
	  }
	}
  }

  my $fileUser = makeFileParser("/etc/quota.conf");
  # Get disk quotas
  $ptrQuota = $fileUser->{'PARSE'}->('KEYSEPARATOR' => '=',
									 'VALUESEPARATOR' => '');

  unless (ref($ptrQuota) =~ /HASH/) {
	$ptrQuota = {};
  }

  #
  # prepare reseller's info
  #
  $ptrResellerInfo={};

  if ( -f "$installDir/reseller-limits-bynumber") {
	$fileUser = makeFileParser("$installDir/reseller-limits-bynumber");
	$ptrHash = $fileUser->{'PARSE'}->('KEYSEPARATOR' => '=', 'VALUESEPARATOR' => '');

	if (ref($ptrHash) =~ /HASH/) {
	  while (($reseller, $value) = each(%{$ptrHash})) {
		$ptrResellerInfo->{$reseller} = {'limits-bynumber' => $value};
	  }
	}
  }


  if ( -f "$installDir/reseller-resource-limits") {
	$fileUser = makeFileParser("$installDir/reseller-resource-limits");
	$ptrHash = $fileUser->{'PARSE'}->('KEYSEPARATOR' => ' ', 'VALUESEPARATOR' => '');

	if (ref($ptrHash) =~ /HASH/) {
	  my ($tail, $keyLimit, $valLimit, $ptrLimits);

	  while (($reseller, $value) = each(%{$ptrHash})) {
		($key, $tail) = split(' ', $value, 2);
		$ptrLimits = {};

		foreach my $chunk (split(/\s*:\s*/, $tail)) {
		  if ($chunk =~ /(\S+)\s*=\s*(\S+)/) {
			$ptrLimits->{$1} = $2;
		  }
		}

		if (exists $ptrResellerInfo->{$reseller}) {
		  $ptrResellerInfo->{$reseller}->{'resource-limits'} = [$key, $ptrLimits];
		} else {
		  $ptrResellerInfo->{$reseller}={'resource-limits' => [$key, $ptrLimits]};
		}
	  }
	}
  }

  #
  # end prepare reseller's info
  #

  #
  # parse httpd.conf
  #
  if (open (INPUT,"</etc/httpd/conf/httpd.conf")) {
	binmode INPUT;
	my ($ip,$state,@vDomains,$vDomain);
	$state=0;
	while (<INPUT>) {
	  chomp;
	  next unless $_;
	  next if (/^#/);
	  if ($state==1) {
		if (/<\/VirtualHost>/) {
		  foreach $vDomain (@vDomains) {
			$vDomain=~s/^www\.//;
			$servers_ips{$vDomain}=$ip;
		  }
		  @vDomains=();
		  $state=0;
		} elsif (/^\s*Server(?:Name|Alias)\s+(.*)$/) {
		  push @vDomains,split(' ',$1);
		}
	  } else {
		if (/<VirtualHost\s(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})>/) {
		  $ip=$1;
		  @vDomains=();
		  $state=1;
		}
	  }
	}
	close(INPUT);
  } else {
	&printToError("Error: unamle to open '/etc/httpd/conf/httpd.conf'");
  }
  #
  # end parse httpd.conf
  #

  ##
  # Determine number of objects to dump
  ##
  my $clientCount = scalar(keys %resellers)+1;
  my $domainCount = 0;
  my %domainCounts;

  foreach $reseller (keys %resellers) {
	  $domainCounts{$reseller}=0;
  }

  unless(keys %trueUserDomains){
	  my $fileUser = makeFileParser("/etc/trueuserdomains");
	  $ptrHash = $fileUser->{'PARSE'}->('KEYSEPARATOR'=>':');
	  if (ref($ptrHash)=~/HASH/) {
		  %trueUserDomains=(%{$ptrHash});
	  }
  }
  my ($domain);
  while ( ($domain,$owner) = each %trueUserDomains) {
	  if (exists($domainCounts{$owner})) {
		  $domainCounts{$owner}++;
		  $domainCount++;
	  }
  }

  if (ref($ptrDomains) =~ /ARRAY/) {
	  $domainCount = $#{$ptrDomains};
  }

  $objDumpStatus->{'COUNTS'}->($clientCount,$domainCount);

  foreach $reseller (sort keys %resellers) {

	&printToLog("Client '$reseller' is started");

	$objDumpStatus->{'ACCOUNT'}->($reseller);
	$objDumpStatus->{'PRINT'}->();

	$item = &makeResellerNode($reseller,\@{$resellers{$reseller}});

	if (ref($item)=~/HASH/) {
	  $root->{'ADDCHILD'}->($item);
	  &printToLog("Client '$reseller' is successfully dumped");
	} else {
	  &printToLog("Client '$reseller' is not dumped");
	}

	$domainCount-=$domainCounts{$reseller};
	foreach $owner (@{$resellers{$reseller}}) {
	  $domainCount-=$domainCounts{$owner};
	}

	$objDumpStatus->{'COUNTS'}->(undef,$domainCount);

  }

  &printToLog("Default client is started");

  $objDumpStatus->{'ACCOUNT'}->('[default]');
  $objDumpStatus->{'PRINT'}->();

  if (@default_reseller_account) {

	$item = &makeResellerNode('',\@default_reseller_account);

	if (ref($item)=~/HASH/) {
	  $root->{'ADDCHILD'}->($item);
	  &printToLog("Default client is successfully dumped");
	} else {
	  &printToLog("Default client is not dumped");
	}
  }


  $objDumpStatus->{'ACCOUNT'}->('');
  $objDumpStatus->{'DOMAIN'}->('');
  $objDumpStatus->{'COUNTS'}->(0,0);
  $objDumpStatus->{'PRINT'}->();

  return $root;
}

#
# makeResellerNode
#
#   arguments:
#       $reseller - name of rteseller
#       $ptrAccounts - pointer to array of accounts
#
sub makeResellerNode {
  my ($reseller,$ptrAccounts) = @_;

  my ($account,$item,$ptrUserHash,$fileParser);

  my $root = &makeXmlNode('reseller');
  $root->{'ATTRIBUTE'}->('name',$reseller);


  if ($reseller) {

	$fileParser = &makeFileParser("$installDir/users/$reseller");
	$ptrUserHash = $fileParser->{'PARSE'}->('KEYSEPARATOR'=>'=');

	my($ptrHash,$value,$key,$xmlKey,$userKey);
	$ptrHash = $ptrResellerInfo->{$reseller};

	if (ref($ptrHash)=~/HASH/) {
	  #
	  # limits
	  #
	  $value = $ptrHash->{'limits-bynumber'};
	  if ($value) {
		$root->{'ADDCHILD'}->(&makeLimitNode('max_accounts',$value));
	  }
	  if (exists $ptrHash->{$ptrHash->{'resource-limits'}}) {
		($key,$ptrHash)=@{$ptrHash->{'resource-limits'}};
	  }

	  if ($key) {
		$value = $ptrHash->{'bw'};
		if ($value=~/(\d+)/) {
		  $value = $1 * 1024*1024;
		  $root->{'ADDCHILD'}->(&makeLimitNode('max_traffic',$value));
		}
		$value = $ptrHash->{'disk'};
		if ($value=~/(\d+)/) {
		  $value = $1 * 1024*1024;
		  $root->{'ADDCHILD'}->(&makeLimitNode('disk_space',$value));
		}
	  }
	}

	while (($xmlKey,$userKey)=each(%mapLimits)) {
	  $value = $ptrUserHash->{$userKey};
	  if ($value=~/\d+/) {
		$root->{'ADDCHILD'}->(&makeLimitNode($xmlKey,$value));
	  } else {
		$root->{'ADDCHILD'}->(&makeLimitNode($xmlKey));
	  }
	}
	#
	# end limits
	#

	#
	# permissions
	#
	$ptrHash = $ptrRowResellers->{$reseller};
	if (ref($ptrHash)=~/HASH/) {
	  if (exists $ptrHash->{'create-acct'}) {
		$root->{'ADDCHILD'}->(&makePermissionNode('create_accounts'));
	  }
	  if (exists $ptrHash->{'create-dns'}) {
		$root->{'ADDCHILD'}->(&makePermissionNode('create_domains'));
	  }
	  if (exists $ptrHash->{'edit-dns'}) {
		$root->{'ADDCHILD'}->(&makePermissionNode('manage_dns'));
	  }
	}
	#
	# end permissions
	#

	$item = &makeAccountNode($reseller, $ptrUserHash, 'is reseller');
	$root->{'ADDCHILD'}->($item);

  }

  foreach $account (sort @{$ptrAccounts}) {
	$item = &makeAccountNode($account);
	$root->{'ADDCHILD'}->($item);
  }

  return $root;
}
sub getDomainName {
  my $account = shift;

  unless(keys %trueUserDomains){
	#
	# calculate $ptrTrueUserDomainsHash
	#
	my $fileParser = &makeFileParser('/etc/trueuserdomains');
	my $ptrHash = $fileParser->{'PARSE'}->('KEYSEPARATOR'=>':');
	if (ref($ptrHash)=~/HASH/) {
	  %trueUserDomains=(%{$ptrHash});
	}
  }

  my ($domain,$owner);
  while (($domain,$owner) = each %trueUserDomains ) {
	if ($owner eq $account) {
	  return $domain;
	}
  }
  return undef;
}

sub getDomainOwner {
  my $domain = shift;
  unless(keys %trueUserDomains){
	#
	# calculate $ptrTrueUserDomainsHash
	#
	my $fileParser = &makeFileParser('/etc/trueuserdomains');
	my $ptrHash = $fileParser->{'PARSE'}->('KEYSEPARATOR'=>':');
	if (ref($ptrHash)=~/HASH/) {
	  %trueUserDomains=(%{$ptrHash});
	}
  }

  return $trueUserDomains{$domain};

}

sub makeAccountNode {

  my ($account, $ptrUserHash, $isReseller) = @_;

  my ($ptrLines, $ptrHash, $ptrQuota,
	  $xmlKey, $userKey, $value, $key, $domain, $item);
  my $home = getHomeDir($account);

  my $root = &makeXmlNode('account');
  $root->{'ATTRIBUTE'}->('name', $account);

  if (-T "$home/.contactemail") {

	my $fileParser = makeFileParser("$home/.contactemail");
	$ptrLines = $fileParser->{'PARSE'}->();
	if (ref($ptrLines) =~ /ARRAY/) {
	  $root->{'ATTRIBUTE'}->('email', $ptrLines->[0]||'');
	}

  } else {
	$root->{'ATTRIBUTE'}->('email', '');
  }

  unless(ref($ptrUserHash) =~ /HASH/) {
	my $fileParser = makeFileParser("$installDir/users/$account");
	$ptrUserHash = $fileParser->{'PARSE'}->('KEYSEPARATOR' => '=');
  }

  if (ref($ptrUserHash) =~ /HASH/) {
	$root->{'ATTRIBUTE'}->('date', $ptrUserHash->{'STARTDATE'}||'');
  }

  unless ($isReseller) {
	my $diskSpace = $ptrQuota->{$account};
	if ($diskSpace =~ /(\d+)/) {
	  $diskSpace = $1*1024*1024;
	  $root->{'ADDCHILD'}->(&makeLimitNode('disk_space', $diskSpace));
	}

	if (ref($ptrUserHash) =~ /HASH/) {

	  # handle limits
	  $mapLimits{'max_traffic'} = 'BWLIMIT';
	  while (($xmlKey,$userKey) = each(%mapLimits)) {
		$value = $ptrUserHash->{$userKey};
		if ($value =~ /\d+/) {
		  $root->{'ADDCHILD'}->(&makeLimitNode($xmlKey, $value));
		} else {
		  $root->{'ADDCHILD'}->(&makeLimitNode($xmlKey));
		}
	  }

	  delete $mapLimits{'max_traffic'}; ## clean map
	}
  }

  #
  # Handle domain
  #
  $domain = &getDomainName($account);

  if ($domain) {

	&printToLog("Domain '$domain' is started");

	$objDumpStatus->{'DOMAIN'}->($domain);
	$objDumpStatus->{'PRINT'}->();

	$item = &makeIpNode($domain, $ptrUserHash);
	$root->{'ADDCHILD'}->($item);

	$item = &makeDomainNode($domain, $account);
	$root->{'ADDCHILD'}->($item);

	&printToLog("Domain '$domain' is dumped");
  }

  return $root;
}

sub makeIpNode {
  my ($domain,$ptrUserHash) = @_;

  my ($ptrRows,$ip,$item,$ptrRow,$dmn);

  unless ( $ip = $servers_ips{$domain}) {
	my $fileParser = makeFileParser("$installDir/accts.db");
	$ptrRows = $fileParser->{'PARSE'}->('VALUESEPARATOR'=>',');
	foreach $ptrRow (@{$ptrRows}) {
	  $dmn = $ptrRow->[0];
	  $dmn=~s/^\s+//;
	  $dmn=~s/\s+$//;
	  if ($ptrRow->[2]=~/(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})/) {
		$ip=$1;
		unless (exists($servers_ips{$domain})) {
		  $servers_ips{$domain}=$ip;
		}
	  }
	}
	unless ($ip = $servers_ips{$domain}) {
	  unless ($ip = $ptrUserHash->{'IP'}) {
		my $zone = "/var/named/$domain.db";
		my $pattern = qr/^\Q$domain\E\.\s+(?:\d+\w?\s+)IN\s+A\s+(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})/;
		if (-T $zone) {
		  if (open(INPUT,"< $zone")) {
			binmode INPUT;
			while (<INPUT>) {
			  if (/$pattern/) {
				$ip=$1;
				last;
			  }
			}
			close(INPUT);
			if ($ip) {
			  $servers_ips{$domain}=$ip;
			}
		  }
		}
	  }
	}
  }
  $item = &makeXmlNode('ip',$ip);

  return $item;
}


sub makeDomainNode {
  my ($domain,$account) = @_;

  my $root = &makeXmlNode('domain');
  $root->{'ATTRIBUTE'}->('name',$domain);

  my(@subdomains,@ftpusersdirs,$subdomain,$acc,
	 $ptrRows,$ptrRow,$path,@dirs,$dir,@subdompatterns,@exclude,
	 $dumpFile,$item,$ptrProftpdRows);

  my $home = getHomeDir($account);

  #
  # subdomain dirs
  #
  unless(ref($ptrUserdomainsHash)=~/HASH/){
	my $fileParser = makeFileParser('/etc/userdomains');
	$ptrUserdomainsHash = $fileParser->{'PARSE'}->('KEYSEPARATOR'=>':');
  }
  if (ref($ptrUserdomainsHash)=~/HASH/) {
	my $pattern = qr/^([^\.]+)\.\Q$domain\E$/;
	while ( ($subdomain,$acc) = each %{$ptrUserdomainsHash} ) {
	  next unless ($acc eq $account);
	  if ($subdomain=~/$pattern/) {
		$dir = $1;
		push @subdomains,$dir;
		push @subdompatterns,qr/^\Q$dir\E(?:\/.*)?/;
	  }
	}
  }
  #
  # end subdomain dirs
  #

  #
  # ftp dirs
  #
  if (-T "/etc/proftpd/$account") {
	my $fileParser = makeFileParser("/etc/proftpd/$account");
	$ptrProftpdRows = $fileParser->{'PARSE'}->('VALUESEPARATOR'=>':',
											   'KEYSEPARATOR'=>'',);
	if (ref($ptrProftpdRows)=~/ARRAY/) {
	  my ($subdompattern,$pattern,$is_subdomain_directory);
	  $pattern = qr/\Q$home\E\/\Q$account\E\/public_html\/(.*)/;
	  foreach $ptrRow (@{$ptrProftpdRows}) {
		next unless ($ptrRow->[4] eq $account);
		$path=$ptrRow->[5];
		if ($path=~/$pattern/) {
		  $dir =$1;
		  $is_subdomain_directory=0;
		  foreach $subdompattern (@subdompatterns) {
			if ($dir=~/$subdompattern/) {
			  $is_subdomain_directory=1;
			  last;
			}
		  }
		  next if ($is_subdomain_directory);
		  push (@ftpusersdirs, $dir);
		}
	  }
	}
  }
  @subdompatterns=();
  #
  # end ftp dirs
  #

  #
  # content of domain
  #
  if (-d "$home/cgi-bin") {
	@exclude = ('cgi-bin');
  } else {
	@exclude = ();
	$dumpFile = &makeDumpFile("$dumpDir/cgi_domain_$account",
							  "$home/public_html/cgi-bin");
	$root->{'ATTRIBUTE'}->('cid_cgi',$dumpFile) if $dumpFile;
  }
  push @exclude,@subdomains,@ftpusersdirs;
  $dumpFile = &makeDumpFile("$dumpDir/docroot_domain_$account",
							"$home/public_html",['.'],\@exclude);
  $root->{'ATTRIBUTE'}->('cid_docroot',$dumpFile) if $dumpFile;

  #
  # end content of domain
  #

  if ($item = &makeAnonFtpdNode($account)) {
	$root->{'ADDCHILD'}->($item);
  }

  &addFtpUserNodes($root,$domain,$account,
				   \@subdomains,\@ftpusersdirs,$ptrProftpdRows);

  if ($noContent) {
	return $root;
  }

  &addProtectedDirs($root,$domain,$account,
					\@subdomains);

  &addSubDomains($root, $domain, $account,
				 \@subdomains);

  &addAddOnDomains($root,$domain,$account);

  addMailNode($root, $domain, $account);

  &addDatabases($root,$account);

  return $root;
}

sub addDatabases {
  my ($root,$account) = @_;

  my ($ptrRow,$database,$item,$dumpFile,$sql,@dbNames);

  unless((ref($wrapUserMysql)=~/HASH/) &&
		 (ref($wrapUserMysql->{'EXECUTE'})=~/CODE/)){

	$wrapUserMysql = &getDbConnect(\%bin,'mysql',$userMysqlDbAddress,
								   $mysqlUserUser, $mysqlUserPw, 'mysql',
								   $mysqlUserServer);

  }

  unless((ref($wrapUserMysql)=~/HASH/) &&
		 (ref($wrapUserMysql->{'EXECUTE'})=~/CODE/)){

	&printToError("Error: unable to connect to 'mysql' database\n",
				  "host: $mysqlUserServer, user: $mysqlUserUser, pass: $mysqlUserUser");
	return;
  }
  unless ($wrapUserMysql->{'PING'}->()) {
	unless ($wrapUserMysql->{'RECONNECT'}->()) {
	  &printToError("Error: unable to connect to 'mysql' database\n",
					"host: $mysqlUserServer, user: $mysqlUserUser, pass: $mysqlUserUser");
	  return;
	}
  }

  $sql="SELECT DISTINCT db FROM db WHERE db like '".$account."_%'";
  if ($wrapUserMysql->{'EXECUTE'}->($sql)) {
	while ($ptrRow = $wrapUserMysql->{'FETCHROW'}->()) {
	  push @dbNames,$ptrRow->[0];
	}
  }
  $wrapUserMysql->{'FINISH'}->();

  foreach $database (@dbNames) {
	$database =~ s/\\_/_/g;
	$item = &makeXmlNode('database');
	$item->{'ATTRIBUTE'}->('name',$database);
	$item->{'ATTRIBUTE'}->('type','mysql');

	printToLog("Started $database dumping...");

	if ($dumpFile = &makeDumpDb($database, 'mysql', 'root')) {
	  $item->{'ATTRIBUTE'}->('cid',$dumpFile);
	}

	addDbUsers($item, $database, $wrapUserMysql);

	$root->{'ADDCHILD'}->($item);

	printToLog("done");
  }
}

sub addSubDomains {

  my($root, $domain, $account, $ptrSubDomains) = @_;

  my($nodeSubDomain, $subdomain, $dumpFile, $subdomainRoot, $fileParser,
	 $ptrRows, $ptrRow, $item, @exclude);

  my $home = getHomeDir($account);


  foreach $subdomain (@{$ptrSubDomains}) {

	&printToLog("Subdomain '$subdomain.$domain' ...", 1);

	$subdomainRoot = "$home/public_html/$subdomain";

	$nodeSubDomain = &makeXmlNode('subdomain');
	$nodeSubDomain->{'ATTRIBUTE'}->('name', $subdomain, 'noencode');
	$root->{'ADDCHILD'}->($nodeSubDomain);

	#
	# content
	#
	if (-d "$subdomainRoot/cgi-bin") {

	  if ($dumpFile = &makeDumpFile("$dumpDir/cgi_subdomain_$subdomain.$domain", $subdomainRoot, 'cgi-bin')) {
		$nodeSubDomain->{'ATTRIBUTE'}->('cid_cgi', $dumpFile);
	  }
	  @exclude=('cgi-bin');

	} else {
	  @exclude=();
	}

	if ($dumpFile = &makeDumpFile("$dumpDir/docroot_subdomain_$subdomain.$domain", $subdomainRoot, ['.'], \@exclude)) {
	  $nodeSubDomain->{'ATTRIBUTE'}->('cid_docroot', $dumpFile);
	}

	#
	# ftp user
	#
	if (-T "/etc/proftpd/$account") {
	  $fileParser = &makeFileParser("/etc/proftpd/$account");
	  $ptrRows = $fileParser->{'PARSE'}->('VALUESEPARATOR' => ':');
	  if (ref($ptrRows) =~ /ARRAY/) {
		foreach $ptrRow (@{$ptrRows}) {
		  if ( ($ptrRow->[0] eq $subdomain) && ($ptrRow->[4] eq $account) ) {
			if ($ptrRow->[1]) {
			  $item = &makeFtpUserNode($subdomain, $ptrRow->[1]);
			  $nodeSubDomain->{'ADDCHILD'}->($item);
			}
			last;
		  }
		}
	  }
	}

	#
	# mail
	#
	addMailNode($nodeSubDomain, "$subdomain.$domain", $account);

	&addAddOnDomains($nodeSubDomain, "$subdomain.$domain", $account);

	&printToLog(' OK');
  }
}

my %apacheAliases = ();

sub addAddOnDomains {
  my ($root,$domain,$account)=@_;
  my ($userdomain,$owner,%userdomains,$maindomain,%aliases,$alias,$item);
  unless (keys %apacheAliases){
	unless(open (INPUT,"/etc/httpd/conf/httpd.conf")){
	  return;
	}
	binmode INPUT;
	my ($state,$vserver,%valiases);
	$state=0;
	while (<INPUT>) {
	  chomp;
	  if ($state==1) {
		if (/^\s*<\/VirtualHost\s*>/) {
		  if ($vserver && (keys %valiases)) {
			push @{$apacheAliases{$vserver}},keys %valiases;
		  }
		  $vserver='';
		  %valiases=();
		  $state=0;
		} elsif (/^\s*ServerName\s+(\S+)/) {
		  $vserver=$1;
		  $vserver=~s/^www\.//;
		} elsif (/^\s*ServerAlias\s+(.*)/) {
		  map {s/^www\.//;$valiases{$_}=1} split(' ',$1);
		}
	  } else {
		if (/^\s*<VirtualHost\s.*>/) {
		  $state=1;
		  $vserver='';
		  %valiases=();
		}
	  }
	}
	close (INPUT);
  }
  unless (exists $apacheAliases{$domain}) {
	return;
  }
  unless(ref($ptrUserdomainsHash)=~/HASH/){
	my $fileParser = &makeFileParser('/etc/userdomains');
	$ptrUserdomainsHash = $fileParser->{'PARSE'}->('KEYSEPARATOR'=>':');
  }

  $maindomain = &getDomainName($account);

  if (ref($ptrUserdomainsHash)=~/HASH/) {
	my $pattern = qr/\Q$maindomain\E$/;
	while (($userdomain,$owner)=each(%{$ptrUserdomainsHash})) {
	  next unless $owner eq $account;
	  next if ($userdomain=~/$pattern/); ## exclude subdomains
	  $userdomains{$userdomain}=1;
	}
  }


  %aliases = map {$_=>1} @{$apacheAliases{$domain}};
  foreach $alias (keys %aliases) {
	next unless exists $userdomains{$alias};
	$item = &makeXmlNode('addondomain',$alias);
	$root->{'ADDCHILD'}->($item);
  }
}

sub addMailNode {
  my ($root, $domain, $account) = @_;
  my ($mailNode,$ptrRows,$ptrRow,$default,$item,$box,
	  $ptrHash,$value,$password,$dumpFile,$ptrAliases,$src,$dst,
	  $from);
  my $home = getHomeDir($account);

  printToLog("Mail for $domain migration started...");
  $mailNode = &makeXmlNode('mail');
  $ptrAliases=[];
  if (-T "/etc/valiases/$domain") {
	my $fileParser = &makeFileParser("/etc/valiases/$domain");
	$ptrAliases=$fileParser->{'PARSE'}->('VALUESEPARATOR'=>':');
	if (ref($ptrAliases)=~/ARRAY/) {
	  foreach $ptrRow (@{$ptrAliases}) {
		if ($ptrRow->[0] eq '*') {
		  if ($ptrRow->[1]=~/([^@]+\@[^@+]+)/) {
			$default=$1;
			last;
		  } elsif ($ptrRow->[2]=~/blackhole/) {
			$default=':blackhole:';
			last;
		  } elsif ($ptrRow->[2]=~/fail/) {
			$default=':fail:'.$ptrRow->[3];
			last;
		  }
		}
	  }
	}
  }

  if ($default) {
	$item = &makeXmlNode('default');
	$mailNode->{'ADDCHILD'}->($item);
	if ($default eq ':blackhole:') {
	  $item->{'ATTRIBUTE'}->('target','ignore');
	} elsif ($default =~ /:fail:(.*)/) {
	  $item->{'ATTRIBUTE'}->('target','fail');
	  $item->{'CONTENT'}->($1);
	} else {
	  $item->{'ATTRIBUTE'}->('target','email');
	  $item->{'CONTENT'}->($default);
	}
  }

  printToLog("Examining $home/etc/$domain");

  if (-d "$home/etc/$domain" && -T "$home/etc/$domain/passwd") {
	my $fileParser = makeFileParser("$home/etc/$domain/passwd");
	$ptrRows=$fileParser->{'PARSE'}->('VALUESEPARATOR'=>':');

	if (ref($ptrRows)=~/ARRAY/) {
	  foreach $ptrRow (@{$ptrRows}) {
		$box = $ptrRow->[0];
		$item = &makeXmlNode('mailname');
		$item->{'ATTRIBUTE'}->('name',$box);
		$mailNode->{'ADDCHILD'}->($item);

		printToLog("migrating " .$box . "@" ."$domain");

		#
		# password
		#
		$password='';
		my $shadowParser = makeFileParser("$home/etc/$domain/shadow");
		$ptrHash = $shadowParser->{'PARSE'}->('KEYSEPARATOR'=>':',
											  'VALUESEPARATOR'=>'');

		if (ref($ptrHash)=~/HASH/) {
		  if ($value=$ptrHash->{$box}) {
			($password)=split(':',$value);
		  }
		}
		$item->{'ATTRIBUTE'}->('password',$password);
		#
		# end password
		#

		#
		# quota
		#
		if ( -T "$home/etc/$domain/quota") {
		  my $quotaParser = makeFileParser("$home/etc/$domain/quota");
		  $ptrHash = $quotaParser->{'PARSE'}->('KEYSEPARATOR' => ':', 'VALUESEPARATOR' => '');
		  if (ref($ptrHash) =~ /HASH/ && ($value = $ptrHash->{$box})) {
			$item->{'ATTRIBUTE'}->('quota', $value);
		  }
		}
		#
		# end quota
		#
		if ($dumpFile = &makeDumpFile("$dumpDir/mailname_$box.$domain",
									  "$home/mail/$domain/$box")) {
		  $item->{'ATTRIBUTE'}->('cid',$dumpFile);
		}
	  }
	}
  }
  if (@{$ptrAliases}) {
	addForwarders($mailNode, $ptrAliases);
	addAutoresponders($mailNode, $account, $ptrAliases);
  }

  $root->{'ADDCHILD'}->($mailNode);
  printToLog("Mail for $domain migration finished...");
}

sub addAutoresponders {
  my ($mailNode, $account, $ptrAliases) = @_;
  my $home = getHomeDir($account);

  foreach my $ptrRow (@{$ptrAliases}) {

	my $src = $ptrRow->[0];
	my $dst = $ptrRow->[1];
	chomp $src;

	next unless ($dst =~ /^\s*\|.*\/cpanel\/bin\/autorespond/);
	next unless ($src =~ /^([^@]+@[^@]+)$/);
	next unless (open(INPUT, "$home/.autorespond/$src"));
	binmode INPUT;

	my $autoresponderNode = makeXmlNode('autoresponder');
	$mailNode->{'ADDCHILD'}->($autoresponderNode);
	$src =~ /^([^@]+)@/;
	$autoresponderNode->{'ATTRIBUTE'}->('mailname', $1);

	while (<INPUT>) {
	  chomp;
	  last unless $_;

	  if (/^From:.+<(.+)>/) {
		$autoresponderNode->{'ATTRIBUTE'}->('from', $1) if ($src ne $1);
	  } elsif (/^Content-type:\s+text\/([^;]+);\s+charset=(.+)/) {
		$autoresponderNode->{'ATTRIBUTE'}->('type', $1);
		$autoresponderNode->{'ATTRIBUTE'}->('charset', $2);
	  } elsif (/^Subject:\s*(.*)/) {
		$autoresponderNode->{'ATTRIBUTE'}->('subject', $wrapBase64->{'ENCODE'}->($1));
	  }
	}

	$autoresponderNode->{'CONTENT'}->($wrapBase64->{'ENCODE'}->(join('',(<INPUT>))));

	close(INPUT);
  }
}

sub addForwarders {
  my ($mailNode, $ptrAliases) = @_;

  foreach my $ptrRow (@{$ptrAliases}) {

	my $src = $ptrRow->[0];
	my $dst = $ptrRow->[1];

	next unless ($src =~ /^([^@]+)@[^@]+$/);
	$src = $1;
	next unless ($dst=~/^\s+([^@|]+@[^@\s]+)/);
	$dst = $1;

	my $forwardNode = makeXmlNode('forward');
	$mailNode->{'ADDCHILD'}->($forwardNode);
	$forwardNode->{'ATTRIBUTE'}->('mailname', $src);
	$forwardNode->{'ATTRIBUTE'}->('redirect', $dst);
  }
}

sub addProtectedDirs {

  my ($root, $domain, $account, $ptrSubDomains) = @_;

  my ($cmd, $exclude, $pdir, $find, $item, $title, $ptrHash,
	  $nodePdir, $name, $password);

  my $home = getHomeDir($account);
  my $rootDir = "$home/.htpasswds";
  unless(-d $rootDir){
	return;
  }

  $find = $bin{'find'}||'find';
  $cmd = " cd $rootDir; $find . ";
  if (@{$ptrSubDomains}) {
	$exclude = join(' -o ',map {"-path './$_' -prune"} @{$ptrSubDomains});
	$cmd .= "\\( $exclude \\) -o";
  }
  $cmd .= " \\( -type f -name passwd -printf '\%h\\n' \\)";
  foreach $pdir (`$cmd`) {
	chomp $pdir;
	$pdir=~s/^\.\///;
	$nodePdir = &makeXmlNode('pdir');
	$nodePdir->{'ATTRIBUTE'}->('name',$pdir);
	$root->{'ADDCHILD'}->($nodePdir);

	$title = '';
	if (-T "$home/public_html/$pdir/.htaccess") {
	  my $fileParser = makeFileParser("$home/public_html/$pdir/.htaccess");
	  $ptrHash = $fileParser->{'PARSE'}->();

	  if (ref($ptrHash)=~/HASH/) {
		if (exists($ptrHash->{'AuthName'})) {
		  if ($ptrHash->{'AuthName'}=~/"([^"]+)"/) {
			$title = $1;
		  }
		}
	  }
	}

	$nodePdir->{'ATTRIBUTE'}->('title',$title);
	my $fileParser = makeFileParser("$rootDir/$pdir/passwd");
	$ptrHash = $fileParser->{'PARSE'}->('KEYSEPARATOR'=>':');
	if (ref($ptrHash)=~/HASH/) {
	  while (($name,$password)=each(%{$ptrHash})) {
		$item = &makeXmlNode('pduser');
		$nodePdir->{'ADDCHILD'}->($item);
		$item->{'ATTRIBUTE'}->('name',$name);
		if ($password !~ /^[\x20-\x7f]*$/) {
		  $item->{'ATTRIBUTE'}->('password', $wrapBase64->{'ENCODE'}->($password));
		  $item->{'ATTRIBUTE'}->('encoding', 'base64');
		} else {
		  $item->{'ATTRIBUTE'}->('password',$password);
		}
	  }
	}
  }
}

sub addFtpUserNodes
  {
	my ($root, $domain, $account, $ptrSubDomains, $ptrFtpDirs, $ptrProftpdRows) = @_;

	my ($password, @users);

	my $home    = getHomeDir($account);
	my $pattern = qr/\Q$home\E\/\Q$account\E\/public_html\/(.*)/;

	foreach my $ptrRow (@{$ptrProftpdRows}) {

	  if ($ptrRow->[5] eq $account) {
		if ($ptrRow->[6] =~ /$pattern/) {
		  my $subdir = $1;
		  my $pass = ($ptrRow->[1] ne '*') ? $ptrRow->[1] : '';

		  if (grep(($_ eq $subdir), @{$ptrFtpDirs})) {
			push @users, [$ptrRow->[0], $pass, $subdir];
		  }
		}
	  }

	  if (!$password && ($ptrRow->[0] eq $account)) {
		$password = $ptrRow->[1];
	  }
	}

	if ('*' eq $password) {
	  $password = getSystemPassword($account);
	  if ('*' eq $password) {
		$password = '';
	  }
	}

	# makes main domains's ftp user node
	$root->{'ADDCHILD'}->(makeFtpUserNode($account, $password));

	# other ftp users of domain
	foreach my $ptrRow (@users) {

	  my ($acc, $pass, $path) = @{$ptrRow};
	  my $ftpUserNode = makeFtpUserNode($acc, $path);

	  if ($path) {
		$ftpUserNode->{'ATTRIBUTE'}->('directory', $path);
		if (my $dumpFile = makeDumpFile("$dumpDir/ftpuser_$acc.$account", "$home/public_hmtl/$path")) {
		  $ftpUserNode->{'ATTRIBUTE'}->('cid', $dumpFile);
		}
	  }

	  $root->{'ADDCHILD'}->($ftpUserNode);
	}
  }

sub makeFtpUserNode
  {
	my ($account, $password) = @_;
	my $item = &makeXmlNode('ftpuser');
	$item->{'ATTRIBUTE'}->('name', $account);
	if ($password) {
	  $item->{'ATTRIBUTE'}->('password', $password);
	}
	return $item;
  }

sub makeAnonFtpdNode {
  my $account = shift;
  my ($node,$publicFtp,$pureFtpd);
  my $home = getHomeDir($account);

  $publicFtp="$home/public_ftp";
  unless(-d $publicFtp){
	return undef;
  }
  $node = &makeXmlNode('anonftp');

  if ($dumpFile = &makeDumpFile("$dumpDir/anonftp_pub_$account",
								$publicFtp,['.'],['incoming'])) {
	$node->{'ATTRIBUTE'}->('cid',$dumpFile);
  }
  if ($dumpFile = &makeDumpFile("$dumpDir/anonftp_incoming_$account",
								"$publicFtp/incoming")) {
	$node->{'ATTRIBUTE'}->('cid_incoming',$dumpFile);
  }
  #
  # prepare anonftps
  #
  $pureFtpd = '/etc/pure-ftpd';
  if (-d $pureFtpd) {
	unless(keys %anonftps){
	  my ($link,$linkTo,$fullPath);
	  if (opendir(FTPD,$pureFtpd)) {
		foreach $link (readdir(FTPD)) {
		  $fullPath = "$pureFtpd/$link";
		  next unless (-l $fullPath);
		  $linkTo = readlink($fullPath);
		  if ($linkTo=~/\Q$home\E\/([^\/]+)\/public_ftp/) {
			push @{$anonftps{$1}},$link;
		  }
		}
		closedir(FTPD);
	  }
	}
  }
  #
  # end prepare anonftps
  #

  #
  # check permissions
  #
  if (exists($anonftps{$account}) &&
	  ( (stat($publicFtp))[2] & 05)
	 ) {

	$node->{'ATTRIBUTE'}->('pub','true');

	if ( (stat("$publicFtp/incoming"))[2] & 03 ) {

	  $node->{'ATTRIBUTE'}->('incoming','true');

	}
  }

  #
  # end check permissions
  #

  return $node;
}

sub makePermissionNode {
  my($name,$value)=@_;
  my $item = &makeXmlNode('permission',$value);
  $item->{'ATTRIBUTE'}->('name',$name);
  return $item;
}

sub makeLimitNode {
  my($name,$value)=@_;
  unless(defined($value) && ($value !~ '/^\s*$/')){
	$value = '-1';
  }
  my $item = &makeXmlNode('limit',$value);
  $item->{'ATTRIBUTE'}->('name',$name);
  return $item;
}

sub makeSysuserNode {
  my ($account,$password,$pswdType,$quota,$shell)=@_;
  my ($ptrHash,$item);
  if (ref($account)=~/HASH/) {
	$ptrHash=$account;
	$account=$ptrHash->{'ACCOUNT'}||$ptrHash->{'NAME'};
	$password=$ptrHash->{'PASSWORD'};
	$pswdType=$ptrHash->{'PSWDTYPE'};
	$shell=$ptrHash->{'SHELL'};
	$quota=$ptrHash->{'QUOTA'};
  }
  $item = &makeXmlNode('sysuser');
  $item->{'ATTRIBUTE'}->('name',$account);
  if ($quota) {
	$item->{'ATTRIBUTE'}->('quota',$quota);
  }
  if ($shell) {
	$item->{'ATTRIBUTE'}->('shell',$shell);
  }
  $item->{'ADDCHILD'}->(&makePasswordNode($password,$pswdType));
  return $item;
}

sub getCpanelVersion {
  my ($engineRoot, $util);
  if ($engineRoot = $cPanelConfig{'engineroot'}) {
	unless (-d $engineRoot) {
	  return;
	}
	$util="$engineRoot/cpanel";
	unless (-x $util) {
	  return;
	}
	my @outs = `$util`;
	foreach my $line (@outs) {
	  chomp $line;
	  if ($line=~/cPanel\s+\[(\d+)\.(\d+).\d+/) {
		if ($1 ne '9' && $1 ne '10') {
		  return;
		}
		return "$1.$2";
	  }
	}
  } elsif (-f "$installDir/cpanel.config") {
	return '9';
  }
}

sub printAgentStatus {
  my $dumpFile = shift;
  my $statusFile = shift;
  my $outFh = shift || *STDOUT;
  my $rootName = shift || 'agentstatus';

  my ($root,$util,$out,$item);

  $root = &makeXmlNode($rootName);

  unless (getCpanelVersion() && defined $bin{'iconv'}) {
	if (defined $bin{'iconv'}) {
	  $item = &makeXmlNode('wrong-platform', '');
	} else {
	  $item = &makeXmlNode('wrong-platform', 'no iconv found on the source host');
	}

	$root->{'ADDCHILD'}->($item);
	&printXml($root,$outFh,$rootName.'.dtd');
	return;
  }

  $item = &makeXmlNode('dumpfile',$dumpFile);
  $root->{'ADDCHILD'}->($item);
  $item = &makeXmlNode('statusfile',$statusFile);
  $root->{'ADDCHILD'}->($item);

  &printXml($root,$outFh,$rootName.'.dtd');
}


sub loadMainConfig
  {
	my $pathToConf = shift || '';
	my $ptrHash    = shift;

	unless (-T $pathToConf) {
	  foreach my $prefix ('./', '/var/cpanel/') {
		$pathToConf = $prefix . 'cpanel.config';
		if (-T $pathToConf) {
		  last;
		}
	  }
	}

	if (-T $pathToConf) {
	  if (ref($ptrHash) =~ /HASH/) {
		if ($pathToConf =~ /(.+)\/[^\/]+/) {
		  $ptrHash->{'installDir'} = $1;
		}
		if (open(CF, "<$pathToConf")) {
		  binmode CF;
		  my ($key, $value);
		  while (my $line = <CF>) {
			chomp $line;
			next unless $line;
			next if ($line =~ /^#/);
			($key, $value) = split(/\s*=\s*/, $line, 2);
			$key =~ s/^\s+//;
			$ptrHash->{$key} = $value;
		  }
		  close (CF);
		}
	  }
	} else {
	  $pathToConf = '';
	}
	return $pathToConf;
  }

sub printHelp {

  print <<"HELP";

Agent of migration from 'cPanel 9' to 'Plesk 7'

Usage:
$0 <options>

Options:
-c |--config=<path>        path to cpanel.conf

-s |--get-status           get status of the agent
-dc|--dump-accounts=<list> a coma separated list of resellers to dump
-dd|--dump-domains=<list>  a coma separated list of customers to dump
-da|--dump-all             make a full dump

-lc|--get-content-list     get list of content files
-nc|--no-content           do not make content files
-nz|--no-compress          do not compress content files

-o |--output=<path>        redirect output to file

-h |--help                 this help
-d |--debug=<level>        level of debug messages
-v |--version

HELP
}


